/*
** mm_cgen: Matrix Multiply Core Generator, 
**    the PHiPAC matrix-matrix multiply core code generator.
** 
** $Revision: 1.27 $
**
** See the man page mm_cgen.1 for details.
**
** Written by: 
**      Jeff Bilmes <bilmes@icsi.berkeley.edu>
**      http://www.icsi.berkeley.edu/~bilmes/phipac
**
** 
** "Copyright (c) 1995 The Regents of the University of California.  All
** rights reserved."  Permission to use, copy, modify, and distribute
** this software and its documentation for any purpose, without fee, and
** without written agreement is hereby granted, provided that the above
** copyright notice and the following two paragraphs appear in all copies
** of this software.
** 
** IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
** DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
** OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
** CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**  
** THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
** INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
** AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
** ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
** PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
** 
*/


static const char rcsid[] = "$Header: /db/bilmes/phipac/mm_gen/RCS/mm_cgen.c,v 1.27 1998/06/08 08:05:09 bilmes Exp $";
static const char rcsrevision[] = "$Revision: 1.27 $";


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>


#include <assert.h>



/**********************************************************************
 *			TYPEDEFS AND ENUMS                            *
 **********************************************************************/

typedef enum { True = 1, False = 0 } bool;

typedef enum {     Zero, /* the value is fixed at zero */
		   One,  /* the value is fixed at one */
		   nOne, /* the value is fixed at negative one */
		   Fix,  /* the value is fixed at some constant */
		   Arb   /* the value is parameterized  */
	       } scaleType;

/* matrix transposition */
typedef enum { Normal, Transpose } op;

/* controls the code generation operation order */
typedef enum { Astripe, 
	       Bstripe, 
	       minStripe 
} regBlockStripe;

/* software pipelining options */
typedef enum {
     OneStage,  /* i.e., no software pipelining */
     TwoStageLoadMul, /* Two stage, {Load-Mul, Add} */
     TwoStageMulAdd,  /* Two stage, {Load, Mul-Add} */
     ThreeStage,  /* Three stage, {Load, Mul, Add} */
} SoftPipeType;

/**********************************************************************
 *			COMMAND LINE ARGUMENTS                        *
 **********************************************************************/

/* where to put the resulting routine */
char *outfile=NULL;
FILE *outf = NULL;

/* the C code name of the resulting routine */
char routine_name[BUFSIZ] = { '\0' };

/* The blocking parameters. We must have
   the Level 0 parameters set. If the level 1 parameters
   are set, we produce a routine for a fixed matrix size. */
size_t M0=0,K0=0,N0=0;
size_t M1=0,K1=0,N1=0;
#define L1SET (M1 != 0)
size_t Mc,Kc,Nc; /* the L1 fringes if L1SET is true */

/* determins the operation on either the A or B matrix. */
op opA = Normal;
op opB = Normal;

/* Sotware Pipelining Option */
SoftPipeType softpipe = OneStage;

/* Hold a stripe of the (A, B, or min(M0,N0)) matrix register block
   in registers while stepping through the other matrix block
   loading one register at a time. */
regBlockStripe rbStripe = Bstripe;

/* True if we do the register block fringes, False if we don't. */
bool do_m_fringe=True;
bool do_k_fringe=True;
bool do_n_fringe=True;

/* true if we compute linear rather than logarithmic fringe code */
bool linear_m_fringe = False;
bool linear_k_fringe = False;
bool linear_n_fringe = False;


/*
 * This next option only pertains to when we are not
 * generating fringe code. 
 * True if we are to never generate code that contains
 * loops with loop head branches (i.e., should we use for() loops
 * which contain an intial compare and branch or instead use a 
 * do {} while (); loop. This gets autmatically set when
 * we're doing a fixed size routine and may be optionally
 * set when we're doing a general no-fringe routine, but may
 * not be set when we're doing a general fringe routine.
 * When this option is true, the routine is correct
 * only when M>=M0,K>=K0, and N>=N0.
 */
bool no_loop_head_branch=False;

scaleType alpha = One, beta = One;
char *alpha_val = "1.0", *beta_val = "1.0";

/* The following are the C floating point types that will be used.
   They may be safely changed if needed for a particular compiler.  */
const char* fp_types[] = {   
  "float",   /* presumably IEEE single prec */
  "double",  /* presumably IEEE double prec */
  "long double" /* presumably IEEE double extended prec */
};
const char fp_chars[] = { 'f', 'd', 'l' };
#define FP_TYPE_INVALID (-1)
#define FP_TYPE_FLOAT 0
#define FP_TYPE_DOUBLE 1
#define FP_TYPE_LONG_DOUBLE 2
#define FP_TYPE_DEFAULT FP_TYPE_FLOAT
int fp_type_num = FP_TYPE_INVALID;
int fp_atype_num = FP_TYPE_INVALID; /* accumulator type */
int fp_dtype_num = FP_TYPE_INVALID; /* destination type */
const char* fp_type;
const char* fp_atype; /* accumulator type */
const char* fp_dtype; /* destination type */
char  fp_char;
char  fp_achar;
char  fp_dchar;


/*
 * Code generation Nesting parameters 
 */
#define DEFAULT_SPACE_PER_NEST 3
int nesting_level=0;
int space_per_nest = DEFAULT_SPACE_PER_NEST;
char space_char = ' ';


/*
 * If (dummy_size_args==true and L1SET), routine interface always
 * has size arguments, even when useless under L1SET.
 */
bool dummy_size_args=False;

/**********************************************************************
 *			MISCELLANEOUS ROUTINES                        * 
 **********************************************************************/

#define MIN(a,b) ((a)<(b)?(a):(b))

#define SP_ADJ_K(K) ((softpipe == OneStage) ? (K) : ((softpipe == TwoStageMulAdd || softpipe == TwoStageLoadMul) ? ((K)+1) : ((K)+2)))


#define DO_ANY_FRINGE (do_m_fringe == True || do_k_fringe == True || do_n_fringe == True)

/*
 * Nest routines.
 */
void nst() { nesting_level++; }
void unst() { nesting_level--; }
void idnt()
{
  int i;
  for (i=0;i<space_per_nest*nesting_level;i++)
    fprintf(outf,"%c",space_char);
}


/* Returns true of arg is a 2^k for some k */
int pow_of_2(int arg)
{
  int i;
  int bits = 0;
  if (arg < 0) return 0;
  for (i=0;i<sizeof(arg)*8;i++) {
    bits += (arg & 0x1);
    if (bits > 1) return 0;
    arg >>= 1;
  }
  return 1;
}

/* returns the highest bit of u */
int highbit( int u )
{
  int res = 1;
  if (u == 0)
    return 0;
  while (u >>= 1) {
    res <<= 1;
  }
  return res;
}




/**********************************************************************
 *			COMMAND LINE PARSING                          *
 **********************************************************************/


void
usage(char *name)
{
  fprintf(stderr,"usage: %s [OPTIONS]\n",name);
  fprintf(stderr,"where [OPTIONS] include:\n");  

  fprintf(stderr,"Semantics options:\n");
  fprintf(stderr,"  -opA [N|T]                 : A matrix op. Normal|Transpose\n");
  fprintf(stderr,"  -opB [N|T]                 : B matrix op. Normal|Transpose\n");
  fprintf(stderr,"  -no_m_fringe               : don't generate an M reg block fringe\n");
  fprintf(stderr,"  -no_k_fringe               : don't generate a K reg block fringe\n");
  fprintf(stderr,"  -no_n_fringe               : don't generate an N reg block fringe\n");
  fprintf(stderr,"  -no_fringes                : don't generate an M,K, or N reg block fringes\n");
  fprintf(stderr,"  -no_loop_head_branch       : Use do-while loops where possible, requiring M>=M0,K>=K0,N>=K0.\n");
  fprintf(stderr,"  -alpha [<val>|c]           : fix alpha at value <val> or arbitrary\n");
  fprintf(stderr,"  -beta [<val>|c]            : fix beta at value <val> or arbitrary\n");


  fprintf(stderr,"Optimization options:\n");
  fprintf(stderr,"  -l0 M0 K0 N0               : register (L0) blocking parameters\n");
  fprintf(stderr,"  -l1 M1 K1 N1               : L1 blocking parameters\n");
  fprintf(stderr,"  -sp [1|2lm|2ma|3]          : software pipelining options\n");
  fprintf(stderr,"  -holdstripe [A|B|m]        : hold regs. stripe A or B (default), or min(M0,N0) (-sp 1 only)\n");
  fprintf(stderr,"  -lin_m_fringe               : generate linear rather than log m-fringe code stripes\n");
  fprintf(stderr,"  -lin_k_fringe               : generate linear rather than log k-fringe code stripes\n");
  fprintf(stderr,"  -lin_n_fringe               : generate linear rather than log n-fringe code stripes\n");
  fprintf(stderr,"  -lin_fringe                 : generate linear rather than log fringe code stripes\n");


  fprintf(stderr,"Precision options:\n");
  fprintf(stderr,"  -prec [single|double|ldouble]  : Precision\n");
  fprintf(stderr,"  -sprec [single|double|ldouble] : Source Precision\n");
  fprintf(stderr,"  -aprec [single|double|ldouble] : Accumulator Precision\n");
  fprintf(stderr,"  -dprec [single|double|ldouble] : Destination Precision\n");

  fprintf(stderr,"Code generation & Misc. options:\n");
  fprintf(stderr,"  -help                      : Print this message\n"); 
  fprintf(stderr,"  -file name                 : Write to file 'name'\n"); 
  fprintf(stderr,"  -routine_name name         : Name of routines\n");
  fprintf(stderr,"  -spacechar c               : char to use as space\n");
  fprintf(stderr,"  -numspaces i               : spaces per nest\n");
  fprintf(stderr,"  -version                   : print version and exit.\n");

}

/*
 * Check if n additional parameters are available.
 */
void
check_remaining_params(int argc,int i,int n)
{
  if ((i+n+1)>argc) {
    fprintf(stderr,"Error: too few command line parameters.\n");
    exit(-1);
  }
}

void
parse_args(int argc,char*argv[])
{
  int i;
  if (argc == 1) {
    usage(argv[0]);
    exit(-1);
  }
  /* args initialize */
  outf = stdout;
  for (i=1;i<argc;i++) {
    if (argv[i][0] == '-') {
      if (!strcmp(&(argv[i][1]),"file")) {
	check_remaining_params(argc,i,1);
        outfile = argv[++i];
	if (!strcmp(outfile,"-")) {
	  /* do nothing */
	} else {
	  if ((outf = fopen(outfile,"w")) == NULL) {
	    fprintf(stderr,"Argument Error: Unable to open file %s\n",outfile);
	    usage(argv[0]);
	    exit(1);
	  }
	}
      } else if (!strncmp(&(argv[i][1]),"l0",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,3);
	if (
	    ((M0 = atoi(argv[++i])) <= 0)
	      ||
	    ((K0 = atoi(argv[++i])) <= 0)
	      ||
	    ((N0 = atoi(argv[++i])) <= 0)
	      ) {
	    fprintf(stderr,"Argument Error: Invalid l0 blocking values.\n");
	    usage(argv[0]);
	    exit(-1);
	  }
      } else if (!strncmp(&(argv[i][1]),"l1",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,3);
	if (
	    ((M1 = atoi(argv[++i])) <= 0)
	      ||
	    ((K1 = atoi(argv[++i])) <= 0)
	      ||
	    ((N1 = atoi(argv[++i])) <= 0)
	      ) {
	    fprintf(stderr,"Argument Error: Invalid l1 blocking values.\n");
	    usage(argv[0]);
	    exit(-1);
	  }
      } else if (!strncmp(&(argv[i][1]),"sp",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"1")) {
	  softpipe = OneStage;
	} else if (!strcmp(argv[i],"2lm")) {
	  softpipe = TwoStageLoadMul;
	} else if (!strcmp(argv[i],"2ma")) {
	  softpipe = TwoStageMulAdd;
	} else if (!strcmp(argv[i],"3")) {
	  softpipe = ThreeStage;
	} else {
	  fprintf(stderr,"Argument Error: Invalid sp option.\n");
	  usage(argv[0]);
	  exit(-1);
	}
      } else if (!strncmp(&(argv[i][1]),"opA",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"N") || !strcmp(argv[i],"n")) { 
	  opA = Normal;
	} else if (!strcmp(argv[i],"T") || !strcmp(argv[i],"t")) { 
	  opA = Transpose;
	} else {
	  fprintf(stderr,"Argument Error : Invalid option (%s), ignored\n",argv[i]);
	  usage(argv[0]);
	  exit(-1);
	}
      } else if (!strncmp(&(argv[i][1]),"opB",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"N") || !strcmp(argv[i],"n")) { 
	  opB = Normal;
	} else if (!strcmp(argv[i],"T") || !strcmp(argv[i],"t")) { 
	  opB = Transpose;
	} else {
	  fprintf(stderr,"Argument Error:Invalid option (%s), ignored\n",argv[i]);
	  usage(argv[0]);
	  exit(-1);
	}
      } else if (!strncmp(&(argv[i][1]),"holdstripe",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"A") || !strcmp(argv[i],"a")) { 
	  rbStripe  = Astripe;
	} else if (!strcmp(argv[i],"B") || !strcmp(argv[i],"b")) { 
	  rbStripe  = Bstripe;	  
	} else if (!strcmp(argv[i],"M") || !strcmp(argv[i],"m")) { 
	  rbStripe  = minStripe;
	} else {
	  fprintf(stderr,"Argument Error:Invalid option (%s), ignored\n",argv[i]);
	  usage(argv[0]);
	  exit(-1);
	}
      } else if (!strncmp(&(argv[i][1]),"alpha",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"c")) {
	  alpha = Arb;
	} else {
	  char *ptr;
	  double tmp;
	  tmp = strtod(argv[i],&ptr);
	  if (ptr == argv[i]) {
	    fprintf(stderr,"Argument Error: Invalid numeric parameter (%s)\n",argv[i]);
	    usage(argv[0]);
	    exit(-1);
	  }
	  alpha_val = argv[i];
	  if (tmp == 1.0)
	    alpha = One;
	  else if (tmp == -1.0)
	    alpha = nOne;
	  else if (tmp == 0.0) {
	    fprintf(stderr,"Argument Error: Fixed zero value for alpha nulify's operation.\n");
	    usage(argv[0]);
	    exit(-1);	    
	  } else
	    alpha = Fix;
	}
      } else if (!strncmp(&(argv[i][1]),"beta",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"c")) {
	  beta = Arb;
	} else {
	  char *ptr;
	  double tmp;
	  tmp = strtod(argv[i],&ptr);
	  if (ptr == argv[i]) {
	    fprintf(stderr,"Argument Error: Invalid numeric parameter (%s)\n",argv[i]);
	    usage(argv[0]);
	    exit(-1);
	  }
	  beta_val = argv[i];
	  if (tmp == 1.0)
	    beta = One;
	  else if (tmp == 0.0)
	    beta = Zero;
	  else 
	    beta = Fix;
	}
      } else if (!strncmp(&(argv[i][1]),"lin_m_fringe",strlen(&(argv[i][1])))) {
	linear_m_fringe = True;
      } else if (!strncmp(&(argv[i][1]),"lin_k_fringe",strlen(&(argv[i][1])))) {
	linear_k_fringe = True;
      } else if (!strncmp(&(argv[i][1]),"lin_n_fringe",strlen(&(argv[i][1])))) {
	linear_n_fringe = True;
      } else if (!strncmp(&(argv[i][1]),"lin_fringe",strlen(&(argv[i][1])))) {
	linear_m_fringe = linear_k_fringe = linear_n_fringe = True;
      } else if (!strncmp(&(argv[i][1]),"no_fringes",strlen(&(argv[i][1])))) {
	do_m_fringe = do_k_fringe = do_n_fringe = False;
      } else if (!strncmp(&(argv[i][1]),"no_m_fringe",strlen(&(argv[i][1])))) {
	do_m_fringe = False;
      } else if (!strncmp(&(argv[i][1]),"no_k_fringe",strlen(&(argv[i][1])))) {
	do_k_fringe = False;
      } else if (!strncmp(&(argv[i][1]),"no_n_fringe",strlen(&(argv[i][1])))) {
	do_n_fringe = False;
      } else if (!strncmp(&(argv[i][1]),"no_loop_head_branch",strlen(&(argv[i][1])))) {
	no_loop_head_branch = True;
      } else if (!strncmp(&(argv[i][1]),"dummy_size_args",strlen(&(argv[i][1])))) {
	dummy_size_args = True;
      } else if (!strncmp(&(argv[i][1]),"routine_name",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	strncpy(routine_name,argv[i],MIN(BUFSIZ,strlen(argv[i])));
      } else if ((!strncmp(&(argv[i][1]),"precision",strlen(&(argv[i][1])))) ||
		 (!strncmp(&(argv[i][1]),"sprecision",strlen(&(argv[i][1]))))) {
	check_remaining_params(argc,i,1);
	if (!strncmp(argv[i+1],"single",strlen(argv[i+1])) ||
	    !strncmp(argv[i+1],"float",strlen(argv[i+1])))
	  fp_type_num = FP_TYPE_FLOAT;
	else if (!strncmp(argv[i+1],"double",strlen(argv[i+1])))
	  fp_type_num = FP_TYPE_DOUBLE;
	else if (!strncmp(argv[i+1],"ldouble",strlen(argv[i+1])))
	  fp_type_num = FP_TYPE_LONG_DOUBLE;
	else {
	  fprintf(stderr,"Argument Error: Invalid precision (%s).\n",argv[i+1]);
	  usage(argv[0]);
	  exit(-1);
	}
	i++;
      } else if (!strncmp(&(argv[i][1]),"aprecision",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	if (!strncmp(argv[i+1],"single",strlen(argv[i+1])) ||
	    !strncmp(argv[i+1],"float",strlen(argv[i+1])))
	  fp_atype_num = FP_TYPE_FLOAT;
	else if (!strncmp(argv[i+1],"double",strlen(argv[i+1])))
	  fp_atype_num = FP_TYPE_DOUBLE;
	else if (!strncmp(argv[i+1],"ldouble",strlen(argv[i+1])))
	  fp_atype_num = FP_TYPE_LONG_DOUBLE;
	else {
	  fprintf(stderr,"Argument Error: Invalid precision (%s).\n",argv[i+1]);
	  usage(argv[0]);
	  exit(-1);
	}
	i++;
      } else if (!strncmp(&(argv[i][1]),"dprecision",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	if (!strncmp(argv[i+1],"single",strlen(argv[i+1])) ||
	    !strncmp(argv[i+1],"float",strlen(argv[i+1])))
	  fp_dtype_num = FP_TYPE_FLOAT;
	else if (!strncmp(argv[i+1],"double",strlen(argv[i+1])))
	  fp_dtype_num = FP_TYPE_DOUBLE;
	else if (!strncmp(argv[i+1],"ldouble",strlen(argv[i+1])))
	  fp_dtype_num = FP_TYPE_LONG_DOUBLE;
	else {
	  fprintf(stderr,"Argument Error:Invalid precision (%s), ignored\n",argv[i+1]);
	  usage(argv[0]);
	  exit(-1);
	}
	i++;
      } else if (!strncmp(&(argv[i][1]),"spacechar",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	space_char = *argv[++i];
      } else if (!strncmp(&(argv[i][1]),"numspaces",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	space_per_nest = atoi(argv[++i]);
      } else if (!strncmp(&(argv[i][1]),"help",strlen(&(argv[i][1])))) {
	usage(argv[0]);
	exit(0);
      } else if (!strncmp(&(argv[i][1]),"-help",strlen(&(argv[i][1])))) {
	usage(argv[0]);
	exit(0);
      } else if (!strncmp(&(argv[i][1]),"version",strlen(&(argv[i][1])))) {
	fprintf(stderr,"%s: %s\n",argv[0],rcsrevision);
	exit(0);
      } else  {
        fprintf(stderr,"Argument Error: Unknown or invalid argument (%s)\n",argv[i]);
	usage(argv[0]);
	exit(-1);
      }
    }
  }

  /* finish initialize & check */
  if (fp_type_num == FP_TYPE_INVALID)
    fp_type_num = FP_TYPE_DEFAULT;
  if (fp_atype_num == FP_TYPE_INVALID)
    fp_atype_num = fp_type_num;
  if (fp_dtype_num == FP_TYPE_INVALID)
    fp_dtype_num = fp_type_num;


  fp_type = fp_types[fp_type_num];
  fp_atype = fp_types[fp_atype_num];
  fp_dtype = fp_types[fp_dtype_num];
  fp_char = fp_chars[fp_type_num];
  fp_achar = fp_chars[fp_atype_num];
  fp_dchar = fp_chars[fp_dtype_num];

  if (M0 == 0 || K0 == 0 || N0 == 0) {
    usage(argv[0]);
    exit(-1);
  }
  if (strlen(routine_name) == 0)
    sprintf(routine_name,"mul_m%c%sm%c%s_m%c",
	    fp_char,(opA==Normal?"":"t"),
	    fp_char,(opB==Normal?"":"t"),
	    fp_dchar);

  if (L1SET) {
    Mc = M0*M1;
    Kc = K0*K1;
    Nc = N0*N1;

    /* we don't do fringes when we do a constant routine */
    do_m_fringe = do_k_fringe = do_n_fringe = False;
    no_loop_head_branch = True;
  }

  if (no_loop_head_branch && DO_ANY_FRINGE) {
    fprintf(stderr,"Argument Error: can't specify -no_loop_head_branch with fringe code.\n");
    usage(argv[0]);
    exit(-1);
  }

}


/**********************************************************************
 *			CODE GENERATION (i.e., the meat)              *
 **********************************************************************/


/***************/
/* K-LOOP Code */
/***************/

void
generate_core_k_loop_open(size_t lM0,size_t lK0, size_t lN0)
{
  size_t m,n;
  

  /* declare the core K-loop variables */

  idnt();fprintf(outf,"register %s ",fp_type);  
  for (n=0;n<lN0;n++)    
    fprintf(outf,"_b%d%s",n,(n<(lN0-1)?",":";\n"));
  idnt();
  fprintf(outf,"register %s ",fp_type);
  for (m=0;m<lM0;m++)
    fprintf(outf,"_a%d%s",m,(m < (lM0-1)?",":";\n"));
  idnt();
  fprintf(outf,"%s *_cp;\n",fp_dtype);


  if (opA == Normal) {
    idnt(); fprintf(outf,"ap_0 = a;\n");
    for (m=1;m<lM0;m++) {
      idnt(); fprintf(outf,"ap_%d = apc_%d;\n",m,m);
    }
  } else {
    idnt();
    fprintf(outf,"ap=a;\n");
  }

  if (opB == Transpose) {
    idnt(); fprintf(outf,"bp_0 = b;\n");
    for (n=1;n<lN0;n++) {
      idnt(); fprintf(outf,"bp_%d = bp_%d + Bstride;\n",n,n-1);
    }
  } else {
    idnt();
    fprintf(outf,"bp=b;\n");
  }


  /* load or zero the accumulator */
  if (alpha == One || alpha == nOne) { /* load */

    idnt(); fprintf(outf,"_cp=cp;");
    for (m=0;m<lM0;m++) {
      for (n=0;n<lN0;n++)
	fprintf(outf,"c%d_%d=_cp[%d];",m,n,n);
      fprintf(outf,"\n");
      if (m < (lM0-1)) {
	idnt(); fprintf(outf,"_cp+=Cstride;");
      }
    }
  } else { /* zero */
    for (m=0;m<lM0;m++) {
      idnt();
      for (n=0;n<lN0;n++)
	fprintf(outf,"c%d_%d = 0.0; ",m,n);
      fprintf(outf,"\n");
    }
  }  


}

  
/*
** Generate an unrolled lM0xlK0xlN0 matrix multiply.
*/
void
generate_unrolled_mat_mul(size_t lM0, size_t lK0, size_t lN0) 
{
  int m,k,n;
  char *ac_str; /* accumulation operation string */

  if (alpha == nOne)
    ac_str = "-=";
  else 
    ac_str = "+=";

  idnt();
  fprintf(outf,
	  "/* Fixed M,K,N = %d,%d,%d fully-unrolled matrix matrix multiply. */\n",
	  lM0,lK0,lN0);

  if (opA == opB) {
    if (opA == Normal) { /* opB == Normal */
      if (rbStripe == Bstripe || (rbStripe == minStripe && (lN0<=lM0))) {
	for (k=0;k<lK0;k++) {
	  idnt(); fprintf(outf,"\n");
	  idnt();
	  for (n=0;n<lN0;n++)
	    fprintf(outf,"_b%d = bp[%d]; ",n,n);
	  fprintf(outf,"\n");
	  idnt(); fprintf(outf,"bp += Bstride;\n");
	  for(m=0;m<lM0;m++) {
	    idnt(); fprintf(outf,"_a%d = ap_%d[%d];\n",m,m,k);
	    idnt();
	    for (n=0;n<lN0;n++)
	      fprintf(outf,"c%d_%d %s _a%d*_b%d; ",m,n,ac_str,m,n);
	    fprintf(outf,"\n");
	  }
	}
      } else {
	for (k=0;k<lK0;k++) {
	  idnt(); fprintf(outf,"\n");
	  idnt();
	  for (m=0;m<lM0;m++)
	    fprintf(outf,"_a%d = ap_%d[%d]; ",m,m,k);
	  fprintf(outf,"\n");
	  for(n=0;n<lN0;n++) {
	    idnt(); fprintf(outf,"_b%d = bp[%d];\n",n,n);
	    idnt();
	    for (m=0;m<lM0;m++)
	      fprintf(outf,"c%d_%d %s _a%d*_b%d; ",m,n,ac_str,m,n);
	    fprintf(outf,"\n");
	  }
	  idnt(); fprintf(outf,"bp += Bstride;\n");
	}
      }
    } else { /* opA == Transpose && opB == Transpose */
      if (rbStripe == Bstripe || (rbStripe == minStripe && (lN0<=lM0))) {
	for (k=0;k<lK0;k++) {
	  idnt(); fprintf(outf,"\n");    
	  idnt();
	  for (n=0;n<lN0;n++)
	    fprintf(outf,"_b%d = bp_%d[%d]; ",n,n,k);
	  fprintf(outf,"\n");
	  for(m=0;m<lM0;m++) {
	    idnt(); fprintf(outf,"_a%d = ap[%d];\n",m,m);
	    idnt();
	    for (n=0;n<lN0;n++)
	      fprintf(outf,"c%d_%d %s _b%d*_a%d; ",m,n,ac_str,n,m);
	    fprintf(outf,"\n");
	  }
	  idnt(); fprintf(outf,"ap += Astride;\n");
	}
      } else {
	for (k=0;k<lK0;k++) {
	  idnt(); fprintf(outf,"\n");    
	  idnt();
	  for (m=0;m<lM0;m++)
	    fprintf(outf,"_a%d = ap[%d]; ",m,m);
	  fprintf(outf,"\n");
	  idnt(); fprintf(outf,"ap += Astride;\n");
	  for(n=0;n<lN0;n++) {
	    idnt(); fprintf(outf,"_b%d = bp_%d[%d];\n",n,n,k);
	    idnt();
	    for (m=0;m<lM0;m++)
	      fprintf(outf,"c%d_%d %s _b%d*_a%d; ",m,n,ac_str,n,m);
	    fprintf(outf,"\n");
	  }
	}
      }
    }
  } else if (opA == Normal) { /* opB == Transpose */
    if (rbStripe == Bstripe || (rbStripe == minStripe && (lN0<=lM0))) {
      for (k=0;k<lK0;k++) {
	idnt(); fprintf(outf,"\n");    
	idnt();
	for (n=0;n<lN0;n++)
	  fprintf(outf,"_b%d = bp_%d[%d]; ",n,n,k);
	fprintf(outf,"\n");
	for(m=0;m<lM0;m++) {
	  idnt(); fprintf(outf,"_a%d = ap_%d[%d];\n",m,m,k);
	  idnt();
	  for (n=0;n<lN0;n++)
	    fprintf(outf,"c%d_%d %s _a%d*_b%d; ",m,n,ac_str,m,n);
	  fprintf(outf,"\n");
	}
      }
    } else {
      for (k=0;k<lK0;k++) {
	idnt(); fprintf(outf,"\n");    
	idnt();
	for (m=0;m<lM0;m++)
	  fprintf(outf,"_a%d = ap_%d[%d]; ",m,m,k);
	fprintf(outf,"\n");
	for(n=0;n<lN0;n++) {
	  idnt(); fprintf(outf,"_b%d = bp_%d[%d];\n",n,n,k);
	  idnt();
	  for (m=0;m<lM0;m++)
	    fprintf(outf,"c%d_%d %s _a%d*_b%d; ",m,n,ac_str,m,n);
	  fprintf(outf,"\n");
	}
      }
    }
  } else { /* opA == Transpose && opB == Normal */
    if (rbStripe == Bstripe || (rbStripe == minStripe && (lN0<=lM0))) {
      for (k=0;k<lK0;k++) {
	idnt(); fprintf(outf,"\n");    
	idnt();
	for (n=0;n<lN0;n++)
	  fprintf(outf,"_b%d = bp[%d]; ",n,n);
	fprintf(outf,"\n");
	idnt(); fprintf(outf,"bp += Bstride;\n");
	for(m=0;m<lM0;m++) {
	  idnt(); fprintf(outf,"_a%d = ap[%d];\n",m,m);
	  idnt();
	  for (n=0;n<lN0;n++)
	    fprintf(outf,"c%d_%d %s _a%d*_b%d; ",m,n,ac_str,m,n);
	  fprintf(outf,"\n");
	}
	idnt(); fprintf(outf,"ap += Astride;\n");
      }
    } else {
      for (k=0;k<lK0;k++) {
	idnt(); fprintf(outf,"\n");    
	idnt();
	for (m=0;m<lM0;m++)
	  fprintf(outf,"_a%d = ap[%d]; ",m,m);
	fprintf(outf,"\n");
	idnt(); fprintf(outf,"ap += Astride;\n");
	for(n=0;n<lN0;n++) {
	  idnt(); fprintf(outf,"_b%d = bp[%d];\n",n,n);
	  idnt();
	  for (m=0;m<lM0;m++)
	    fprintf(outf,"c%d_%d %s _a%d*_b%d; ",m,n,ac_str,m,n);
	  fprintf(outf,"\n");
	}
	idnt(); fprintf(outf,"bp += Bstride;\n");
      }
    }
  }
  fprintf(outf,"\n");

}
 



void
gen_core_k_loop_1stg(size_t lM0,size_t lK0, size_t lN0)
{
  size_t m,n,k;
  
  /* generate the loop start control */
  if (!no_loop_head_branch) {
    idnt();
    fprintf(outf,"for (;");
    fprintf(outf,"ap%s!=ap_endp; ",(opA == Normal ? "_0" : ""));
    if (opA == Normal) {
      for (m=0;m<lM0;m++)
	fprintf(outf,"ap_%d+=%d%s",m,lK0,(m<(lM0-1))?",":"");
    }
    if (opB == Transpose) {
      if (opA == Normal) 
	fprintf(outf,",");
      for (n=0;n<lN0;n++)
	fprintf(outf,"bp_%d+=%d%s",n,lK0,
		(n<(lN0-1))?",":"");
    }
    fprintf(outf,") {\n");
    nst();
  } else {
    /* We assume that this is not fringe code. */
    assert (lM0 == M0 && lK0 == K0 && lN0 == N0);
    idnt();
    if (L1SET) {
      if (K1 > 1) {
	fprintf(outf,"do {\n");
      } else {
	fprintf(outf,"{\n");
      }
    } else {
      fprintf(outf,"do {\n");
    }
    nst();
  }

  /* generate the core unrolled matrix multiply */
  generate_unrolled_mat_mul(lM0,lK0,lN0);

  /* generate the loop end control */
  if (!no_loop_head_branch) {
    unst(); idnt(); fprintf(outf,"}\n");
  } else {
    if ((L1SET && K1 > 1) || !L1SET) {
      if (opA == Normal) {
	idnt();
	for (m=0;m<lM0;m++)
	  fprintf(outf,"ap_%d+=%d;",m,lK0);
	fprintf(outf,"\n");
      }
      if (opB == Transpose) {
	idnt();
	for (n=0;n<lN0;n++)
	  fprintf(outf,"bp_%d+=%d;",n,lK0);
	fprintf(outf,"\n");
      }
      unst();
      idnt();
      fprintf(outf,"} while (ap%s!=ap_endp);\n",
	      (opA == Normal ? "_0" : ""));
    } else { /* L1SET && K1 == 1 */
      /* no need to update pointers here since
	 there this must not be fringe code (see assertion
	 above) */
      unst();
      idnt(); 
      fprintf(outf,"};\n");
    }
  }


  /* K fringe */
  if (do_k_fringe) {
    if (!linear_k_fringe) {
      if (pow_of_2(lK0))
	k = (lK0 >> 1);
      else 
	k = highbit(lK0);
      for (;k>=1;k>>=1) {
	idnt(); fprintf(outf,"if (k_marg_el & 0x%X) {\n",k);
	nst();
	generate_unrolled_mat_mul(lM0,k,lN0);
	/*  pointer updates */
	if (k > 1) {
	  if (opA == Normal) {
	    idnt();
	    for (m=0;m<lM0;m++)
	      fprintf(outf,"ap_%d+=%d;",m,k);
	    fprintf(outf,"\n");
	  }
	  if (opB == Transpose) {
	    idnt();
	    for (n=0;n<lN0;n++)
	      fprintf(outf,"bp_%d+=%d;",n,k);
	    fprintf(outf,"\n");
	  }
	}
	unst();
	idnt(); fprintf(outf,"}\n");
      }
    } else {
      for (k=(lK0-1);k>=1;k--) {
	idnt(); 
	if (k == (lK0-1) || (lK0 == 2))
	  fprintf(outf,"if (k_marg_el == %d) {\n",k);
	else 
	  fprintf(outf,"else if (k_marg_el == %d) {\n",k);
	nst();
	generate_unrolled_mat_mul(lM0,k,lN0);
	unst();
	idnt(); fprintf(outf,"}\n");
      }
    }
  }
}

/*
 *  Support routines for software pipelining.
 */

void
gen_core_k_loop_sp_load(size_t lM0, size_t lN0, size_t k)
{
  size_t m,n;

  /* load a and b */
  idnt();
  if (opB == Normal) {
    for (n=0;n<lN0;n++)
      fprintf(outf,"_b%d = bp[%d];",n,n);
  } else {
    for (n=0;n<lN0;n++)
      fprintf(outf,"_b%d = bp_%d[%d];",n,n,k);
  }
  if (opA == Normal) {
    for(m=0;m<lM0;m++)
      fprintf(outf,"_a%d = ap_%d[%d];",m,m,k);
  } else {
    for(m=0;m<lM0;m++)
      fprintf(outf,"_a%d = ap[%d];",m,m);
  }
  fprintf(outf,"\n");
}


void
gen_core_k_loop_sp_mul(size_t lM0, size_t lN0)
{
  size_t m,n;

  /* multiply into temporaries */
  for(m=0;m<lM0;m++) {
    idnt();
    for (n=0;n<lN0;n++)
      fprintf(outf,"t%d_%d=_a%d*_b%d;",m,n,m,n);
    fprintf(outf,"\n");
  }
}

void
gen_core_k_loop_sp_add(size_t lM0, size_t lN0)
{
  size_t m,n;
  char *ac_str; /* accumulation operation string */

  if (alpha == nOne)
    ac_str = "-=";
  else 
    ac_str = "+=";


  /* accumulate */
  for (m=0;m<lM0;m++) {
    idnt();
    for (n=0;n<lN0;n++)
      fprintf(outf,"c%d_%d%st%d_%d;",m,n,ac_str,m,n);
    fprintf(outf,"\n");
  }
}


void
gen_core_k_loop_sp_update_a(size_t lM0,size_t val)
{
  size_t m;

  if (opA == Normal) {
    idnt();
    for (m=0;m<lM0;m++)
      fprintf(outf,"ap_%d+=%d;",m,val);
    fprintf(outf,"\n");  
  } else {
    if (val == 1) {
      idnt(); fprintf(outf,"ap += Astride;\n");
    } else {
      idnt(); fprintf(outf,"ap += (%d*Astride);\n",val);
    }
  }
}

void
gen_core_k_loop_sp_update_b(size_t lN0,size_t val)
{
  size_t n;

  /*  pointer updates */
  if (opB == Normal) {
    if (val == 1) {
      idnt(); fprintf(outf,"bp+=Bstride;\n");
    } else {
      idnt(); fprintf(outf,"bp+=(%d*Bstride);\n",val);
    }
  } else {
    idnt();
    for (n=0;n<lN0;n++)
      fprintf(outf,"bp_%d+=%d;",n,val);
    fprintf(outf,"\n");  
  }
}




/*
 * generate two pipe stage, [ld-mul, add]
 */
void
gen_core_k_loop_2stglm(size_t lM0,size_t lK0, size_t lN0) 
{
  int k;

  /* start of software pipelining */
  if (!no_loop_head_branch) {
    idnt();
    fprintf(outf,"if (K >= %d) {\n",lK0+1);
    nst();
  }

  /* PROLOGUE */
  /* load a and b */
  gen_core_k_loop_sp_load(lM0,lN0,0);
  gen_core_k_loop_sp_update_b(lN0,1);

  /* multiply into temporaries */
  gen_core_k_loop_sp_mul(lM0,lN0);
  gen_core_k_loop_sp_update_a(lM0,1);


  /* loop control */
  if (!L1SET || (L1SET && (K1>1))) {
    idnt(); fprintf(outf,"do {\n");
    nst();
  }

  for (k=0;k<lK0;k++) {

    /* accumulate */
    gen_core_k_loop_sp_add(lM0,lN0);

    /* load */
    gen_core_k_loop_sp_load(lM0,lN0,k);

    /* multiply */
    gen_core_k_loop_sp_mul(lM0,lN0);

    /*  pointer updates */
    if (opA == Transpose)
      gen_core_k_loop_sp_update_a(lM0,1);
    if (opB == Normal)
      gen_core_k_loop_sp_update_b(lN0,1);


  }

  /*  pointer updates */
  if (opA == Normal)
    gen_core_k_loop_sp_update_a(lM0,lK0);
  if (opB == Transpose)
    gen_core_k_loop_sp_update_b(lN0,lK0);


  /* loop control */
  if (!L1SET || (L1SET && (K1>1))) {
    unst();
    idnt(); fprintf(outf,"} while (ap%s != ap_endp);\n",
		    (opA == Normal ? "_0" : ""));
  }

  /* epilogue */

  /* accumulate */
  gen_core_k_loop_sp_add(lM0,lN0);

  /* end of software pipelining */
  if (!no_loop_head_branch) {
    unst();
    idnt(); fprintf(outf,"}\n");
  }

  /* now we do rest of K and perhaps fringe */
  if (!L1SET && do_k_fringe) {
    if (!linear_k_fringe) {
      for (k = highbit(lK0);k>=1;k>>=1) {
	idnt(); fprintf(outf,"if (k_marg_el & 0x%X) {\n",k);
	nst();
	generate_unrolled_mat_mul(lM0,k,lN0);
	/*  pointer updates */
	if (k > 1) {
	  if (opA == Normal)
	    gen_core_k_loop_sp_update_a(lM0,k);
	  if (opB == Transpose)
	    gen_core_k_loop_sp_update_b(lN0,k);
	}
	unst();
	idnt(); fprintf(outf,"}\n");
      }
    } else {
      for (k=lK0;k>=1;k--) {
	idnt(); 
	if (k == lK0 || (lK0 == 1))
	  fprintf(outf,"if (k_marg_el == %d) {\n",k);
	else 
	  fprintf(outf,"else if (k_marg_el == %d) {\n",k);
	nst();
	generate_unrolled_mat_mul(lM0,k,lN0);
	unst();
	idnt(); fprintf(outf,"}\n");
      }
    }
  }
}

/*
 * generate two pipe stage, [ld, mul-add]
 */
void
gen_core_k_loop_2stgma(size_t lM0,size_t lK0, size_t lN0) 
{
  int k;

  /* start of software pipelining */
  if (!no_loop_head_branch) {
    idnt();
    fprintf(outf,"if (K >= %d) {\n",lK0+1);
    nst();
  }

  /* PROLOGUE */
  /* load a and b */
  gen_core_k_loop_sp_load(lM0,lN0,0);
  gen_core_k_loop_sp_update_b(lN0,1);
  gen_core_k_loop_sp_update_a(lM0,1);

  /* loop control */
  if (!L1SET || (L1SET && (K1>1))) {
    idnt(); fprintf(outf,"do {\n");
    nst();
  }

  for (k=0;k<lK0;k++) {
    /* multiply */
    gen_core_k_loop_sp_mul(lM0,lN0);

    /* accumulate */
    gen_core_k_loop_sp_add(lM0,lN0);

    /* load */
    gen_core_k_loop_sp_load(lM0,lN0,k);

    /*  pointer updates */
    if (opA == Transpose)
      gen_core_k_loop_sp_update_a(lM0,1);
    if (opB == Normal)
      gen_core_k_loop_sp_update_b(lN0,1);
  }

  /*  pointer updates */
  if (opA == Normal)
    gen_core_k_loop_sp_update_a(lM0,lK0);
  if (opB == Transpose)
    gen_core_k_loop_sp_update_b(lN0,lK0);


  /* loop control */
  if (!L1SET || (L1SET && (K1>1))) {
    unst();
    idnt(); fprintf(outf,"} while (ap%s != ap_endp);\n",
		    (opA == Normal ? "_0" : ""));
  }

  /* epilogue */

  /* multiply */
  gen_core_k_loop_sp_mul(lM0,lN0);


  /* accumulate */
  gen_core_k_loop_sp_add(lM0,lN0);


  /* end of software pipelining */
  if (!no_loop_head_branch) {
    unst();
    idnt(); fprintf(outf,"}\n");
  }

  /* now we do rest of K and perhaps fringe */
  if (!L1SET && do_k_fringe) {
    if (!linear_k_fringe) {
      for (k = highbit(lK0);k>=1;k>>=1) {
	idnt(); fprintf(outf,"if (k_marg_el & 0x%X) {\n",k);
	nst();
	generate_unrolled_mat_mul(lM0,k,lN0);
	/* pointer updates */
	if (k > 1) {
	  if (opA == Normal)
	    gen_core_k_loop_sp_update_a(lM0,k);
	  if (opB == Transpose)
	    gen_core_k_loop_sp_update_b(lN0,k);
	}
	unst();
	idnt(); fprintf(outf,"}\n");
      }
    } else {
      for (k=lK0;k>=1;k--) {
	idnt(); 
	if (k == lK0 || (lK0 == 1))
	  fprintf(outf,"if (k_marg_el == %d) {\n",k);
	else 
	  fprintf(outf,"else if (k_marg_el == %d) {\n",k);
	nst();
	generate_unrolled_mat_mul(lM0,k,lN0);
	unst();
	idnt(); fprintf(outf,"}\n");
      }
    }
  }
}


void
gen_core_k_loop_3stg(size_t lM0,size_t lK0, size_t lN0) 
{
  int k;

  /* start of software pipelining */
  if (!no_loop_head_branch) {
    idnt();
    fprintf(outf,"if (K >= %d) {\n",lK0+2);
    nst();
  }

  /* PROLOGUE */
  /* load a and b */
  gen_core_k_loop_sp_load(lM0,lN0,0);
  if (opB == Normal)
    gen_core_k_loop_sp_update_b(lN0,1);
  if (opA == Transpose)
    gen_core_k_loop_sp_update_a(lM0,1);

  /* multiply into temporaries */
  gen_core_k_loop_sp_mul(lM0,lN0);

  /* load next stage a and b */
  gen_core_k_loop_sp_load(lM0,lN0,1);
  if (opB == Normal)
    gen_core_k_loop_sp_update_b(lN0,1);
  else 
    gen_core_k_loop_sp_update_b(lN0,2);
  if (opA == Transpose)
    gen_core_k_loop_sp_update_a(lM0,1);
  else
    gen_core_k_loop_sp_update_a(lM0,2);

  /* loop control */
  if (!L1SET || (L1SET && (K1>1))) {
    idnt(); fprintf(outf,"do {\n");
    nst();
  }

  for (k=0;k<lK0;k++) {

    /* accumulate */
    gen_core_k_loop_sp_add(lM0,lN0);

    /* multiply */
    gen_core_k_loop_sp_mul(lM0,lN0);

    /* load */
    gen_core_k_loop_sp_load(lM0,lN0,k);

    /* pointer updates */
    if (opA == Transpose)
      gen_core_k_loop_sp_update_a(lM0,1);
    if (opB == Normal)
      gen_core_k_loop_sp_update_b(lN0,1);
  }

  /* pointer updates */
  if (opA == Normal)
    gen_core_k_loop_sp_update_a(lM0,lK0);
  if (opB == Transpose)
    gen_core_k_loop_sp_update_b(lN0,lK0);


  /* loop control */
  if (!L1SET || (L1SET && (K1>1))) {
    unst();
    idnt(); fprintf(outf,"} while (ap%s != ap_endp);\n",
		    (opA == Normal ? "_0" : ""));
  }

  /* epilogue */

  /* accumulate */
  gen_core_k_loop_sp_add(lM0,lN0);

  /* multiply */
  gen_core_k_loop_sp_mul(lM0,lN0);

  /* accumulate */
  gen_core_k_loop_sp_add(lM0,lN0);


  /* end of software pipelining */
  if (!no_loop_head_branch) {
    unst();
    idnt(); fprintf(outf,"}\n");
  }

  /* now we do rest of K and perhaps fringe */
  if (!L1SET && do_k_fringe) {
    if (!linear_k_fringe) {
      for (k = highbit(lK0+1);k>=1;k>>=1) {
	idnt(); fprintf(outf,"if (k_marg_el & 0x%X) {\n",k);
	nst();
	generate_unrolled_mat_mul(lM0,k,lN0);
	/* pointer updates */
	if (k > 1) {
	  if (opA == Normal)
	    gen_core_k_loop_sp_update_a(lM0,k);
	  if (opB == Transpose)
	    gen_core_k_loop_sp_update_b(lN0,k);
	}
	unst();
	idnt(); fprintf(outf,"}\n");
      }
    } else {
      for (k=(lK0+1);k>=1;k--) {
	idnt(); 
	if (k == (lK0+1))
	  fprintf(outf,"if (k_marg_el == %d) {\n",k);
	else 
	  fprintf(outf,"else if (k_marg_el == %d) {\n",k);
	nst();
	generate_unrolled_mat_mul(lM0,k,lN0);
	unst();
	idnt(); fprintf(outf,"}\n");
      }
    }
  }
}


void
generate_core_k_loop_body(size_t lM0,size_t lK0, size_t lN0)
{
  if (softpipe == OneStage) {
    gen_core_k_loop_1stg(lM0,lK0,lN0);
  } else if (softpipe == TwoStageLoadMul) {
    gen_core_k_loop_2stglm(lM0,lK0,lN0);
  } else if (softpipe == TwoStageMulAdd) {
    gen_core_k_loop_2stgma(lM0,lK0,lN0);
  } else if (softpipe == ThreeStage) {
    gen_core_k_loop_3stg(lM0,lK0,lN0);
  }
}

void
generate_core_k_loop_close(size_t lM0,size_t lK0, size_t lN0)
{
  size_t m,n;

  /* store or scale/accumulate/store the accumulator */
  if (alpha == One || alpha == nOne) { /* store */
    
    idnt(); fprintf(outf,"_cp=cp;");
    for (m=0;m<lM0;m++) {
      for (n=0;n<lN0;n++)
	fprintf(outf,"_cp[%d]=c%d_%d;",n,m,n);
      fprintf(outf,"\n");
      if (m < (lM0-1)) {
	idnt(); fprintf(outf,"_cp+=Cstride;");
      }
    }
  } else { /* scale/accumulator/store */

    idnt(); fprintf(outf,"_cp=cp;");
    for (m=0;m<lM0;m++) {
      for (n=0;n<lN0;n++) {
	if (alpha == Arb)
	  fprintf(outf,"_cp[%d]+=alpha*c%d_%d;",n,m,n);
	else 
	  fprintf(outf,"_cp[%d]+=%s*c%d_%d;",n,alpha_val,m,n);
      }
      fprintf(outf,"\n");
      if (m < (lM0-1)) {
	idnt(); fprintf(outf,"_cp+=Cstride;");
      }
    }
    
  }  

}


void
generate_core_k_loop(size_t lM0,size_t lK0, size_t lN0)
{
  generate_core_k_loop_open(lM0,lK0,lN0);
  generate_core_k_loop_body(lM0,lK0,lN0);
  generate_core_k_loop_close(lM0,lK0,lN0);
}

/***************/
/* N-LOOP Code */
/***************/


void
generate_core_n_loop_open(size_t lM0, size_t lK0, size_t lN0)
{
  size_t m;
  /* generate the N loop */

  idnt(); fprintf(outf,"%s* const cp_endp = ",fp_dtype);
  if (!L1SET)
    fprintf(outf,"c + n_norm");
  else 
    fprintf(outf,"c + %d",Nc);
  fprintf(outf,";\n");

  if (opA == Normal) {
    if (lM0 > 1) {
      idnt(); fprintf(outf,"const %s* const apc_1 = a + Astride;\n",
		      fp_type);
      for (m=2;m<lM0;m++) {
	idnt(); fprintf(outf,"const %s* const apc_%d = apc_%d + Astride;\n",
			fp_type,m,m-1);
      }
    }
  }


  if (!no_loop_head_branch) {
    idnt();
    fprintf(outf,"for (b=B,cp=c; cp!=cp_endp; ");
    fprintf(outf,"b+=");
    if (opB == Normal) {
      fprintf(outf,"%d,",lN0);
    } else {
      fprintf(outf,"B_sbs_stride,");
    }
    fprintf(outf,"cp+=%d) {\n",lN0);
    nst();
  } else {
    /* We assume that this is not fringe code.  */
    assert (lM0 == M0 && lK0 == K0 && lN0 == N0);
    idnt(); fprintf(outf,"b=B;cp=c;\n");
    idnt(); 
    if ((L1SET && N1 > 1) || !L1SET)
      fprintf(outf,"do {\n");
    else 
      fprintf(outf,"{\n");
    nst();
  }
}


/*
 * The opening for the n fringe 
 */
void
generate_core_n_fringe_open(size_t lM0, size_t lK0, size_t lN0)
{
  size_t m;

  if (opA == Normal) {
    if (lM0 > 1) {
      idnt(); fprintf(outf,"const %s* const apc_1 = a + Astride;\n",
		      fp_type);
      for (m=2;m<lM0;m++) {
	idnt(); fprintf(outf,"const %s* const apc_%d = apc_%d + Astride;\n",
			fp_type,m,m-1);
      }
    }
  }
      
  if (opB == Normal)  {
    idnt(); fprintf(outf,"b = B+n_norm;\n");
  } else {
    idnt(); fprintf(outf,"b = B+n_norm*Bstride;\n");
  }
  idnt(); fprintf(outf,"cp = c+n_norm;\n");

}


void
generate_core_n_loop_close(size_t lN0)
{
  if (!no_loop_head_branch) {
    unst();
    idnt(); fprintf(outf,"}\n");
  } else {
    /* We assume that this is not fringe code.  */
    assert (lN0 == N0);
    if ((L1SET && N1 > 1) || !L1SET) {
      idnt();
      fprintf(outf,"b+=");
      if (opB == Normal) {
	fprintf(outf,"%d;",lN0);
      } else {
	fprintf(outf,"B_sbs_stride;");
      }
      fprintf(outf,"cp+=%d;\n",lN0);
      unst();
      idnt();
      fprintf(outf,"} while (cp!=cp_endp);\n");
    } else {
      unst(); idnt();
      fprintf(outf,"}\n");
    }
  }
}

void
generate_core_n_fringe_close()
{
  /* do nothing */
}

void
generate_core_n_fringe_body(size_t lM0, size_t lK0, size_t lN0)
{
  size_t n;

  if (!linear_n_fringe) {
    n = (pow_of_2(lN0) ? (lN0>>1) : highbit(lN0));
    for (;n>=1;n>>=1) {
      idnt(); fprintf(outf,"if (n_marg_el & 0x%X) {\n",n);
      nst();
      generate_core_k_loop(lM0,lK0,n);
      if (n > 1) {
	if (opB == Normal) {
	  idnt(); fprintf(outf,"b+=%d;\n",n);
	} else {
	  idnt(); fprintf(outf,"b+=Bstride*%d;\n",n);
	}
	idnt(); fprintf(outf,"cp+=%d;\n",n);
      }
      unst();
      idnt(); fprintf(outf,"}\n");
    }
  } else {
    for (n=(lN0-1);n>=1;n--) {
      idnt(); 
      if (n == (lN0-1) || (lN0 == 2))
	fprintf(outf,"if (n_marg_el == %d) {\n",n);
      else 
	fprintf(outf,"else if (n_marg_el == %d) {\n",n);
      nst();
      generate_core_k_loop(lM0,lK0,n);
      unst();
      idnt(); fprintf(outf,"}\n");
    }
  }
}


/***************/
/* M-LOOP Code */
/***************/

void
generate_core_m_loop_open()
{
  /* generate M loop */
  if (!no_loop_head_branch) {
    idnt();
    fprintf(outf,"for (c=C,a=A; c!= c_endp; c+=C_sbs_stride,");
    if (opA == Normal) {
      fprintf(outf,"a+=A_sbs_stride");
    } else {
      fprintf(outf,"a+=%d",M0);
    }
    fprintf(outf,") {\n");
    nst();
  } else {
    idnt(); fprintf(outf,"c=C;a=A;\n");
    idnt(); 
    if ((L1SET && M1 > 1) || !L1SET) 
      fprintf(outf,"do {\n");
    else 
      fprintf(outf,"{\n");
    nst();
  }

  idnt(); fprintf(outf,"const %s* const ap_endp = ",fp_type);
  if (!L1SET || opA == Transpose) 
    fprintf(outf,"a + k_norm");
  else {
    fprintf(outf,"a + %d",SP_ADJ_K(Kc));
  }
  fprintf(outf,";\n");

}


void
generate_core_m_fringe_body_using(size_t m, bool update_pointers)
{
  idnt(); fprintf(outf,"const %s* const ap_endp = ",fp_type);
  if (!L1SET || opA == Transpose) 
    fprintf(outf,"a + k_norm");
  else {
    fprintf(outf,"a + %d",SP_ADJ_K(Kc));      
  }
  fprintf(outf,";\n");

  generate_core_n_loop_open(m,K0,N0);
  generate_core_k_loop(m,K0,N0);
  generate_core_n_loop_close(N0);

  if (do_n_fringe) {
    generate_core_n_fringe_body(m,K0,N0);
    generate_core_n_fringe_close();
  }

  if (update_pointers) {
    idnt(); fprintf(outf,"c += Cstride*%d;\n",m);
    if (opA == Normal)  {
      idnt(); fprintf(outf,"a += Astride*%d;\n",m);
    } else {
      idnt(); fprintf(outf,"a += %d;\n",m);
    }
  }
}

void
generate_core_m_fringe_body()
{
  size_t m;
  
  if (!linear_m_fringe) {
    m = (pow_of_2(M0) ? (M0>>1) : highbit(M0));
    for (;m>=1;m>>=1) {
      idnt(); 
      fprintf(outf,"if (m_marg_el & 0x%X) {\n",m);
      nst();
      generate_core_m_fringe_body_using(m,(m>1));
      unst();
      idnt(); fprintf(outf,"}\n");
    }
  } else {
    for (m=(M0-1);m>=1;m--) {
      idnt(); 
      if (m == (M0-1) || (M0 == 2))
	fprintf(outf,"if (m_marg_el == %d) {\n",m);
      else 
	fprintf(outf,"else if (m_marg_el == %d) {\n",m);
      nst();
      generate_core_m_fringe_body_using(m,False);
      unst();
      idnt(); fprintf(outf,"}\n");
    }
  }
}



void
generate_core_m_loop_close(size_t lM0)
{
  if (!no_loop_head_branch) {
    unst();
    idnt(); fprintf(outf,"}\n");
  } else {
    assert (lM0 == M0);
    if ((L1SET && M1 > 1) || !L1SET) {
      idnt();
      fprintf(outf,"c+=C_sbs_stride;");
      if (opA == Normal) {
	fprintf(outf,"a+=A_sbs_stride;\n");
      } else {
	fprintf(outf,"a+=%d;\n",lM0);
      }
      unst(); idnt(); fprintf(outf,"} while (c!=c_endp);\n");
    } else {
      unst(); idnt(); fprintf(outf,"}\n");
    }
  }
}


/*******************/
/* PROLOGUE Code   */
/*******************/


void
generate_core_prologue()
{
  int m,n;

  idnt();
  fprintf(outf,"const %s *a,*b;\n",fp_type);
  idnt();
  fprintf(outf,"%s *c;\n",fp_dtype);

  idnt();
  fprintf(outf,"const %s ",fp_type);
  if (opA == Normal) {
    for (m=0;m<M0;m++) 
      fprintf(outf,"*ap_%d%s",m,((m==(M0-1))?";":","));
    fprintf(outf,"\n");
  } else {
    fprintf(outf,"*ap;\n");
  }
  idnt();
  fprintf(outf,"const %s ",fp_type);
  if (opB == Normal) {
    fprintf(outf,"*bp;\n");
  } else {
    for (n=0;n<N0;n++) 
      fprintf(outf,"*bp_%d%s",n,((n==(N0-1))?";":","));
    fprintf(outf,"\n");
  }

  idnt();
  fprintf(outf,"%s *cp;\n",fp_dtype);

  if (opA == Normal) {
    idnt(); fprintf(outf,"const int A_sbs_stride = Astride*%d;\n",M0);
  }
  if (opB == Transpose) {
    idnt(); fprintf(outf,"const int B_sbs_stride = Bstride*%d;\n",N0);
  }
  idnt(); fprintf(outf,"const int C_sbs_stride = Cstride*%d;\n",M0);

  if (!L1SET) {
    idnt(); fprintf(outf,"const int k_marg_el = ");
    if (softpipe == OneStage) {
      if (pow_of_2(K0))
	fprintf(outf,"K & %d;\n",K0 - 1);
      else 
	fprintf(outf,"K %% %d;\n",K0);
    } else {
      if (softpipe == ThreeStage) {
	fprintf(outf,"((K>=%d)?",K0+2);
	if (pow_of_2(K0))
	  fprintf(outf,"((K-2)&%d):K);\n",K0-1);
	else 
	  fprintf(outf,"((K-2)%%%d):K);\n",K0);
      } else {
	fprintf(outf,"((K>=%d)?",K0+1);
	if (pow_of_2(K0))
	  fprintf(outf,"((K-1)&%d):K);\n",K0-1);
	else 
	  fprintf(outf,"((K-1)%%%d):K);\n",K0);
      }
    }

    idnt(); 
    if (opA == Normal)
      fprintf(outf,"const int k_norm = K - k_marg_el;\n");
    else 
      fprintf(outf,"const int k_norm = (K - k_marg_el)*Astride;\n");
    idnt(); fprintf(outf,"const int m_marg_el = ");
    if (pow_of_2(M0))
      fprintf(outf,"M & %d;\n",M0 - 1);
    else 
      fprintf(outf,"M %% %d;\n",M0);
    idnt(); fprintf(outf,"const int m_norm = M - m_marg_el;\n");
    idnt(); fprintf(outf,"const int n_marg_el = ");
    if (pow_of_2(N0))
      fprintf(outf,"N & %d;\n",N0 - 1);
    else 
      fprintf(outf,"N %% %d;\n",N0);
    idnt(); fprintf(outf,"const int n_norm = N - n_marg_el;\n");
  } else {
    if (opA == Transpose) {
      idnt();
      fprintf(outf,"const int k_norm = %d*Astride;\n",SP_ADJ_K(Kc));
    }
  }

  if (!L1SET) {
    idnt();
    fprintf(outf,"%s *const c_endp = C+m_norm*Cstride;\n",fp_dtype);
  } else {
    idnt();
    fprintf(outf,"%s *const c_endp = C+%d*Cstride;\n",fp_dtype,Mc);
  }


  idnt(); fprintf(outf,"register %s ",fp_atype);
  for (m=0;m<M0;m++)
    for (n=0;n<N0;n++)
      fprintf(outf,"c%d_%d%s",m,n,(m==(M0-1)&&(n==(N0-1))?"":","));
  fprintf(outf,";\n");

  if (softpipe != OneStage) {
    /* Add local variables for temporaries. */
    idnt(); fprintf(outf,"register %s ",fp_atype);
    for (m=0;m<M0;m++)
      for (n=0;n<N0;n++)
	fprintf(outf,"t%d_%d%s",m,n,(m==(M0-1)&&(n==(N0-1))?"":","));
    fprintf(outf,";\n");
  }

}

/***************/
/* BETA Code   */
/***************/


void
generate_core_beta_code()
{
  if (beta == One)
    return;  /* no scaling necessary */

  if (beta == Arb || beta == Fix) {
    if (beta == Fix) {
      idnt(); fprintf(outf,"{\n");
      nst(); /* 1 */
      idnt(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      idnt(); fprintf(outf,"cpre = C + M*Cstride;\n");
      idnt();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nst(); /* 2 */
      idnt(); fprintf(outf,"cpe = cprb + N;\n");
      idnt(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nst(); /* 3 */
      idnt(); fprintf(outf,"*cp *= (%s);\n",(beta==Arb?"beta":beta_val));
      unst(); /* 3 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 2 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 1 */
      idnt(); fprintf(outf,"}\n");
    } else { /* beta == Arb */
      idnt(); fprintf(outf,"if (beta == 0.0) {\n");
      nst(); /* 1 */
      idnt(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      idnt(); fprintf(outf,"cpre = C + M*Cstride;\n");
      idnt();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nst(); /* 2 */
      idnt(); fprintf(outf,"cpe = cprb + N;\n");
      idnt(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nst(); /* 3 */
      idnt(); fprintf(outf,"*cp = 0.0;\n");
      unst(); /* 3 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 2 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 1 */
      idnt(); fprintf(outf,"} else if (beta != 1.0) {\n");
      nst(); /* 1 */
      idnt(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      idnt(); fprintf(outf,"cpre = C + M*Cstride;\n");
      idnt();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nst(); /* 2 */
      idnt(); fprintf(outf,"cpe = cprb + N;\n");
      idnt(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nst(); /* 3 */
      idnt(); fprintf(outf,"*cp *= (beta);\n");
      unst(); /* 3 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 2 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 1 */
      idnt(); fprintf(outf,"}\n");
    }
  } else if (beta == Zero) {
      idnt(); fprintf(outf,"{\n");
      nst(); /* 1 */
      idnt(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      idnt(); fprintf(outf,"cpre = C + M*Cstride;\n");
      idnt();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nst(); /* 2 */
      idnt(); fprintf(outf,"cpe = cprb + N;\n");
      idnt(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nst(); /* 3 */
      idnt(); fprintf(outf,"*cp = 0.0;\n");
      unst(); /* 3 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 2 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 1 */
      idnt(); fprintf(outf,"}\n");
  }
}

/***************/
/* CORE Code   */
/***************/

void
generate_core_n_fringe_code()
{
  if (!L1SET && do_n_fringe && N0 > 1) {
    generate_core_m_loop_open();
    generate_core_n_fringe_open(M0,K0,N0);
    generate_core_n_fringe_body(M0,K0,N0);
    generate_core_n_fringe_close();
    generate_core_m_loop_close(M0);
  }
}

void
generate_core_m_fringe_code()
{
  if (!L1SET && do_m_fringe && M0 > 1)
    generate_core_m_fringe_body();
}


void
generate_core_code() 
{
  generate_core_prologue();
  generate_core_beta_code();

  /* core code, possibly including k frindge */
  generate_core_m_loop_open();
  generate_core_n_loop_open(M0,K0,N0);
  generate_core_k_loop(M0,K0,N0);
  generate_core_n_loop_close(N0);
  generate_core_m_loop_close(M0);

  /* N fringes */
  generate_core_n_fringe_code();

  /* M fringes */
  generate_core_m_fringe_code();

}



void
generate_core_routine_comment()
{
  if (!L1SET) {
    fprintf(outf,"General (M,K,N) = (");
    if (do_m_fringe)
      fprintf(outf,"M, ");
    else 
      fprintf(outf,"m*%d:m>=%s, ",M0,(no_loop_head_branch?"1":"0"));
    if (do_k_fringe)
      fprintf(outf,"K, ");
    else {
      fprintf(outf,"k*%d",K0);
      if (softpipe == OneStage)
	fprintf(outf,":k>=%s, ",(no_loop_head_branch?"1":"0"));
      else if (softpipe == TwoStageLoadMul || softpipe == TwoStageMulAdd) {
	/* This isn't a completely correct comment since calling
           the resulting routine with K==0 would still work. */
	fprintf(outf,"+1:k>=1, ");
      } else if (softpipe == ThreeStage) {
	/* This isn't a completely correct comment since calling
           the resulting routine with K==0 would still work. */
	fprintf(outf,"+2:k>=1, ");
      }
    }
    if (do_n_fringe)
      fprintf(outf,"N");
    else 
      fprintf(outf,"n*%d:n>=%s",N0,(no_loop_head_branch?"1":"0"));
    fprintf(outf,") matrix multiply");
  } else {
    fprintf(outf,"Fixed (M,K,N) = (%d,%d,%d) matrix multiply",
	    Mc,SP_ADJ_K(Kc),Nc);
  }

  if (softpipe == OneStage) {
  } else if (softpipe == TwoStageLoadMul) {
    fprintf(outf,": Two stage software pipe [load-mul, add]");
  } else if (softpipe == TwoStageMulAdd) {
    fprintf(outf,": Two stage software pipe [load, mul-add]");
  } else if (softpipe == ThreeStage) {
    fprintf(outf,": Three stage software pipe [load, mul, add]");
  }

}

void
generate_core_routine_header()
{

  if (!L1SET || dummy_size_args) {
    fprintf(outf,"%s(const int M, const int K, const int N, const %s *const A, const %s *const B, %s *const C, const int Astride, const int Bstride, const int Cstride",routine_name,fp_type,fp_type,fp_dtype);
  } else {
    fprintf(outf,"%s(const %s *const A, const %s *const B, %s *const C, const int Astride, const int Bstride, const int Cstride",routine_name,fp_type,fp_type,fp_dtype);
  }

  if (alpha == Arb)
    fprintf(outf,", const %s alpha",fp_type);
  if (beta == Arb)
    fprintf(outf,", const %s beta",fp_type);
  fprintf(outf,")\n");
}


void
generate_core_routine()
{
  fprintf(outf,"/*\n * ");
  generate_core_routine_comment();
  fprintf(outf,"\n */\n");
  fprintf(outf,"void\n");
  generate_core_routine_header();
  fprintf(outf,"{\n");

  nst();
  generate_core_code();
  unst();

  fprintf(outf,"}\n");  
}

/**********************************************************************
 *			 HEADER PRINTING CODE                         *
 **********************************************************************/

void
print_date_string()
{
  time_t tloc;
  struct tm*tms;
  char buf[BUFSIZ];
  time(&tloc);
  tms = localtime(&tloc);
  strftime(buf,BUFSIZ,"%A %B %d %Y, %T %Z",tms);
  fprintf(outf,"%s",buf);
}


void
printUsageMsg()
{

  idnt(); 
  fprintf(outf,"**\n");

  idnt(); 
  fprintf(outf,"** Routine Usage: ");
  generate_core_routine_comment();
  fprintf(outf,"\n");

  idnt(); 
  fprintf(outf,"**    ");
  generate_core_routine_header();

  idnt(); 
  fprintf(outf,"** where\n");

  idnt(); 
  fprintf(outf,"**  ");
  if (opA == Transpose)
    fprintf(outf,"transpose(A)");
  else 
    fprintf(outf,"A");
  fprintf(outf," is an MxK matrix\n");

  idnt(); 
  fprintf(outf,"**  ");
  if (opB == Transpose)
    fprintf(outf,"transpose(B)");
  else 
    fprintf(outf,"B");
  fprintf(outf," is an KxN matrix\n");

  idnt(); 
  fprintf(outf,"**  ");
  fprintf(outf,"C");
  fprintf(outf," is an MxN matrix\n");

  idnt(); 
  fprintf(outf,"**  Astride is the number of entries between the start of each row of A\n");
  

  idnt(); 
  fprintf(outf,"**  Bstride is the number of entries between the start of each row of B\n");
  

  idnt(); 
  fprintf(outf,"**  Cstride is the number of entries between the start of each row of C\n");

  idnt(); 
  fprintf(outf,"**\n");


}


void
generate_header(int argc,char *argv[])
{
  int i;

  idnt();
  fprintf(outf,"/*\n");

  idnt();
  fprintf(outf,"**\n");


  idnt();
  fprintf(outf,"** PHiPAC Matrix-Matrix Code for the operation:\n");

  idnt();
  fprintf(outf,"**    C = ");
  if (alpha == Arb)
    fprintf(outf,"alpha*");
  else if (alpha == Fix)
    fprintf(outf,"%s*",alpha_val);
  else if (alpha == nOne)
    fprintf(outf,"-");
  if (opA == Transpose)
    fprintf(outf,"transpose(A)*");
  else 
    fprintf(outf,"A*");
  if (opB == Transpose)
    fprintf(outf,"transpose(B)");
  else 
    fprintf(outf,"B");
  if (beta == One)
    fprintf(outf," + C");
  else if (beta == Arb)
    fprintf(outf," + beta*C");
  else if (beta == Fix)
    fprintf(outf," + %s*C",beta_val);
  fprintf(outf,"\n");

  idnt();
  fprintf(outf,"**\n");

  idnt();
  fprintf(outf,"** Automatically Generated by mm_cgen (%s) using the command:\n",rcsrevision);

  idnt();
  fprintf(outf,"**    %s ",argv[0]);
  for (i=1;i<argc;i++)
    fprintf(outf,"%s ",argv[i]);
  fprintf(outf,"\n");

  idnt();
  fprintf(outf,"**\n");

  idnt();
  fprintf(outf,"** Run '%s -help' for help.\n",argv[0]);

  idnt();
  fprintf(outf,"**\n");


  if (!do_m_fringe || !do_k_fringe || !do_n_fringe) {
    idnt();
    fprintf(outf,"** WARNING: This routine was generated without code for\n");

    idnt();
    fprintf(outf,"** fringes in the following dimension(s): ");
    
    if (!do_m_fringe)
      fprintf(outf,"M ");
    if (!do_k_fringe)
      fprintf(outf,"K ");
    if (!do_n_fringe)
      fprintf(outf,"N ");
    fprintf(outf,"\n");

    idnt();
    fprintf(outf,"** That means that the routine will only work for matrices with the\n");

    idnt(); 
    fprintf(outf,"** appropriate size(s). See the 'Routine Usage' below for details.\n");

    idnt();
    fprintf(outf,"**\n");
  }



  idnt();
  fprintf(outf,"** Generated on: "); print_date_string();
  fprintf(outf,"\n");

  idnt();
  fprintf(outf,"** Created by: Jeff Bilmes <bilmes@cs.berkeley.edu>\n");
  idnt();
  fprintf(outf,"**             http://www.icsi.berkeley.edu/~bilmes/phipac\n");

  idnt(); 
  fprintf(outf,"**\n");

  printUsageMsg();

  idnt();
  fprintf(outf,"**\n");

  fprintf(outf,"** \"Copyright (c) 1995 The Regents of the University of California.  All\n");
  fprintf(outf,"** rights reserved.\"  Permission to use, copy, modify, and distribute\n");
  fprintf(outf,"** this software and its documentation for any purpose, without fee, and\n");
  fprintf(outf,"** without written agreement is hereby granted, provided that the above\n");
  fprintf(outf,"** copyright notice and the following two paragraphs appear in all copies\n");
  fprintf(outf,"** of this software.\n");
  fprintf(outf,"**\n");
  fprintf(outf,"** IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n");
  fprintf(outf,"** DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n");
  fprintf(outf,"** OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n");
  fprintf(outf,"** CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n");
  fprintf(outf,"**\n");
  fprintf(outf,"** THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n");
  fprintf(outf,"** INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n");
  fprintf(outf,"** AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n");
  fprintf(outf,"** ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n");
  fprintf(outf,"** PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n");

  idnt();
  fprintf(outf,"**\n");

  idnt();
  fprintf(outf,"*/\n\n");

}

/**********************************************************************
 *			 MAIN ROUTINE                                 *
 **********************************************************************/


int main(int argc,char*argv[])
{
  parse_args(argc,argv);
  generate_header(argc,argv);
  generate_core_routine();
  exit(0);
}

