/*
** mm_lgen: Matrix Multiply Level n code Generator
**    the PHiPAC matrix-matrix multiply Level-N generator.
** 
** $Revision: 1.5 $
**
** See the man page mm_lgen.1 for details.
**
** Written by: 
**      Jeff Bilmes <bilmes@icsi.berkeley.edu>
**      http://www.icsi.berkeley.edu/~bilmes/phipac
**
** 
** "Copyright (c) 1995 The Regents of the University of California.  All
** rights reserved."  Permission to use, copy, modify, and distribute
** this software and its documentation for any purpose, without fee, and
** without written agreement is hereby granted, provided that the above
** copyright notice and the following two paragraphs appear in all copies
** of this software.
** 
** IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
** DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
** OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
** CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**  
** THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
** INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
** AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
** ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
** PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
** 
*/

/*
Use three routine names:
   1) fixed-size l0 core with l1 fixed size, no_fringes -fixed_l0_routine
   2) general-sized l0 core with variable size, no_fringes -gen_l0_routine
   2) general l0 core, fringes.    -gen_routine

Options:
  mm_lgen -l0 M0 K0 N0 -l1 M1 K1 N1 -sp [] 
   -gen_routine  (if only this option given, use this)
   -gen_nf_routine no fring routine
   -fixed_routine



Also, write a general perl script that will
take the options and generate a complete
*/

static const char rcsid[] = "$Header: /db/bilmes/phipac/mm_gen/RCS/mm_lgen.c,v 1.5 1997/04/10 04:07:45 bilmes Exp $";
static const char rcsrevision[] = "$Revision: 1.5 $";


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>

#include <assert.h>

/**********************************************************************
 *			TYPEDEFS AND ENUMS                            *
 **********************************************************************/


#define MIN(a,b) ((a)<(b)?(a):(b))

typedef enum { True = 1, False = 0 } bool;


typedef enum {     Zero, /* the value is fixed at zero */
		   One,  /* the value is fixed at one */
		   Fix,  /* the value is fixed at some constant */
		   Arb   /* the value is parameterized  */
	       } scaleType;

/* matrix transposition */
typedef enum { Normal, Transpose } op;

/* The two types of N and M fringe */
typedef enum { Not_Fringe,   /* This is not fringe code */
	       L1_Fringe,   /* In the L1 fringe but not L0 */
               L0_Fringe    /* In the L0 fringe */
} FringeType;

/* software pipelining options */
typedef enum {
     OneStage,  /* i.e., no software pipelining */
     TwoStageLoadMul, /* Two stage, {Load-Mul, Add} */
     TwoStageMulAdd,  /* Two stage, {Load, Mul-Add} */
     ThreeStage,  /* Three stage, {Load, Mul, Add} */
} SoftPipeType;


/**********************************************************************
 *			COMMAND LINE ARGUMENTS                        *
 **********************************************************************/

/* where to put the resulting routine */
char *outfile=NULL;
FILE *outf = NULL;

/* the C code name of the resulting routine */
char routine_name[BUFSIZ] = { '\0' };

/* Name of the general L0-blocked matrix multiply routine to call. */
char *gen_routine=NULL;

/* Name of the general L0-blocked matrix multiply fringeless routine to call. */
char *gen_nf_routine=NULL;
/* Name of the fixed (M0*M1,K0*K1+c,N0*N1) L0/L1-blocked routine to call. */
char *fixed_routine=NULL;


/* The blocking parameters. We must have
   the Level 0 parameters set. If the level 1 parameters
   are set, we produce a routine for a fixed matrix size. */
size_t M0=0,K0=0,N0=0;
size_t M1=0,K1=0,N1=0;
#define L1SET (M1 != 0)
#define L0SET (M0 != 0)
size_t Mc,Kc,Nc; /* the L1 fringes if L1SET is true */

/* determins the operation on either the A or B matrix. */
op opA = Normal;
op opB = Normal;


/* Sotware Pipelining Option */
SoftPipeType softpipe = OneStage;

/* True if we do the register block fringes, False if we don't. */
bool do_m_fringe=True;
bool do_k_fringe=True;
bool do_n_fringe=True;

scaleType alpha = One, beta = One;
char *alpha_val = "1.0", *beta_val = "1.0";

/* The following are the C floating point types that will be used.
   They may be safely changed if needed for a particular compiler.  */
const char* fp_types[] = {   
  "float",   /* presumably IEEE single prec */
  "double",  /* presumably IEEE double prec */
  "long double" /* presumably IEEE double extended prec */
};
const char fp_chars[] = { 'f', 'd', 'l' };
#define FP_TYPE_INVALID (-1)
#define FP_TYPE_FLOAT 0
#define FP_TYPE_DOUBLE 1
#define FP_TYPE_LONG_DOUBLE 2
#define FP_TYPE_DEFAULT FP_TYPE_FLOAT
int fp_type_num = FP_TYPE_INVALID;
int fp_atype_num = FP_TYPE_INVALID; /* accumulator type */
int fp_dtype_num = FP_TYPE_INVALID; /* destination type */
const char* fp_type;
const char* fp_atype; /* accumulator type */
const char* fp_dtype; /* destination type */
char  fp_char;
char  fp_achar;
char  fp_dchar;

/*
 * Do we immediately call down the lower level routine or not.
 */
bool  call_down = False;

/*
 * Code generation Nesting parameters 
 */
#define DEFAULT_SPACE_PER_NEST 3
int nesting_level=0;
int space_per_nest = DEFAULT_SPACE_PER_NEST;
char space_char = ' ';

/**********************************************************************
 *			MISCELLANEOUS ROUTINES                        * 
 **********************************************************************/


#define SP_ADJ_K(K) ((softpipe == OneStage) ? (K) : ((softpipe == TwoStageMulAdd || softpipe == TwoStageLoadMul) ? ((K)+1) : ((K)+2)))

#define SP_C  ((softpipe == OneStage) ? (0) : ((softpipe == TwoStageMulAdd || softpipe == TwoStageLoadMul) ? (1) : (2)))




void nst() { nesting_level++; }
void unst() { nesting_level--; }
void idnt()
{
  int i;
  for (i=0;i<space_per_nest*nesting_level;i++)
    fprintf(outf,"%c",space_char);
}

/* Returns true of arg is a 2^k for some k */
int pow_of_2(int arg)
{
  int i;
  int bits = 0;
  if (arg < 0) return 0;
  for (i=0;i<sizeof(arg)*8;i++) {
    bits += (arg & 0x1);
    if (bits > 1) return 0;
    arg >>= 1;
  }
  return 1;
}

/* integer log2(u) */
int ilog2 (int u)
{
  int   i;
  if (u == 0)
    return (-1);
  for (i = 0; (u & 1) == 0; i++, u >>= 1);
  return (u == 1 ? i : -1);
}

/* returns the highest bit of u */
int highbit( int u )
{
  int res = 1;
  if (u == 0)
    return 0;
  while (u >>= 1) {
    res <<= 1;
  }
  return res;
}




/**********************************************************************
 *			COMMAND LINE PARSING                          *
 **********************************************************************/


void
usage(char *name)
{
  fprintf(stderr,"usage: %s [OPTIONS]\n",name);
  fprintf(stderr,"where [OPTIONS] include:\n");  

  fprintf(stderr,"Semantics options:\n");
  fprintf(stderr,"  -opA [N|T]                 : A matrix op. Normal|Transpose\n");
  fprintf(stderr,"  -opB [N|T]                 : B matrix op. Normal|Transpose\n");
  fprintf(stderr,"  -no_m_fringe               : don't generate an M reg block fringe\n");
  fprintf(stderr,"  -no_k_fringe               : don't generate a K reg block fringe\n");
  fprintf(stderr,"  -no_n_fringe               : don't generate an N reg block fringe\n");
  fprintf(stderr,"  -no_fringes                : don't generate an M,K, or N reg block fringes\n");
  fprintf(stderr,"  -alpha [<val>|c]           : fix alpha at value <val> or arbitrary\n");
  fprintf(stderr,"  -beta [<val>|c]            : fix beta at value <val> or arbitrary\n");


  fprintf(stderr,"Optimization options:\n");
  fprintf(stderr,"  -l0 M0 K0 N0               : register (L0) blocking parameters\n");
  fprintf(stderr,"  -l1 M1 K1 N1               : L1 blocking parameters\n");

  fprintf(stderr,"  -calldown                  : check/call down lower routines first\n");

  fprintf(stderr,"Precision options:\n");
  fprintf(stderr,"  -prec [single|double|ldouble]  : Precision\n");
  fprintf(stderr,"  -sprec [single|double|ldouble] : Source Precision\n");
  fprintf(stderr,"  -aprec [single|double|ldouble] : Accumulator Precision\n");
  fprintf(stderr,"  -dprec [single|double|ldouble] : Destination Precision\n");

  fprintf(stderr,"Core code routine names options:\n");

  fprintf(stderr,"  -gen_rout  name         : General MM routine\n"); 

  fprintf(stderr,"  -gen_nf_rout  name      : General nofring (M0,K0,N0) routine\n"); 
  fprintf(stderr,"  -fixed_rout  name       : Fixed (M0M1,K0K1,N0N0) routine\n");
  fprintf(stderr,"  -sp [1|2lm|2ma|3]       : software pipelining option for core routine.\n");

  fprintf(stderr,"Code generation & Misc. options:\n");
  fprintf(stderr,"  -help                      : Print this message\n"); 
  fprintf(stderr,"  -file name                 : Write to file 'name'\n"); 
  fprintf(stderr,"  -routine_name name         : Name of resulting routine\n");
  fprintf(stderr,"  -spacechar c               : char to use as space\n");
  fprintf(stderr,"  -numspaces i               : spaces per nest\n");
  fprintf(stderr,"  -version                   : print version and exit.\n");

}

/*
 * Check if n additional parameters are available.
 */
void
check_remaining_params(int argc,int i,int n)
{
  if ((i+n+1)>argc) {
    fprintf(stderr,"Error: too few command line parameters.\n");
    exit(-1);
  }
}

void
parse_args(int argc,char*argv[])
{
  int i;
  if (argc == 1) {
    usage(argv[0]);
    exit(-1);
  }
  /* args initialize */
  outf = stdout;
  for (i=1;i<argc;i++) {
    if (argv[i][0] == '-') {
      if (!strcmp(&(argv[i][1]),"file")) {
	check_remaining_params(argc,i,1);
        outfile = argv[++i];
	if (!strcmp(outfile,"-")) {
	  /* do nothing */
	} else {
	  if ((outf = fopen(outfile,"w")) == NULL) {
	    fprintf(stderr,"Unable to open file %s\n",outfile);
	    exit(1);
	  }
	}
      } else if (!strncmp(&(argv[i][1]),"l0",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,3);
	if (
	    ((M0 = atoi(argv[++i])) <= 0)
	      ||
	    ((K0 = atoi(argv[++i])) <= 0)
	      ||
	    ((N0 = atoi(argv[++i])) <= 0)
	      ) {
	    fprintf(stderr,"Error: Invalid l0 blocking values.\n");
	    exit(-1);
	  }
      } else if (!strncmp(&(argv[i][1]),"l1",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,3);
	if (
	    ((M1 = atoi(argv[++i])) <= 0)
	      ||
	    ((K1 = atoi(argv[++i])) <= 0)
	      ||
	    ((N1 = atoi(argv[++i])) <= 0)
	      ) {
	    fprintf(stderr,"Error: Invalid l1 blocking values.\n");
	    exit(-1);
	  }
      } else if (!strncmp(&(argv[i][1]),"sp",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"1")) {
	  softpipe = OneStage;
	} else if (!strcmp(argv[i],"2lm")) {
	  softpipe = TwoStageLoadMul;
	} else if (!strcmp(argv[i],"2ma")) {
	  softpipe = TwoStageMulAdd;
	} else if (!strcmp(argv[i],"3")) {
	  softpipe = ThreeStage;
	} else {
	  fprintf(stderr,"Error: Invalid sp option.\n");
	  usage(argv[0]);
	  exit(-1);
	}
      } else if (!strncmp(&(argv[i][1]),"opA",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"N") || !strcmp(argv[i],"n")) { 
	  opA = Normal;
	} else if (!strcmp(argv[i],"T") || !strcmp(argv[i],"t")) { 
	  opA = Transpose;
	} else
	  fprintf(stderr,"Invalid option (%s), ignored\n",argv[i]);
      } else if (!strncmp(&(argv[i][1]),"opB",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"N") || !strcmp(argv[i],"n")) { 
	  opB = Normal;
	} else if (!strcmp(argv[i],"T") || !strcmp(argv[i],"t")) { 
	  opB = Transpose;
	} else
	  fprintf(stderr,"Invalid option (%s), ignored\n",argv[i]);
      } else if (!strncmp(&(argv[i][1]),"alpha",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"c")) {
	  alpha = Arb;
	} else {
	  char *ptr;
	  double tmp;
	  tmp = strtod(argv[i],&ptr);
	  if (ptr == argv[i]) {
	    fprintf(stderr,"Invalid numeric parameter (%s)\n",argv[i]);
	    exit(-1);
	  }
	  alpha_val = argv[i];
	  if (tmp == 1.0)
	    alpha = One;
	  else if (tmp == 0.0) {
	    fprintf(stderr,"Error: Fixed zero value for alpha nulify's operation.\n");
	    exit(-1);	    
	  } else
	    alpha = Fix;
	}
      } else if (!strncmp(&(argv[i][1]),"beta",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"c")) {
	  beta = Arb;
	} else {
	  char *ptr;
	  double tmp;
	  tmp = strtod(argv[i],&ptr);
	  if (ptr == argv[i]) {
	    fprintf(stderr,"Invalid numeric parameter (%s)\n",argv[i]);
	    exit(-1);
	  }
	  beta_val = argv[i];
	  if (tmp == 1.0)
	    beta = One;
	  else if (tmp == 0.0)
	    beta = Zero;
	  else 
	    beta = Fix;
	}
      } else if (!strncmp(&(argv[i][1]),"no_fringes",strlen(&(argv[i][1])))) {
	do_m_fringe = do_k_fringe = do_n_fringe = False;
      } else if (!strncmp(&(argv[i][1]),"no_m_fringe",strlen(&(argv[i][1])))) {
	do_m_fringe = False;
      } else if (!strncmp(&(argv[i][1]),"no_k_fringe",strlen(&(argv[i][1])))) {
	do_k_fringe = False;
      } else if (!strncmp(&(argv[i][1]),"no_n_fringe",strlen(&(argv[i][1])))) {
	do_n_fringe = False;
      } else if (!strncmp(&(argv[i][1]),"gen_routine",strlen(&(argv[i][1])))) {
	i++;
	gen_routine = argv[i];
      } else if (!strncmp(&(argv[i][1]),"gen_nf_routine",strlen(&(argv[i][1])))) {
	i++;
	gen_nf_routine = argv[i];
      } else if (!strncmp(&(argv[i][1]),"fixed_routine",strlen(&(argv[i][1])))) {
	i++;
	fixed_routine = argv[i];
      } else if (!strncmp(&(argv[i][1]),"routine_name",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	strncpy(routine_name,argv[i],MIN(BUFSIZ,strlen(argv[i])));
      } else if ((!strncmp(&(argv[i][1]),"precision",strlen(&(argv[i][1])))) ||
		 (!strncmp(&(argv[i][1]),"sprecision",strlen(&(argv[i][1]))))) {
	check_remaining_params(argc,i,1);
	if (!strncmp(argv[i+1],"single",strlen(argv[i+1])) ||
	    !strncmp(argv[i+1],"float",strlen(argv[i+1])))
	  fp_type_num = FP_TYPE_FLOAT;
	else if (!strncmp(argv[i+1],"double",strlen(argv[i+1])))
	  fp_type_num = FP_TYPE_DOUBLE;
	else if (!strncmp(argv[i+1],"ldouble",strlen(argv[i+1])))
	  fp_type_num = FP_TYPE_LONG_DOUBLE;
	else 
	  fprintf(stderr,"Invalid precision (%s), ignored\n",argv[i+1]);
	i++;
      } else if (!strncmp(&(argv[i][1]),"aprecision",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	if (!strncmp(argv[i+1],"single",strlen(argv[i+1])) ||
	    !strncmp(argv[i+1],"float",strlen(argv[i+1])))
	  fp_atype_num = FP_TYPE_FLOAT;
	else if (!strncmp(argv[i+1],"double",strlen(argv[i+1])))
	  fp_atype_num = FP_TYPE_DOUBLE;
	else if (!strncmp(argv[i+1],"ldouble",strlen(argv[i+1])))
	  fp_atype_num = FP_TYPE_LONG_DOUBLE;
	else 
	  fprintf(stderr,"Invalid precision (%s), ignored\n",argv[i+1]);
	i++;
      } else if (!strncmp(&(argv[i][1]),"dprecision",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	if (!strncmp(argv[i+1],"single",strlen(argv[i+1])) ||
	    !strncmp(argv[i+1],"float",strlen(argv[i+1])))
	  fp_dtype_num = FP_TYPE_FLOAT;
	else if (!strncmp(argv[i+1],"double",strlen(argv[i+1])))
	  fp_dtype_num = FP_TYPE_DOUBLE;
	else if (!strncmp(argv[i+1],"ldouble",strlen(argv[i+1])))
	  fp_dtype_num = FP_TYPE_LONG_DOUBLE;
	else 
	  fprintf(stderr,"Invalid precision (%s), ignored\n",argv[i+1]);
	i++;
      } else if (!strncmp(&(argv[i][1]),"spacechar",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	space_char = *argv[++i];
      } else if (!strncmp(&(argv[i][1]),"numspaces",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	space_per_nest = atoi(argv[++i]);
      } else if (!strncmp(&(argv[i][1]),"nocalldown",strlen(&(argv[i][1])))) {
	call_down = False;
      } else if (!strncmp(&(argv[i][1]),"calldown",strlen(&(argv[i][1])))) {
	call_down = True;
      } else if (!strncmp(&(argv[i][1]),"help",strlen(&(argv[i][1])))) {
	usage(argv[0]);
	exit(0);
      } else if (!strncmp(&(argv[i][1]),"-help",strlen(&(argv[i][1])))) {
	usage(argv[0]);
	exit(0);
      } else if (!strncmp(&(argv[i][1]),"version",strlen(&(argv[i][1])))) {
	fprintf(stderr,"%s: %s\n",argv[0],rcsrevision);
	exit(0);
      } else  {
        fprintf(stderr,"Error: Unknown or invalid argument (%s)\n",argv[i]);
	exit(-1);
      }
    }
  }

  /* finish initialize & check */
  if (fp_type_num == FP_TYPE_INVALID)
    fp_type_num = FP_TYPE_DEFAULT;
  if (fp_atype_num == FP_TYPE_INVALID)
    fp_atype_num = fp_type_num;
  if (fp_dtype_num == FP_TYPE_INVALID)
    fp_dtype_num = fp_type_num;


  fp_type = fp_types[fp_type_num];
  fp_atype = fp_types[fp_atype_num];
  fp_dtype = fp_types[fp_dtype_num];
  fp_char = fp_chars[fp_type_num];
  fp_achar = fp_chars[fp_atype_num];
  fp_dchar = fp_chars[fp_dtype_num];


  if (gen_routine == NULL) {
    fprintf(stderr,"Error: Must provide a gen_routine parameter\n");
    usage(argv[0]);
    exit(-1);
  }

  if (L1SET && !L0SET) {
    /* assume we're doing Ln:n>=2 blocking where the parameters
       M1, K1, and N1 indicate the sub block sizes in elements. */
    if (gen_nf_routine != NULL) {
      fprintf(stderr,"Error: no gen_nf_routine possible for Ln:n>=2\n");
      usage(argv[0]);
      exit(-1);
    }
    M0=K0=N0=1;
  } else if (!L1SET && !L0SET) {
    fprintf(stderr,"Error with -l0 and -l1 options\n");
    usage(argv[0]);
    exit(-1);
  }

  if (gen_routine == NULL) {
    fprintf(stderr,"Error: must provide at least a gen_routine option.\n");
    usage(argv[0]);
    exit(-1);
  }
 
  if (strlen(routine_name) == 0)
    sprintf(routine_name,"mul_m%c%sm%c%s_m%c",
	    fp_char,(opA==Normal?"":"t"),
	    fp_char,(opB==Normal?"":"t"),
	    fp_dchar);

 Mc = M0*M1;
 Kc = K0*K1;
 Nc = N0*N1;
 
}


/**********************************************************************
 *			CODE GENERATION (i.e., the meat)              *
 **********************************************************************/

/***************/
/* BETA Code   */
/***************/


void
generate_beta_code()
{
  if (beta == One)
    return;  /* no scaling necessary */

  if (beta == Arb || beta == Fix) {
    if (beta == Fix) {
      idnt(); fprintf(outf,"{\n");
      nst(); /* 1 */
      idnt(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      idnt(); fprintf(outf,"cpre = C + M*Cstride;\n");
      idnt();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nst(); /* 2 */
      idnt(); fprintf(outf,"cpe = cprb + N;\n");
      idnt(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nst(); /* 3 */
      idnt(); fprintf(outf,"*cp *= (%s);\n",(beta==Arb?"beta":beta_val));
      unst(); /* 3 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 2 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 1 */
      idnt(); fprintf(outf,"}\n");
    } else { /* beta == Arb */
      idnt(); fprintf(outf,"if (beta == 0.0) {\n");
      nst(); /* 1 */
      idnt(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      idnt(); fprintf(outf,"cpre = C + M*Cstride;\n");
      idnt();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nst(); /* 2 */
      idnt(); fprintf(outf,"cpe = cprb + N;\n");
      idnt(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nst(); /* 3 */
      idnt(); fprintf(outf,"*cp = 0.0;\n");
      unst(); /* 3 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 2 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 1 */
      idnt(); fprintf(outf,"} else if (beta != 1.0) {\n");
      nst(); /* 1 */
      idnt(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      idnt(); fprintf(outf,"cpre = C + M*Cstride;\n");
      idnt();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nst(); /* 2 */
      idnt(); fprintf(outf,"cpe = cprb + N;\n");
      idnt(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nst(); /* 3 */
      idnt(); fprintf(outf,"*cp *= (beta);\n");
      unst(); /* 3 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 2 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 1 */
      idnt(); fprintf(outf,"}\n");
    }
  } else if (beta == Zero) {
      idnt(); fprintf(outf,"{\n");
      nst(); /* 1 */
      idnt(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      idnt(); fprintf(outf,"cpre = C + M*Cstride;\n");
      idnt();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nst(); /* 2 */
      idnt(); fprintf(outf,"cpe = cprb + N;\n");
      idnt(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nst(); /* 3 */
      idnt(); fprintf(outf,"*cp = 0.0;\n");
      unst(); /* 3 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 2 */
      idnt(); fprintf(outf,"}\n");    
      unst(); /* 1 */
      idnt(); fprintf(outf,"}\n");
  }
}



/***************/
/* K-LOOP Code */
/***************/


void
generate_k_loop_open()
{
  /* K loop */
  idnt();
  fprintf(outf,
	  "for (k=0,bp=b,ap=a; ");
  fprintf(outf,"(k+%d)<=K; k+=%d,",SP_ADJ_K(Kc),SP_ADJ_K(Kc));
  if (opB == Normal) {
    fprintf(outf,
	    "bp+=%d*Bstride,",SP_ADJ_K(Kc));
  } else {
    fprintf(outf,
	    "bp+=%d,",SP_ADJ_K(Kc));
  }
  if (opA == Normal) {
    fprintf(outf,
	    "ap+=%d) {\n",SP_ADJ_K(Kc));
  } else {
    fprintf(outf,
	    "ap+=%d*Astride) {\n",SP_ADJ_K(Kc));
  }
  nst();
}

void
generate_k_loop_body(FringeType mfringe, 
		     FringeType kfringe, 
		     FringeType nfringe)
{
  char Mc_str[BUFSIZ];
  char Kc_str[BUFSIZ];
  char Nc_str[BUFSIZ];

  sprintf(Mc_str,"%d",Mc);
  sprintf(Kc_str,"%d",SP_ADJ_K(Kc));
  sprintf(Nc_str,"%d",Nc);

  idnt();
  if (fixed_routine != NULL && mfringe == Not_Fringe && 
      kfringe == Not_Fringe && nfringe == Not_Fringe) {
    fprintf(outf,
	    "%s(ap,bp,cp,Astride,Bstride,Cstride%s);\n",
	    fixed_routine,
	    ((alpha == Arb) ? ",alpha" : "" ));
  } else if (gen_nf_routine != NULL && mfringe != L0_Fringe && 
	     nfringe != L0_Fringe && kfringe != L0_Fringe) {
    fprintf(outf,
	    "%s(%s,%s,%s,ap,bp,cp,Astride,Bstride,Cstride%s);\n",
	    gen_nf_routine,
	    (mfringe == Not_Fringe ? Mc_str : "m_marg"),
	    (kfringe == Not_Fringe ? Kc_str : "k_marg"),
	    (nfringe == Not_Fringe ? Nc_str : "n_marg"),
	    ((alpha == Arb) ? ",alpha" : "" ));
  } else {
    fprintf(outf,
	    "%s(%s,%s,%s,ap,bp,cp,Astride,Bstride,Cstride%s);\n",
	    gen_routine,
	    (mfringe == Not_Fringe ? Mc_str : (mfringe == L0_Fringe ? "M-m" : "m_marg")),
	    (kfringe == Not_Fringe ? Kc_str : (kfringe == L0_Fringe ? "K-k" : "k_marg")),
	    (nfringe == Not_Fringe ? Nc_str : (nfringe == L0_Fringe ? "N-n" : "n_marg")),
	    ((alpha == Arb) ? ",alpha" : "" ));
    }
}

void
generate_k_loop_close()
{
  unst();
  idnt(); fprintf(outf,"}\n");
}


void
generate_k_fringe_body(FringeType mfringe, FringeType nfringe)
{
  /* finish up K */
  if (gen_nf_routine != NULL) {
    idnt();
    fprintf(outf,"if (k + %d <= K) {\n",SP_ADJ_K(K0));
    nst();

    generate_k_loop_body(mfringe,L1_Fringe,nfringe);

    idnt();
    fprintf(outf,"k+=k_marg;");
    if (opB == Normal) {
      fprintf(outf,"bp+=k_marg*Bstride;");
    } else {
      fprintf(outf,"bp+=k_marg;");
    }
    if (opA == Normal) {
      fprintf(outf,"ap+=k_marg;\n");
    } else {
      fprintf(outf,"ap+=k_marg*Astride;\n");
    }

    unst();
    idnt(); fprintf(outf,"}\n");
  }
  

  idnt();
  fprintf(outf,"if (k < K) {\n");
  nst();

  generate_k_loop_body(mfringe,L0_Fringe,nfringe);

  unst();
  idnt(); fprintf(outf,"}\n");


}




/***************/
/* N-LOOP Code */
/***************/

void
generate_n_loop_open()
{
  /* N loop */
  idnt();
  fprintf(outf,"for (n=0,b=B,cp=c; (n+%d)<=N; n+=%d,",Nc,Nc);
  if (opB == Normal) {
    fprintf(outf,"b+=%d,",Nc);
  } else {
    fprintf(outf,"b+=%d*Bstride,",Nc);
  }
  fprintf(outf,"cp+=%d) {\n",Nc);
  nst();
}

void
generate_n_loop_close()
{
  unst(); 
  idnt(); fprintf(outf,"}\n");    
}

void
generate_n_fringe_body(FringeType mfringe)
{
  /* finish up N */

  if (gen_nf_routine != NULL) {

    idnt(); fprintf(outf,"if (n + %d <= N) {\n",N0);
    nst();

    generate_k_loop_open();
    generate_k_loop_body(mfringe,Not_Fringe,L1_Fringe);
    generate_k_loop_close();  
    generate_k_fringe_body(mfringe,L1_Fringe);

    idnt();
    fprintf(outf,"n+=n_marg;");
    if (opB == Normal) {
      fprintf(outf,"b+=n_marg;");
    } else {
      fprintf(outf,"b+=n_marg*Bstride;");
    }
    fprintf(outf,"cp+=n_marg;\n");

    unst();
    idnt(); fprintf(outf,"}\n");  

  }

  idnt(); fprintf(outf,"if (n < N) {\n");
  nst();

  generate_k_loop_open();
  generate_k_loop_body(mfringe,Not_Fringe,L0_Fringe);
  generate_k_loop_close();  
  generate_k_fringe_body(mfringe,L0_Fringe);

  unst();
  idnt(); fprintf(outf,"}\n");  
}



/***************/
/* M-LOOP Code */
/***************/


void
generate_m_loop_open()
{
  /* M loop */
  idnt();
  fprintf(outf,"for (m=0; (m+%d)<=M; m+=%d) {\n",Mc,Mc);

  nst();
  idnt(); fprintf(outf,"c = C + m*Cstride;\n");
  if (opA == Normal) {
    idnt(); fprintf(outf,"a = A + m*Astride;\n");
  } else {
    idnt(); fprintf(outf,"a = A + m;\n");
  }
}


void
generate_m_loop_close()
{
  unst();
  idnt(); fprintf(outf,"}\n");    
}

void
generate_m_fringe_body()
{
  /* finish up M */

  if (gen_nf_routine != NULL) {
    idnt(); fprintf(outf,"if (m + %d <= M) {\n",M0);
    nst();

    idnt(); fprintf(outf,"c = C + m*Cstride;\n");
    if (opA == Normal) {
      idnt(); fprintf(outf,"a = A + m*Astride;\n");
    } else {
      idnt(); fprintf(outf,"a = A + m;\n");
    }

    generate_n_loop_open();
    generate_k_loop_open();
    generate_k_loop_body(L1_Fringe,Not_Fringe,Not_Fringe);
    generate_k_loop_close();
    generate_k_fringe_body(L1_Fringe,Not_Fringe);
    generate_n_loop_close();
    generate_n_fringe_body(L1_Fringe);

    idnt(); fprintf(outf,"m+=m_marg;\n");

    unst();
    idnt();fprintf(outf,"}\n");  
  }

  idnt(); fprintf(outf,"if (m < M) {\n");
  nst();

  idnt(); fprintf(outf,"c = C + m*Cstride;\n");
  if (opA == Normal) {
    idnt(); fprintf(outf,"a = A + m*Astride;\n");
  } else {
    idnt(); fprintf(outf,"a = A + m;\n");
  }

  generate_n_loop_open();
  generate_k_loop_open();
  generate_k_loop_body(L0_Fringe,Not_Fringe,Not_Fringe);
  generate_k_loop_close();
  generate_k_fringe_body(L0_Fringe,Not_Fringe);
  generate_n_loop_close();
  generate_n_fringe_body(L0_Fringe);

  unst();
  idnt();fprintf(outf,"}\n");  
}



/*******************/
/* PROLOGUE Code   */
/*******************/


void
generate_routine_header()
{
  fprintf(outf,"%s(const int M, const int K, const int N, const %s *const A, const %s *const B, %s *const C, const int Astride, const int Bstride, const int Cstride",routine_name,fp_type,fp_type,fp_dtype);

  if (alpha == Arb)
    fprintf(outf,", const %s alpha",fp_type);
  if (beta == Arb)
    fprintf(outf,", const %s beta",fp_type);
  fprintf(outf,")\n");
}

void
generate_prologue()
{

  idnt();
  fprintf(outf,"int m,k,n;\n");

  idnt();
  fprintf(outf,"const %s *a,*b;\n",fp_type);
  idnt();
  fprintf(outf,"%s *c;\n",fp_dtype);


  idnt();
  fprintf(outf,"const %s *ap,*bp;\n",fp_type);
  idnt();
  fprintf(outf,"%s *cp;\n",fp_dtype);

  if (gen_nf_routine != NULL) {
    idnt();
    fprintf(outf,"const int m_marg = (M %% %d) - (M %% %d);\n",Mc,M0);

    
    idnt();
    fprintf(outf,"const int k_marg = (K %% %d) - ((K%%%d)-%d)%%%d;\n",
	    SP_ADJ_K(Kc),
	    SP_ADJ_K(Kc),SP_C,K0);
	    

    idnt();
    fprintf(outf,"const int n_marg = (N %% %d) - (N %% %d);\n",Nc,N0);
  }

  if (call_down) {
    idnt();
    fprintf(outf,"if (M < %d && K < %d && N < %d) {\n",
	    Mc+1,SP_ADJ_K(Kc)+1,Nc+1);

    nst();
    idnt();
    fprintf(outf,
	      "%s(M,K,N,A,B,C,Astride,Bstride,Cstride%s);\n",
	      gen_routine,
	      ((alpha == Arb) ? ",alpha" : ""));
    idnt(); fprintf(outf,"return;\n");
    unst();
    idnt(); fprintf(outf,"}\n");
  }


}

/***************/
/* Main Code   */
/***************/

void
generate_externs()
{

  if (gen_routine != NULL) {
    idnt();
    fprintf(outf,"extern %s(const int M, const int K, const int N, const %s *const A, const %s *const B, %s *const C, const int Astride, const int Bstride, const int Cstride",gen_routine,fp_type,fp_type,fp_dtype);

    if (alpha == Arb)
      fprintf(outf,", const %s alpha",fp_type);
    fprintf(outf,");\n");
  }

  if (gen_nf_routine != NULL) {
    idnt();
    fprintf(outf,"extern %s(const int M, const int K, const int N, const %s *const A, const %s *const B, %s *const C, const int Astride, const int Bstride, const int Cstride",gen_nf_routine,fp_type,fp_type,fp_dtype);

    if (alpha == Arb)
      fprintf(outf,", const %s alpha",fp_type);
    fprintf(outf,");\n");
  }

  if (fixed_routine != NULL) {
    idnt();
    fprintf(outf,"extern %s(const %s *const A, const %s *const B, %s *const C, const int Astride, const int Bstride, const int Cstride",fixed_routine,fp_type,fp_type,fp_dtype);

    if (alpha == Arb)
      fprintf(outf,", const %s alpha",fp_type);
    fprintf(outf,");\n");
  }

  fprintf(outf,"\n\n");

}


void
generate_code()
{

  generate_externs();

  fprintf(outf,"void\n");
  generate_routine_header();
  fprintf(outf,"{\n");

  nst();

  generate_prologue();
  generate_beta_code();
  generate_m_loop_open();
  generate_n_loop_open();
  generate_k_loop_open();
  generate_k_loop_body(Not_Fringe,Not_Fringe,Not_Fringe);
  generate_k_loop_close();
  generate_k_fringe_body(Not_Fringe,Not_Fringe);
  generate_n_loop_close();
  generate_n_fringe_body(Not_Fringe);
  generate_m_loop_close();
  generate_m_fringe_body();

  unst();
  idnt();fprintf(outf,"}\n");  

}

/**********************************************************************
 *			 HEADER PRINTING CODE                         *
 **********************************************************************/


void
print_date_string()
{
  time_t tloc;
  struct tm*tms;
  char buf[BUFSIZ];
  time(&tloc);
  tms = localtime(&tloc);
  strftime(buf,BUFSIZ,"%A %B %d %Y, %T %Z",tms);
  fprintf(outf,"%s",buf);
}


void
printUsageMsg()
{

  idnt(); 
  fprintf(outf,"**\n");

  idnt(); 
  fprintf(outf,"** Routine Usage: ");
  fprintf(outf,"\n");

  idnt(); 
  fprintf(outf,"**    ");
  generate_routine_header();

  idnt(); 
  fprintf(outf,"** where\n");

  idnt(); 
  fprintf(outf,"**  ");
  if (opA == Transpose)
    fprintf(outf,"transpose(A)");
  else 
    fprintf(outf,"A");
  fprintf(outf," is an MxK matrix\n");

  idnt(); 
  fprintf(outf,"**  ");
  if (opB == Transpose)
    fprintf(outf,"transpose(B)");
  else 
    fprintf(outf,"B");
  fprintf(outf," is an KxN matrix\n");

  idnt(); 
  fprintf(outf,"**  ");
  fprintf(outf,"C");
  fprintf(outf," is an MxN matrix\n");

  idnt(); 
  fprintf(outf,"**  Astride is the number of entries between the start of each row of A\n");
  

  idnt(); 
  fprintf(outf,"**  Bstride is the number of entries between the start of each row of B\n");
  

  idnt(); 
  fprintf(outf,"**  Cstride is the number of entries between the start of each row of C\n");

  idnt(); 
  fprintf(outf,"**\n");


}


void
generate_header(int argc,char *argv[])
{
  int i;

  idnt();
  fprintf(outf,"/*\n");

  idnt();
  fprintf(outf,"**\n");


  idnt();
  fprintf(outf,"** PHiPAC Matrix-Matrix Code for the operation:\n");

  idnt();
  fprintf(outf,"**    C = ");
  if (alpha == Arb)
    fprintf(outf,"alpha*");
  else if (alpha == Fix)
    fprintf(outf,"%s*",alpha_val);
  if (opA == Transpose)
    fprintf(outf,"transpose(A)*");
  else 
    fprintf(outf,"A*");
  if (opB == Transpose)
    fprintf(outf,"transpose(B)");
  else 
    fprintf(outf,"B");
  if (beta == One)
    fprintf(outf," + C");
  else if (beta == Arb)
    fprintf(outf," + beta*C");
  else if (beta == Fix)
    fprintf(outf," + %s*C",beta_val);
  fprintf(outf,"\n");

  idnt();
  fprintf(outf,"**\n");

  idnt();
  fprintf(outf,"** Automatically Generated by mm_lgen (%s) using the command:\n",rcsrevision);

  idnt();
  fprintf(outf,"**    %s ",argv[0]);
  for (i=1;i<argc;i++)
    fprintf(outf,"%s ",argv[i]);
  fprintf(outf,"\n");

  idnt();
  fprintf(outf,"**\n");

  idnt();
  fprintf(outf,"** Run '%s -help' for help.\n",argv[0]);

  idnt();
  fprintf(outf,"**\n");


  if (!do_m_fringe || !do_k_fringe || !do_n_fringe) {
    idnt();
    fprintf(outf,"** WARNING: This routine was generated without code for\n");

    idnt();
    fprintf(outf,"** fringes in the following dimension(s): ");
    
    if (!do_m_fringe)
      fprintf(outf,"M ");
    if (!do_k_fringe)
      fprintf(outf,"K ");
    if (!do_n_fringe)
      fprintf(outf,"N ");
    fprintf(outf,"\n");

    idnt();
    fprintf(outf,"** That means that the routine will only work for matrices with the\n");

    idnt(); 
    fprintf(outf,"** appropriate size(s). See the 'Routine Usage' below for details.\n");

    idnt();
    fprintf(outf,"**\n");
  }



  idnt();
  fprintf(outf,"** Generated on: "); print_date_string();
  fprintf(outf,"\n");

  idnt();
  fprintf(outf,"** Created by: Jeff Bilmes <bilmes@cs.berkeley.edu>\n");
  idnt();
  fprintf(outf,"**             http://www.icsi.berkeley.edu/~bilmes/phipac\n");

  idnt(); 
  fprintf(outf,"**\n");

  printUsageMsg();

  idnt();
  fprintf(outf,"**\n");

  fprintf(outf,"** \"Copyright (c) 1995 The Regents of the University of California.  All\n");
  fprintf(outf,"** rights reserved.\"  Permission to use, copy, modify, and distribute\n");
  fprintf(outf,"** this software and its documentation for any purpose, without fee, and\n");
  fprintf(outf,"** without written agreement is hereby granted, provided that the above\n");
  fprintf(outf,"** copyright notice and the following two paragraphs appear in all copies\n");
  fprintf(outf,"** of this software.\n");
  fprintf(outf,"**\n");
  fprintf(outf,"** IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n");
  fprintf(outf,"** DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n");
  fprintf(outf,"** OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n");
  fprintf(outf,"** CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n");
  fprintf(outf,"**\n");
  fprintf(outf,"** THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n");
  fprintf(outf,"** INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n");
  fprintf(outf,"** AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n");
  fprintf(outf,"** ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n");
  fprintf(outf,"** PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n");

  idnt();
  fprintf(outf,"**\n");

  idnt();
  fprintf(outf,"*/\n\n");

}

/**********************************************************************
 *			 MAIN ROUTINE                                 *
 **********************************************************************/


int main(int argc,char*argv[])
{
  parse_args(argc,argv);
  generate_header(argc,argv);
  generate_code();
  exit(0);
}

