/*
** mm_timer.c:
**      RPRF interface for timing PHiPAC generated code.
**
** Written by:
**      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
**
**
** "Copyright (c) 1997 The Regents of the University of California.  All
** rights reserved."  Permission to use, copy, modify, and distribute
** this software and its documentation for any purpose, without fee, and
** without written agreement is hereby granted, provided that the above
** copyright notice and the following two paragraphs appear in all copies
** of this software.
** 
** IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
** FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
** THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
** 
** THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
** INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
** PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
** CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
** ENHANCEMENTS, OR MODIFICATIONS.
** 
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

/*
 * Load RPRF declarations
 */

#include "rprf.h"

/*
 * Set MM_FP_TYPE to the correct data type according to
 * the precision assigned.
 */

#if defined DOUBLE_PREC
#define MM_FP_TYPE double
#elif defined SINGLE_PREC
#define MM_FP_TYPE float
#else
#undef MM_FP_TYPE		/* make sure this code doesn't compile!!! */
#endif

#if defined ALPHA_EQUALS_ONE

extern void
mm_mult(int,int,int,
	MM_FP_TYPE*,MM_FP_TYPE*,MM_FP_TYPE*,
	int,int,int);
#define FUNC mm_mult(M,K,N,A,B,cC,colA,colB,colC);

#elif defined ALPHA_ARBITRARY

extern void
mm_mult(int,int,int,
	MM_FP_TYPE*,MM_FP_TYPE*,MM_FP_TYPE*,
	int,int,int,
	MM_FP_TYPE);
#define FUNC mm_mult(M,K,N,A,B,cC,colA,colB,colC,alpha);

#else

/* make sure this code doesn't compile!!! */
#undef MM_FP_TYPE
#undef FUNC
#endif /* ALPHA */



/* Matrix initialization routines */

/* double precision */
void copy_d_vd(size_t len, double from, double *to)
{
  size_t i;
  
  for (i=0; i<len; i++)
    *to++ = from;
}

/* single precision */
void copy_f_vf(size_t len, float from, float *to)
{
  size_t i;
  
  for (i=0; i<len; i++)
    *to++ = from;
}



/* Main program */

int
main(int argc, char *argv[])
{
    MM_FP_TYPE *A, *B, *cC, alpha;

    size_t M, K, N;		/* dimension of submatrices to multiply*/
    size_t rowA, colA;		/* actual storage sizes of the matrices */
    size_t rowB, colB;
    size_t rowC, colC;

    if (argc < 9) {
      printf("Usage: %s [<perf_args>] M K N rowA colA rowB colB rowC colC\n",
	      argv[0]);
      printf("\tbenchmarks the MxK, KxN, MxN sub-matrix multiplication in");
      printf("\t\tA * B -> C,\twhere\n");
      printf("\t  A is an rowA x colA matrix,\n");
      printf("\t  B is an rowB x colB matrix,\n");
      printf("\t  C is an rowC x colC matrix.\n");

      exit(0);
    }
  
    /* Parse the arguments */
    rprf_parse_args(argc, argv, "iiiiiiiii", &M, &K, &N, 
		    &rowA, &colA, &rowB, &colB, &rowC, &colC);

#if defined DOUBLE_PREC
    /* double precision */
    
    /* Allocate and initialize. */
    
    A = rprf_malloc_vd(rowA*colA);
    B = rprf_malloc_vd(rowB*colB);
    cC = rprf_malloc_vd(rowC*colC);

    copy_d_vd(M*K, 1.0, A);
    copy_d_vd(K*N, 1.0, B);

    alpha = 1.0;

#elif defined SINGLE_PREC
    /* single precision */    

    /* Allocate and initialize. */
    
    A = rprf_malloc_vf(rowA*colA);
    B = rprf_malloc_vf(rowB*colB);
    cC = rprf_malloc_vf(rowC*colC);

    copy_f_vf(M*K, 1.0f, A);
    copy_f_vf(K*N, 1.0f, B);

    alpha = 1.0f;

#endif
    

/* Run the performance test */

    RPRF_BODY(FUNC, M*N*(2*K-1))

    return 0;
}
