#! /usr/local/bin/perl

##
## SEARCH_L0CORE.PL
##
## Perl script to search for the optimal L0 core blocking sizes 
## (register blocking) for matrix-matrix multiplication code
## generated by MM_CGEN.
##
## Written by:
##      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
##
##
## "Copyright (c) 1997 The Regents of the University of California.  All
## rights reserved."  Permission to use, copy, modify, and distribute
## this software and its documentation for any purpose, without fee, and
## without written agreement is hereby granted, provided that the above
## copyright notice and the following two paragraphs appear in all copies
## of this software.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
## THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
## SUCH DAMAGE.
## 
## THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
## PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
## CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
## ENHANCEMENTS, OR MODIFICATIONS.
## 


##
## Initialization
##

## Add the directory of search to @INC
{
   local(@temp);

   @temp = split(/\//,$0);
   pop(@temp);
   push(@INC, join("/",@temp));
}

## Initialize test configuration parameters.
require "search_init.pl";
$search_mode = "L0CORE";
$useage = "$0 <machine specification file> <single|double> <compiler_options file> <L0 search parameters file>";

## Read search parameters.
require "search_readparams.pl";
&read_machine_specs($ARGV[0]);
$prec = $ARGV[1];
$compiler_options_file = $ARGV[2];
&read_l0_search_params($ARGV[3]);

if ($prec eq "single")
{
    if ($benchmark_l0_out_of_l3) {
	# number of matrix elements fits in L3 cache
	$L1maxelmts = int ($L3_cache_size / 4);
    } elsif  ($benchmark_l0_out_of_l2) {
	# number of matrix elements fits in L2 cache
	$L1maxelmts = int ($L2_cache_size / 4);
    } else {
	# number of matrix elements fits in L1 cache
	$L1maxelmts = int ($L1_cache_size / 4);
    }
   $mn_max = $num_single_prec_registers; # maximum bound for m0*n0 to be searched
}
elsif ($prec eq "double")
{
    if ($benchmark_l0_out_of_l3) {
	# number of matrix elements fits in L3 cache
	$L1maxelmts = int ($L3_cache_size / 8);
    } elsif  ($benchmark_l0_out_of_l2) {
	# number of matrix elements fits in L2 cache
	$L1maxelmts = int ($L2_cache_size / 8);
    } else {
	# number of matrix elements fits in L1 cache
	$L1maxelmts = int ($L1_cache_size / 8);
    }
   $mn_max = $num_double_prec_registers; # maximum bound for m*n to be searched
}
else
{
   die "***Precision must be either single or double.\n";
}



## Load supporting routines.
require "search_supp.pl";

## Skip this search altogether if the top performance file
## already exists.
if (-e "$perf_dir/$search_mode" . "_top.$prec")
{
   warn "***Nothing to be done for $search_mode search.\n";
   exit(0);
}


chop($hostname = `hostname`);
chop($date = `date`);
print STDOUT "\n\nL0CORE Search started on $hostname on $date\n\n";
&flush(STDOUT);


## Change to temporary directory
chdir($tmp_dir);


##
## Create the L0CORE options space.
##

$numpbms = 0;			# total number of test cases to perform
$numdone = 0;			# number of test cases already performed

$numopts = &create_L0CORE_options_space; # this is the number of points in
				     # the whole option space

##
## Read in the matrix sizes to be used for benchmarking
## if an external list is given
##

if ($auto_matrixsizes == 0)
{
   &read_matrixsizes_file;
}


##
## Loop through the L0CORE options space,
## and conduct benchmarking procedure for each option.
##

CCOPT:
foreach $ccopt (@ccopts)
{
   ($ccopttag, $cflags) = split(/\:/, $ccopt, 2);

   ALPHATYPE:
   foreach $alphatype (@alphatypes)
   {
      # compile timer interface (main() code in C)
      $status = &compile_timer;
      next ALPHATYPE if ($status != 0);
      
      SOFTPIPE:
      foreach $softpipeopt (@softpipeopts)
      {
	 MATOP:
	 foreach $matop (@matops)
	 {
	    print STDOUT "Testing $prec.$matop.$alphatype.$softpipeopt.$ccopttag ($cflags)...\n"
	       if $verbosity > 0;

	    # initialize current search statistics (best-so-far info)
	    &init_best_stats;

	    # create the blockings space of all blocking sizes;
	    # this is the list of blocking sizes that HAS NOT
	    # been benchmarked
	    @keys_blockings_space = &create_blockings_space;

	    # randomize the list of blocking sizes
	    {
	       local(@temp);

	       @temp = ();
	       push(@temp,
		    splice(@keys_blockings_space,
			   rand(@keys_blockings_space), 1))
		  while @keys_blockings_space;

	       @keys_blockings_space = @temp;
	    }

	    $numpbms = scalar(@keys_blockings_space) * $numopts;

	    if (-e "$perf_dir/$search_mode.$prec.$matop.$alphatype.$softpipeopt.$ccopttag")
	    {
	       # performance file already exist,
	       # so skip this case altogether

	       warn "***Nothing to be done for $prec.$matop.$alphatype.$softpipeopt.$ccopttag.\n";

	       $numdone += scalar(@keys_blockings_space);
	       next MATOP;
	    }

	    # open and initialize output file for this option
	    open(PERF_FILE,">$perf_dir/$perf_file")
	       || die "***Can't open $perf_dir/$perf_file.\n";
	    printf(PERF_FILE &file_header_str);

	    if ($auto_matrixsizes)
	    {
	       open(PERF_FILE_ALT,">$perf_dir/$perf_file.alt")
		  || die "***Can't open $perf_dir/$perf_file.alt.\n";
	       printf(PERF_FILE_ALT &file_header_str);
	    }

	    BLOCKING:
	    while (scalar(@keys_blockings_space) > 0)
	    {
	       # select a current set of blocking sizes
	       $key = shift(@keys_blockings_space);

	       ($m0,$k0,$n0) = split(/\,/,$key);
	       print STDOUT "\nTesting m0 = $m0, k0 = $k0, n0 = $n0...\n"
		  if $verbosity > 2;

	       # generate source code with this blocking
	       $status = &generate_core("-file ./src/mm_mult.c -routine_name mm_mult",
					$prec, $matop, $alphatype,
					"nofringe",$softpipeopt,
					$m0,$k0,$n0);

	       # compile generated source code
	       $status = &compile_code("mm_mult",$cflags)
		  if ($status == 0);

	       # link with timing routines to produce
	       # an executable timer program
	       $status = &link_timer
		  if ($status == 0);

	       # execute the timer program to benchmark
	       # the current blocking sizes
	       if ($status == 0)
	       {
		  $mflops_rate = &run_timing_test;
	       }
	       else
	       {
		  $mflops_rate = -1;
	       }

	       # delete the timer program to prevent
	       # subsequent failed cases from running it
	       # accidentally
	       &delete_timer;

	       # do housekeeping work
	       $numdone++;
	       &update_best_stats;
	       &print_curr_stats;

	    } # end loop for BLOCKING
	    
	    # close output file for this option
	    printf(PERF_FILE &file_trailer_str);
	    close(PERF_FILE);
	    printf(PERF_FILE_ALT &file_trailer_str);
	    close(PERF_FILE_ALT);

	    # rename file to indicate completion of this case
	    rename("$perf_dir/$perf_file",
		   "$perf_dir/$search_mode.$prec.$matop.$alphatype.$softpipeopt.$ccopttag");
	    if ($use_timing eq "fatdot")
	    {
	       rename("$perf_dir/$perf_file.alt",
		      "$perf_dir/$search_mode" . "_almstsq.$prec.$matop.$alphatype.$softpipeopt.$ccopttag");
	    }
	    elsif ($use_timing eq "almstsq")
	    {
	       rename("$perf_dir/$perf_file.alt",
		      "$perf_dir/$search_mode" . "_fatdot.$prec.$matop.$alphatype.$softpipeopt.$ccopttag");
	    }

	 } # end loop for MATOP
      } # end loop for SOFTPIPE
   } # end loop for ALPHATYPE
} # end loop for CCOPT


##
## Termination
##

## Change back to working directory
chdir($working_dir);

chop($date = `date`);
print STDOUT "\n\nL0CORE Search ended on $hostname at $date\n\n";

exit(0);

##
## END
##




##
## sub create_L0CORE_options_space
##
## Generate all points in the options space.  The options space
## is a subset of the following direct product:
##
## {compilers options} X {software-pipelining options} X
##      {alpha type options} X {matrix options}
##
## {compilers options} is given by the lines in the file
## $compiler_options_file.
##
## {software-pipelining options} is the (sub)set of software-pipelining
## options specified by $one_stage_holdstripe_A, $one_stage_holdstripe_B,
## $two_stage_load_mult, $two_stage_mult_add, $three_stage.
##
## {alpha type options} is the (sub)set of alpha types specified by
## $alpha_equals_one, $alpha_arbitrary.
##
## {matrix options} is the (sub)set of normal/transposed matrix
## operations specified by $matop_NN, $matop_NT, $matop_TN, $matop_TT.
##
## The returned value of this subroutine is the total number of
## points in the whole options space.
##
sub create_L0CORE_options_space
{
   ##
   ## create array @ccopts
   ##

   @ccopts = ();

   open(CCOPTFILE, "$working_dir/$compiler_options_file")
      || die "***Can't open $working_dir/$compiler_options_file\n";

   $line = 0;
   while(<CCOPTFILE>)
   {
      s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

      if (/^%/)
      {
	 # comments --- ignore
      }
      elsif (/^$/)
      {
	 # blank line --- ignore
      }
      elsif (/^(\w+)\s*:\s*(.*)$/)
      {
	 $line++;
	 push(@ccopts, join(':', $1, $2));
      }
      elsif (/^\s*:\s*(.*)$/)
      {
	 $line++;
	 push(@ccopts, join(':', $numpbms_ccopt_mult, $1));
      }
      else	      
      {
	 $line++;
	 push(@ccopts, join(':', $numpbms_ccopt_mult, $_));
      }
   }

   close(CCOPTFILE);

   ##
   ## create array @softpipeopts
   ##

   @softpipeopts = ();

   if ($one_stage_holdstripe_A)
   {
      push(@softpipeopts, ('1A'));
   }
   if ($one_stage_holdstripe_B)
   {
      push(@softpipeopts, ('1B'));
   }
   if ($two_stage_load_mult)
   {
      push(@softpipeopts, ('2lm'));
   }
   if ($two_stage_mult_add)
   {
      push(@softpipeopts, ('2ma'));
   }
   if ($three_stage)
   {
      push(@softpipeopts, ('3'));
   }

   ##
   ## create array @alphatypes
   ##
   
   @alphatypes = ();

   if ($alpha_equals_one)
   {
      push(@alphatypes, ('1'));
   }
   if ($alpha_arbitrary)
   {
      push(@alphatypes, ('c'));
   }

   ##
   ## create array @matops
   ##

   @matops = ();

   if ($matop_NN)
   {
      push(@matops, ('NN'));
   }
   if ($matop_NT)
   {
      push(@matops, ('NT'));
   }
   if ($matop_TN)
   {
      push(@matops, ('TN'));
   }
   if ($matop_TT)
   {
      push(@matops, ('TT'));
   }

   ##
   ## return number of points in options space
   ##

   return(scalar(@ccopts) * scalar(@softpipeopts)
	  * scalar(@alphatypes) * scalar(@matops));
}



##
## sub init_best_stats
##
## Initializes the statistics for the best blocking sizes found
## so far.
##
sub init_best_stats
{
   $best_mflops_rate = 0;	# the maximum mflops_rate so far
   $best_blocking = "";		# and the corresponding blocking sizes
}


##
## sub update_best_stats
##
## Updates the statistics for the best blocking sizes found so
## far.
##
sub update_best_stats
{
   if ($mflops_rate >= $best_mflops_rate)
   {
      $best_mflops_rate = $mflops_rate;
      $best_blocking = "($m0,$k0,$n0)";
   }
}


##
## sub print_curr_stats
##
## Prints the current search statistics to STDOUT
##
sub print_curr_stats
{
   if ($verbosity > 1)
   {
      printf(STDOUT "Performance for $prec.$matop.$alphatype.$softpipeopt.$ccopttag ($m0,$k0,$n0) : %7.1f MFLOPS.\n",
	     $mflops_rate);

      printf(STDOUT "Best Blocking for $prec.$matop.$alphatype.$softpipeopt.$ccopttag so far : $best_blocking : %7.1f MFLOPS.\n",
	     $best_mflops_rate);

      printf(STDOUT "Done %d (%3.1f%%), %d (%3.1f%%) left; ",
	     $numdone, $numdone * 100 / $numpbms,
	     $numpbms - $numdone, ($numpbms - $numdone) * 100 / $numpbms);
      printf(STDOUT "Finishing by %s\n",
	     &ctime($^T + ($numpbms / $numdone) * (time() - $^T)));
      &flush(STDOUT);
   }
}


##
## sub file_header_str
##
## Returns a file header string to be printed at the beginning of
## the output file, documenting the parameters with which this
## search was conducted.
##
sub file_header_str
{
   local($output_str);
   local($opA, $opB);

   $opA = substr($matop,0,1);
   $opB = substr($matop,1,1);

   $output_str = "";
   
   $output_str .= "%%%% Performance from L0 core search:\n";
   $output_str .= "%%%%\n";

   $output_str .= "%%%% (search started on $hostname on $date)\n";
   $output_str .= "%%%% File opened on ";
   $output_str .= `date`;
   $output_str .= "%%%%\n";

   $output_str .= "%%%% Precision:\t\t\t";
   $output_str .= "$prec\n";

   $output_str .= "%%%% Matrix Operations:\t\t";

   if ($opA eq "N")
   {
      $output_str .= "Normal-";
   }
   elsif ($opA eq "T")
   {
      $output_str .= "Transpose-";
   }

   if  ($opB eq "N")
   {
      $output_str .= "Normal\n";
   }
   elsif ($opB eq "T")
   {
      $output_str .= "Transpose\n";
   }

   $output_str .= "%%%% Alpha:\t\t\t";
   
   if ($alphatype eq "1")
   {
      $output_str .= "equals one (constant)\n";
   }
   elsif ($alphatype eq "c")
   {
      $output_str .= "arbitrary\n";
   }

   $output_str .= "%%%% Software Pipelining:\t\t";
   $output_str .= "$softpipeopt\n";
   
   $output_str .= "%%%% Compiler-Options Tag:\t$ccopttag\n";

   $output_str .= "%%%%\n";

   ##
   ## machine information
   ##

   $output_str .= "%%%% Number of single precision registers: $num_single_prec_registers\n";
   $output_str .= "%%%% Number of double precision registers: $num_double_prec_registers\n";
   $output_str .= "%%%%\n";
   $output_str .= "%%%% L1 cache size: $L1_cache_size bytes\n";
   if ($L2_cache_size > 0)
   {
      $output_str .= "%%%% L2 cache size: $L2_cache_size bytes\n";

      if ($L3_cache_size > 0)
      {
	 $output_str .= "%%%% L3 cache size: $L3_cache_size bytes\n";
      }
   }
   $output_str .= "%%%%\n";
   $output_str .= "%%%% Compiler: $cc\n";
   $output_str .= "%%%% Generator options: $generator_opts\n";
   $output_str .= "%%%% Timer arguments: $timer_args\n";
   $output_str .= "%%%%\n\n";

   ##
   ## environment information
   ##

   $output_str .= "# cflags = $cflags\n";
   $output_str .= "\n";

   $output_str .= "%%%%   Mflops   \t m0  k0  n0\n";
   $output_str .= "\n";

   $output_str;			# return value
}


##
## sub file_trailer_str
##
## Returns a file trailer string to be printed at the end of
## the output file.
##
sub file_trailer_str
{
   local($output_str);

   $output_str  = "\n\n";
   $output_str .= "%%%%\n";

   $output_str .= "%%%% File closed on ";
   $output_str .= `date`;
   $output_str .= "%%%%\n";

   $output_str;			# return value
}



##
## sub read_matrixsizes_file
##
## Read in the matrix sizes to be used for benchmarking from the file
## given in $matrixsizes_file.
##
## The performance measure on a set of blocking sizes is specified in
## one of two ways.  If $auto_matrixsizes == 1, the performance of the
## set of blocking sizes is just the MFLOPS rate achieved by the code
## on this single set of matrix sizes.  This mode of performance
## measure is suitable for generating high performance code for large
## matrices.  If $auto_blockings == 0, the performance of the set of
## blocking sizes is the weighted mean MFLOPS rate achieved by the
## code on all sets of matrix sizes listed in the file
## $matrixsizes_file.  This mode of performance measure is suitable
## for generating high performance code for a specific set(s) of
## predetermined matrix sizes (which are typically small).
##
## Here, we are assuming that $auto_matrixsizes == 0, and so we read
## in the given $matrixsize_file.
##
sub read_matrixsizes_file
{
   # create array @matrix_sizes

   @matrix_sizes = ();

   open(MATRIXSIZES_FILE, "$working_dir/$matrixsizes_file")
      || die "***Can't open $working_dir/$matrixsizes_file.\n";

   while(<MATRIXSIZES_FILE>)
   {
      s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

      if (/^%/)
      {
	 # comments --- ignore
      }
      elsif (/^$/)
      {
	 # blank line --- ignore
      }
      else
      {
	 push(@matrix_sizes, join(',', split(/\D+/)));
      }
   }

   close(MATRIXSIZES_FILE);
}



##
## sub create_blockings_space
##
## Generate all points in the blockings space.  The blockings
## space is the set of all triplets (m0,k0,n0) that will be
## used as the register blocking sizes for the generated
## code.
##
## The blockings space is specified in one of two ways.  If
## $auto_blockings == 1, the blocking sizes are specified
## implicitly by $num_single_prec_registers,
## $num_double_prec_registers, @k0_set
## so that the triplets (m0,n0,k0) to be tested are exactly
## those with m0*n0 lying between 1 and $num_regs,
## and with k0 selected from @k0_set ($num_regs =
## $num_single_prec_registers for single precision, and
## $num_double_prec_registers for double precision).
## If $auto_blockings == 0, the blocking sizes are read from
## the file $blockings_file.
##
sub create_blockings_space
{
   # create array %blockings_space

   local(%blocking_space);

   %blockings_space = ();

   if ($auto_blockings)
   {
      # automatically generate blocking sizes to be tested

      local($m0,$k0,$n0);

      foreach $k0 (@k0_set)
      {
	 for ($m0 = 1; $m0 <= $mn_max; $m0++)
	 {
	    for ($n0 = 1; $m0*$n0 <= $mn_max; $n0++)
	    {
	       $blockings_space{"$m0,$k0,$n0"} = 0;
	    }
	 }
      }
   }
   else
   {
      # read in blocking sizes to be tested
      # from the file $blockings_file

      open(BLOCKINGS_FILE, "$working_dir/$blockings_file")
	 || die "***Can't open $working_dir/$blockings_file.\n";

      while(<BLOCKINGS_FILE>)
      {
	 s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

	 if (/^%/)
	 {
	    # comments --- ignore
	 }
	 elsif (/^$/)
	 {
	    # blank line --- ignore
	 }
	 else
	 {
	    $blockings_space{join(',', split(/\D+/))} = 0;
	 }
      }

      close(BLOCKINGS_FILE);
   }

   keys(%blockings_space);
}


##
## sub run_timing_test
##
## Executes the timer program to perform timing tests on the
## generated routine.  Returns the mflops-rating of the routine.
##
sub run_timing_test
{
   local($M, $K, $N);		# these are the matrix sizes

   local($mflops_rate, $mflops_rate_alt);
   local($mflops_rate_fatdot);
   local($mflops_rate_almstsq);

   $mflops_rate = 0;
   $mflops_rate_alt = 0;

   if ($auto_matrixsizes)
   {
      # compute the l1 multiples for the matrix to use for timing
      $M = $m0;
      $K = ($L1maxelmts * ($fill_cache_percentage/100) - $m0 * $n0) /
	       ($m0 + $n0);
      $K = int($K/$k0+0.5)*$k0;
      $N = $n0;
      $spadd = substr($softpipeopt,0,1) - 1; # additional $K for software pipelining
      $K += $spadd;

      $mflops_rate_fatdot = &do_test;

      $sqsize = sqrt($L1maxelmts * ($fill_cache_percentage/100) / 3);

      $M = $m0 * int($sqsize / $m0);
      $K = $k0 * int($sqsize / $k0) + $spadd;
      $N = $n0 * int($sqsize / $n0);

      $mflops_rate_almstsq = &do_test;

      if ($use_timing eq "fatdot")
      {
	 $mflops_rate = $mflops_rate_fatdot;
	 $mflops_rate_alt = $mflops_rate_almstsq;
      }
      elsif ($use_timing eq "almstsq")
      {
	 $mflops_rate = $mflops_rate_almstsq;
	 $mflops_rate_alt = $mflops_rate_fatdot;
      }	 
   }
   else
   {
      local($sum_inv_rate);     # the culmulative sum of the reciprocals
				# of mflops_rates;
				# used to compute the harmonic mean of the
				# mflops_rates as the return value when
				# benchmarking over a list of matrix-sizes
      $sum_inv_rate = 0;

      foreach (@matrix_sizes)
      {
	 local($rate);

	 ($M,$K,$N) = split(/\,/);
	 $rate = &do_test;
	 if ($rate > 0)
	 {
	    $sum_inv_rate += 1 / $rate;
	 }
      }

      if ($sum_inv_rate > 0)
      {
	 $mflops_rate = scalar(@matrix_sizes) / $sum_inv_rate;
      }
      else
      {
	 $mflops_rate = 0;
      }
   }

   return(-1) if ($NO_ACTION);

   printf(PERF_FILE "%12.6f\t%3d %3d %3d\n",
	  $mflops_rate,
	  $m0,$k0,$n0);
   &flush(PERF_FILE);

   if ($auto_matrixsizes)
   {
      printf(PERF_FILE_ALT "%12.6f\t%3d %3d %3d\n",
	     $mflops_rate_alt,
	     $m0,$k0,$n0);
      &flush(PERF_FILE_ALT);
   }

   $mflops_rate;		# return value
}


