#! /usr/local/bin/perl

##
## SEARCH_L0GEN.PL
##
## Perl script to search for the optimal L0 general blocking sizes 
## (register blocking) for matrix-matrix multiplication code
## generated by MM_CGEN.
##
## Written by:
##      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
##
##
## "Copyright (c) 1997 The Regents of the University of California.  All
## rights reserved."  Permission to use, copy, modify, and distribute
## this software and its documentation for any purpose, without fee, and
## without written agreement is hereby granted, provided that the above
## copyright notice and the following two paragraphs appear in all copies
## of this software.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
## THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
## SUCH DAMAGE.
## 
## THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
## PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
## CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
## ENHANCEMENTS, OR MODIFICATIONS.
## 


##
## Initialization
##

## Add the directory of search to @INC
{
   local(@temp);

   @temp = split(/\//,$0);
   pop(@temp);
   push(@INC, join("/",@temp));
}

## Initialize test configuration parameters.
require "search_init.pl";
$search_mode = "L0GEN";
$useage = "$0 <machine specification file> <L1 top performance file> <L0 search parameters file>";

## Read search parameters.
require "search_readparams.pl";
&read_machine_specs($ARGV[0]);
$l1_top_file = $ARGV[1];
&read_l0_search_params($ARGV[2]);

## Load supporting routines.
require "search_supp.pl";

## Skip this search altogether if the top performance file
## already exists.
if (-e "$perf_dir/$search_mode" . "_top.$prec")
{
   warn "***Nothing to be done for $search_mode search.\n";
   exit(0);
}


chop($hostname = `hostname`);
chop($date = `date`);
print STDOUT "\n\nL0GEN Search started on $hostname on $date\n\n";
&flush(STDOUT);

## Change to temporary directory
chdir($tmp_dir);


##
## Create the L0GEN options space.
##

$numpbms = 0;			# total number of test cases to perform
$numdone = 0;			# number of test cases already performed

$numopts = &create_L0GEN_options_space; # this is the number of points in
				     # the whole option space

##
## Read in the matrix sizes to be used for benchmarking
## if an external list is given
##

if ($auto_matrixsizes == 0)
{
   &read_matrixsizes_file;
}


##
## Loop through the L0GEN options space,
## and conduct benchmarking procedure for each option.
##

CASE:
foreach $case (keys(%cases))
{
   ($prec, $matop, $alphatype) = split(/\./, $case);

   if ($prec eq "single")
   {
      $L1maxelmts = int ($L1_cache_size / 4); # number of matrix elements fits in L1 cache
      $mn_max = $num_single_prec_registers; # maximum bound for m0*n0 to be searched
   }
   elsif ($prec eq "double")
   {
      $L1maxelmts = int ($L1_cache_size / 8); # number of matrix elements fits in L1 cache
      $mn_max = $num_double_prec_registers; # maximum bound for m*n to be searched
   }
   else
   {
      $L1maxelmts = -1;	# something is wrong
   }
      
   # select the L1 code options for this case
   ($l1, $l0core, $cflags) = split(/\;/, $cases{$case});
   ($m1,$k1,$n1,$l1codeuse) = split(/\,/, $l1);
   ($m0c,$k0c,$n0c,$l0spopt) = split(/\,/, $l0core);

   # compile timer interface (main() code in C)
   $status = &compile_timer;
   next CASE if ($status != 0);
	 
   SOFTPIPE:
   foreach $softpipeopt (@softpipeopts)
   {
      print STDOUT "Testing $prec.$matop.$alphatype.$softpipeopt ($cflags)...\n"
	 if $verbosity > 0;

      # initialize current search status (best-so-far info)
      &init_best_stats;

      # create the blockings space of all blocking sizes;
      # this is the list of blocking sizes that HAS NOT
      # been benchmarked
      @keys_blockings_space = &create_blockings_space;

      # randomize the list of blocking sizes
      {
	 local(@temp);
	 
	 @temp = ();
	 push(@temp,
	      splice(@keys_blockings_space,
		     rand(@keys_blockings_space), 1))
	    while @keys_blockings_space;
	 
	 @keys_blockings_space = @temp;
      }

      $numpbms = scalar(@keys_blockings_space) * $numopts;

      if (-e "$perf_dir/$search_mode.$prec.$matop.$alphatype.$softpipeopt")
      {
	 # performance file already exist,
	 # so skip this case altogether

	 warn "***Nothing to be done for $prec.$matop.$alphatype.$softpipeopt.\n";

	 $numdone += scalar(@keys_blockings_space);
	 next SOFTPIPE;
      }

      # open and initialize output file for this option
      open(PERF_FILE,">$perf_dir/$perf_file")
	 || die "***Can't open $perf_dir/$perf_file.\n";
      printf(PERF_FILE &file_header_str);

      BLOCKING:
      while (scalar(@keys_blockings_space) > 0)
      {
	 # select a current set of blocking sizes
	 $key = shift(@keys_blockings_space);

	 ($m0,$k0,$n0) = split(/\,/,$key);
	 print STDOUT "\nTesting m0 = $m0, k0 = $k0, n0 = $n0...\n"
	    if $verbosity > 2;
	 
	 # generate source code with this blocking
	 $status = &generate_core("-file ./src/mm_mult.c -routine_name mm_mult",
				  $prec, $matop, $alphatype,
				  "general", $softpipeopt,
				  $m0,$k0,$n0);
	 
	 # compile generated source code
	 $status = &compile_code("mm_mult",$cflags)
	    if ($status == 0);

	 # link with timing routines to produce
	 # an executable timer program
	 $status = &link_timer
	    if ($status == 0);

	 # execute the timer program to benchmark
	 # the current blocking sizes
	 if ($status == 0)
	 {
	    $mflops_rate = &run_timing_test;
	 }
	 else
	 {
	    $mflops_rate = -1;
	 }

	 # delete the timer program to prevent
	 # subsequent failed cases from running it
	 # accidentally
	 &delete_timer;

	 # do housekeeping work
	 $numdone++;
	 &update_best_stats;
	 &print_curr_stats;

      } # end loop for BLOCKING
	       
      # close output file for this option
      printf(PERF_FILE &file_trailer_str);
      close(PERF_FILE);

      # rename file to indicate completion of this case
      rename("$perf_dir/$perf_file",
	     "$perf_dir/$search_mode.$prec.$matop.$alphatype.$softpipeopt");

   } # end loop for SOFTPIPE
} # end loop for CASE


##
## Termination
##

## Change back to working directory
chdir($working_dir);

chop($date = `date`);
print STDOUT "\n\nL0GEN Search ended on $hostname at $date\n\n";

exit(0);

##
## END
##




##
## sub create_L0GEN_options_space
##
## Generate all points in the options space.  The options space
## is a subset of the following direct product:
##
## {cases} X {software-pipelining options}
##
## {software-pipelining options} is the (sub)set of software-pipelining
## options specified by $one_stage_holdstripe_A, $one_stage_holdstripe_B,
## $two_stage_load_mult, $two_stage_mult_add, $three_stage.
##
## {cases} is a (sub)set of combinations of precision, matrix operation
## and alpha type which has some performance recorded in the L1
## top performance file.
##
## The returned value of this subroutine is the total number of
## points in the whole options space.
##
sub create_L0GEN_options_space
{
   ##
   ## create array %cases
   ##

   %cases = ();

   open(L1TOP_FILE, "$working_dir/$l1_top_file")
      || die "***Can't open $working_dir/$l1_top_file.\n";

   while(<L1TOP_FILE>)
   {
      s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

      if (/^%/)
      {
	 # comments --- ignore
      }
      elsif (/^$/)
      {
	 # blank line --- ignore
      }
      else
      {
	 /^(\w+)\.(\w+)\.(\w+):\s+([0-9.]+)/;
	 # the fields are:
	 #   (1) precision, (2) matrix operation, (3) alpha type,
	 #   (4) performance

	 $key = "$1.$2.$3";

	 local($l1, $l0core, $cflags);

	 $_ = <L1TOP_FILE>;
	 while (substr($_,0,1) eq "#")
	 {
	    $_ =~ /^\# (.+) = (.*)$/;

	    if ($1 eq "L1")
	    {
	       $l1 = $2;
	    }
	    elsif ($1 eq "L0CORE")
	    {
	       $l0core = $2;
	    }
	    elsif ($1 eq "cflags")
	    {
	       $cflags = $2;
	    }
	    else
	    {
	       warn "***Ignoring information $_.\n";
	    }

	    $_ = <L1TOP_FILE>;
	 }

	 $cases{$key} = "$l1;$l0core;$cflags";
      }
   }

   close(L1TOP_FILE);

   ##
   ## create array @softpipeopts
   ##

   @softpipeopts = ();

   if ($one_stage_holdstripe_A)
   {
      push(@softpipeopts, ('1A'));
   }
   if ($one_stage_holdstripe_B)
   {
      push(@softpipeopts, ('1B'));
   }
   if ($two_stage_load_mult)
   {
      push(@softpipeopts, ('2lm'));
   }
   if ($two_stage_mult_add)
   {
      push(@softpipeopts, ('2ma'));
   }
   if ($three_stage)
   {
      push(@softpipeopts, ('3'));
   }

   ##
   ## return number of points in options space
   ##

   return(scalar(keys(%cases)) * scalar(@softpipeopts));
}



##
## sub init_best_stats
##
## Initializes the statistics for the best blocking sizes found
## so far.
##
sub init_best_stats
{
   $best_mflops_rate = 0;	# the maximum mflops_rate so far
   $best_blocking = "";		# and the corresponding blocking sizes
}


##
## sub update_best_stats
##
## Updates the statistics for the best blocking sizes found so
## far.
##
sub update_best_stats
{
   if ($mflops_rate >= $best_mflops_rate)
   {
      $best_mflops_rate = $mflops_rate;
      $best_blocking = "($m0,$k0,$n0)";
   }
}


##
## sub print_curr_stats
##
## Prints the current search statistics to STDOUT
##
sub print_curr_stats
{
   if ($verbosity > 1)
   {
      printf(STDOUT "Performance for $prec.$matop.$alphatype.$softpipeopt.$ccopttag ($m0,$k0,$n0) : %7.1f MFLOPS.\n",
	     $mflops_rate);

      printf(STDOUT "Best Blocking for $prec.$matop.$alphatype.$softpipeopt.$ccopttag so far : $best_blocking : %7.1f MFLOPS.\n",
	     $best_mflops_rate);

      printf(STDOUT "Done %d (%3.1f%%), %d (%3.1f%%) left; ",
	     $numdone, $numdone * 100 / $numpbms,
	     $numpbms - $numdone, ($numpbms - $numdone) * 100 / $numpbms);
      printf(STDOUT "Finishing by %s\n",
	     &ctime($^T + ($numpbms / $numdone) * (time() - $^T)));
      &flush(STDOUT);
   }
}


##
## sub file_header_str
##
## Returns a file header string to be printed at the beginning of
## the output file, documenting the parameters with which this
## search was conducted.
##
sub file_header_str
{
   local($output_str);
   local($opA, $opB);

   $opA = substr($matop,0,1);
   $opB = substr($matop,1,1);

   $output_str = "";
   
   $output_str .= "%%%% Performance from L0GEN search:\n";
   $output_str .= "%%%%\n";

   $output_str .= "%%%% (search started on $hostname at $date)\n";
   $output_str .= "%%%% File opened on ";
   $output_str .= `date`;
   $output_str .= "%%%%\n";

   $output_str .= "%%%% Precision:\t\t\t";
   $output_str .= "$prec\n";

   $output_str .= "%%%% Matrix Operations:\t\t";

   if ($opA eq "N")
   {
      $output_str .= "Normal-";
   }
   elsif ($opA eq "T")
   {
      $output_str .= "Transpose-";
   }

   if  ($opB eq "N")
   {
      $output_str .= "Normal\n";
   }
   elsif ($opB eq "T")
   {
      $output_str .= "Transpose\n";
   }

   $output_str .= "%%%% Alpha:\t\t\t";
   
   if ($alphatype eq "1")
   {
      $output_str .= "equals one (constant)\n";
   }
   elsif ($alphatype eq "c")
   {
      $output_str .= "arbitrary\n";
   }

   $output_str .= "%%%% Software Pipelining:\t\t";
   $output_str .= "$softpipeopt\n";
   
   $output_str .= "%%%%\n";

   ##
   ## machine information
   ##

   $output_str .= "%%%% Number of single precision registers: $num_single_prec_registers\n";
   $output_str .= "%%%% Number of double precision registers: $num_double_prec_registers\n";
   $output_str .= "%%%%\n";
   $output_str .= "%%%% L1 cache size: $L1_cache_size bytes\n";
   if ($L2_cache_size > 0)
   {
      $output_str .= "%%%% L2 cache size: $L2_cache_size bytes\n";

      if ($L3_cache_size > 0)
      {
	 $output_str .= "%%%% L3 cache size: $L3_cache_size bytes\n";
      }
   }
   $output_str .= "%%%%\n";
   $output_str .= "%%%% Compiler: $cc\n";
   $output_str .= "%%%% Generator options: $generator_opts\n";
   $output_str .= "%%%% Timer arguments: $timer_args\n";
   $output_str .= "%%%%\n\n";

   ##
   ## environment information
   ##

   $output_str .= "# L1 = $l1\n";
   $output_str .= "# L0CORE = $l0core\n";
   $output_str .= "# cflags = $cflags\n";
   $output_str .= "\n";

   $output_str .= "%%%%   Mflops   \t m0  k0  n0\n";
   $output_str .= "\n";

   $output_str;			# return value
}



##
## sub file_trailer_str
##
## Returns a file trailer string to be printed at the end of
## the output file.
##
sub file_trailer_str
{
   local($output_str);

   $output_str  = "\n\n";
   $output_str .= "%%%%\n";

   $output_str .= "%%%% File closed on ";
   $output_str .= `date`;
   $output_str .= "%%%%\n";

   $output_str;			# return value
}



##
## sub read_matrixsizes_file
##
## Read in the matrix sizes to be used for benchmarking from the file
## given in $matrixsizes_file.
##
## The performance measure on a set of blocking sizes is specified in
## one of two ways.  If $auto_matrixsizes == 1, the performance of the
## set of blocking sizes is just the MFLOPS rate achieved by the code
## on this single set of matrix sizes.  This mode of performance
## measure is suitable for generating high performance code for large
## matrices.  If $auto_blockings == 0, the performance of the set of
## blocking sizes is the weighted mean MFLOPS rate achieved by the
## code on all sets of matrix sizes listed in the file
## $matrixsizes_file.  This mode of performance measure is suitable
## for generating high performance code for a specific set(s) of
## predetermined matrix sizes (which are typically small).
##
## Here, we are assuming that $auto_matrixsizes == 0, and so we read
## in the given $matrixsize_file.
##
sub read_matrixsizes_file
{
   # create array @matrix_sizes

   @matrix_sizes = ();

   open(MATRIXSIZES_FILE, "$working_dir/$matrixsizes_file")
      || die "***Can't open $working_dir/$matrixsizes_file.\n";

   while(<MATRIXSIZES_FILE>)
   {
      s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

      if (/^%/)
      {
	 # comments --- ignore
      }
      elsif (/^$/)
      {
	 # blank line --- ignore
      }
      else
      {
	 push(@matrix_sizes, join(',', split(/\D+/)));
      }
   }

   close(MATRIXSIZES_FILE);
}



sub MIN
{
    local($min) = pop(@_);
    foreach $foo (@_) {
	$min = $foo if $min > $foo;
    }
    $min;
}


##
## sub create_blockings_space
##
## Generate all points in the blockings space.  The blockings
## space is the set of all triplets (m0,k0,n0) that will be
## used as the register blocking sizes for the generated
## code.
##
## The blockings space is specified in one of two ways.  If
## $auto_blockings == 1, the blocking sizes are specified
## implicitly by $num_single_prec_registers,
## $num_double_prec_registers, $k0_min and $k0_max,
## so that the triplets (m0,n0,k0) to be tested are exactly
## those with m0*n0 lying between $num_regs/8 and $num_regs,
## and with k0 between $k0_min and $k0_max ($num_regs =
## $num_single_prec_registers for single precision, and
## $num_double_prec_registers for double precision).
## If $auto_blockings == 0, the blocking sizes are read from
## the file $blockings_file.
##
## The return value is the total number of test cases
## generated.
##
sub create_blockings_space
{
   # create array %blockings_space

   local(%blocking_space);

   %blockings_space = ();

   if ($auto_blockings)
   {
      # automatically generate blocking sizes to be tested

      local($m0,$k0,$n0);

      foreach $k0 (@k0_set)
      {
	 for ($m0 = 1; $m0 <= $mn_max; $m0++)
	 {
	    for ($n0 = 1; $m0*$n0+&MIN($m0,$n0) < $mn_max; $n0++)
	    {
	       $blockings_space{"$m0,$k0,$n0"} = 0;
	    }
	 }
      }
   }
   else
   {
      # read in blocking sizes to be tested
      # from the file $blockings_file

      open(BLOCKINGS_FILE, "$working_dir/$blockings_file")
	 || die "***Can't open $working_dir/$blockings_file.\n";

      while(<BLOCKINGS_FILE>)
      {
	 s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

	 if (/^%/)
	 {
	    # comments --- ignore
	 }
	 elsif (/^$/)
	 {
	    # blank line --- ignore
	 }
	 else
	 {
	    $blockings_space{join(',', split(/\D+/))} = 0;
	 }
      }

      close(BLOCKINGS_FILE);
   }

   keys(%blockings_space);
}


##
## sub run_timing_test
##
## Executes the timer program to perform timing tests on the
## generated routine.  Returns the mflops-rating of the routine.
##
sub run_timing_test
{
   local($M, $K, $N);		# these are the matrix sizes

   local($mflops_rate);

   if ($auto_matrixsizes)
   {
      # compute the array of matrix sizes to be used for timing this case

      @matrix_sizes = ();

      $K = $k0c * $k1;
      $N = $n0c * $n1;
      for ($M = 1; $M < $m0c; $M *= 2)
      {
	 push(@matrix_sizes, "$M,$K,$N");
      }

      $M = $m0c * $m1;
      $N = $n0c * $n1;
      for ($K = 1; $K < $k0c; $K *= 2)
      {
	 push(@matrix_sizes, "$M,$K,$N");
      }

      $M = $m0c * $m1;
      $K = $k0c * $k1;
      for ($N = 1; $N < $n0c; $N *= 2)
      {
	 push(@matrix_sizes, "$M,$K,$N");
      }
   }

   local($sum_inv_rate);	# the culmulative sum of the reciprocals
				# of mflops_rates;
				# used to compute the harmonic mean of the
				# mflops_rates as the return value when
				# benchmarking over a list of matrix-sizes
   $sum_inv_rate = 0;

   foreach (@matrix_sizes)
   {
      local($rate);

      ($M,$K,$N) = split(/\,/);
      $rate = &do_test;
      if ($rate > 0)
      {
	 $sum_inv_rate += 1 / $rate;
      }
   }

   if ($sum_inv_rate > 0)
   {
      $mflops_rate = scalar(@matrix_sizes) / $sum_inv_rate;
   }
   else
   {
      $mflops_rate = 0;
   }
   
   return(-1) if ($NO_ACTION);

   printf(PERF_FILE "%12.6f\t%3d %3d %3d\n",
	  $mflops_rate,
	  $m0,$k0,$n0);
   &flush(PERF_FILE);

   $mflops_rate;		# return value
}


