#! /usr/local/bin/perl

##
## SEARCH_L1.PL
##
## Perl script to search for the optimal L1-blocking sizes 
## for matrix-matrix multiplication code generated by MM_LGEN.
##
## Written by:
##      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
##      Jeff Bilmes <bilmes@icsi.berkeley.edu>
##
## $Header: /db/bilmes/phipac/dist/PHiPAC-beta/PHiPAC/search-2.2/RCS/search_L1.pl,v 1.2 1998/01/23 23:08:56 bilmes Exp $
##
## "Copyright (c) 1997 The Regents of the University of California.  All
## rights reserved."  Permission to use, copy, modify, and distribute
## this software and its documentation for any purpose, without fee, and
## without written agreement is hereby granted, provided that the above
## copyright notice and the following two paragraphs appear in all copies
## of this software.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
## THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
## SUCH DAMAGE.
## 
## THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
## PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
## CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
## ENHANCEMENTS, OR MODIFICATIONS.
## 


##
## Initialization
##

## Add the directory of search to @INC
{
   local(@temp);

   @temp = split(/\//,$0);
   pop(@temp);
   push(@INC, join("/",@temp));
}

## Initialize test configuration parameters.
require "search_init.pl";
$search_mode = "L1";
$useage = "$0 <machine specification file> <L0 top performance file> <L1 search parameters file>";

## Read search parameters.
require "search_readparams.pl";
&read_machine_specs($ARGV[0]);
$l0_top_file = $ARGV[1];
&read_l1_search_params($ARGV[2]);

## Load supporting routines.
require "search_supp.pl";

## Skip this search altogether if the top performance file
## already exists.
if (-e "$perf_dir/$search_mode" . "_top.$prec")
{
   warn "***Nothing to be done for $search_mode search.\n";
   exit(0);
}

chop($hostname = `hostname`);
chop($date = `date`);
print STDOUT "\n\nL1 Search started on $hostname on $date\n\n";
&flush(STDOUT);


## Change to temporary directory
chdir($tmp_dir);


##
## Create the L1 options space.
##

$numpbms = 0;			# total number of test cases to perform
$numdone = 0;			# number of test cases already performed

$numopts = &create_L1_options_space; # this is the number of points in
				     # the whole option space

##
## Loop through the L1 options space,
## and conduct benchmarking procedure for each option.
##

CASE:
foreach $case (keys(%cases))
{
   ($prec, $matop, $alphatype) = split(/\./, $case);

   if ($prec eq "single")
   {
      $L1maxelmts = int ($L1_cache_size / 4); # number of matrix elements fits in L1 cache

      if ($benchmark_l1_out_of_l3) {
	  # number of matrix elements fits in L3 cache
	  $L2maxelmts = int ($L3_cache_size / 4);
      } else {
	  # number of matrix elements fits in L2 cache
	  $L2maxelmts = int ($L2_cache_size / 4);
      }
   }
   elsif ($prec eq "double")
   {
      $L1maxelmts = int ($L1_cache_size / 8); # number of matrix elements fits in L1 cache

      if ($benchmark_l1_out_of_l3) {
	  # number of matrix elements fits in L3 cache
	  $L2maxelmts = int ($L3_cache_size / 8);
      } else {
	  # number of matrix elements fits in L2 cache
	  $L2maxelmts = int ($L2_cache_size / 8);
      }
   }
   else
   {
      $L1maxelmts = -1;	# something is wrong
      $L2maxelmts = -1;
   }

   # select the best L0 code options for this case
   ($l0core, $cflags) = split(/\;/, $cases{$case});
   ($m0,$k0,$n0,$l0spopt) = split(/\,/, $l0core);

   # compile timer interface (main() code in C)
   $status = &compile_timer;
   next CASE if ($status != 0);

   # generate dummy general core code (just to get L1 code to compile)
   $status = &generate_core("-file ./src/mm_general.c -routine_name mm_general",
			    $prec, $matop, $alphatype,
			    "general", "",
			    1,1,1);
   next CASE if ($status != 0);

   # compile dummy general core code
   $status = &compile_code("mm_general","");
   next CASE if ($status != 0);

   # generate nofringe core code
   $status = &generate_core("-file ./src/mm_nofringe.c -routine_name mm_nofringe",
			    $prec, $matop, $alphatype,
			    "nofringe",$l0spopt,
			    $m0,$k0,$n0);
   next CASE if ($status != 0);

   # compile nofringe core code
   $status = &compile_code("mm_nofringe",$cflags);
   next CASE if ($status != 0);

   L1CODEUSE:
   foreach $l1codeuse (@l1codeuses)
   {
      print STDOUT "Testing $prec.$matop.$alphatype.$l1codeuse...\n"
	 if $verbosity > 0;
      
      # initialize current search status (best-so-far info)
      &init_best_stats;

      # create the blockings space of all blocking sizes;
      # this is the list of blocking sizes that HAS NOT
      # been benchmarked
      @keys_blockings_space = &create_blockings_space;

      # randomize the list of blocking sizes
      {
	 local(@temp);

	 @temp = ();
	 push(@temp,
	      splice(@keys_blockings_space,
		     rand(@keys_blockings_space), 1))
	    while @keys_blockings_space;

	 @keys_blockings_space = @temp;
      }

      $numpbms = scalar(@keys_blockings_space) * $numopts;

      if (-e "$perf_dir/$search_mode.$prec.$matop.$alphatype.$l1codeuse")
      {
	 # performance file already exist,
	 # so skip this case altogether

	 warn "***Nothing to be done for $prec.$matop.$alphatype.$l1codeuse.\n";

	 $numdone += scalar(@keys_blockings_space);
	 next L1CODEUSE;
      }

      # open and initialize output file for this option
      open(PERF_FILE,">$perf_dir/$perf_file")
	 || die "***Can't open $perf_dir/$perf_file.\n";
      printf(PERF_FILE &file_header_str);

      BLOCKING:
      while (scalar(@keys_blockings_space) > 0)
      {
	 # select a current set of blocking sizes
	 $key = shift(@keys_blockings_space);

	 ($m1,$k1,$n1) = split(/\,/,$key);
	 print STDOUT "\nTesting m1 = $m1, k1 = $k1, n1 = $n1...\n"
	    if $verbosity > 2;

	 # generate L1 source code with this blocking

	 if ($l1codeuse eq "fixed_nofringe_general")
	 {
	    # also generate fixed core code
	    $status = &generate_core("-file ./src/mm_fixed.c -routine_name mm_fixed",
				     $prec, $matop, $alphatype,
				     "fixed",$l0spopt,
				     $m0,$k0,$n0,
				     $m1,$k1,$n1);

	    # also compile fixed core code
	    $status = &compile_code("mm_fixed", $cflags)
	       if ($status == 0);
	 }

	 $generator_controls = " -gen_rout mm_general -gen_nf_rout mm_nofringe";
	 if ($l1codeuse eq "fixed_nofringe_general")
	 {
	    $generator_controls .= " -fixed_rout mm_fixed";
	 }

	 $generator_controls .= " -file ./src/mm_mult.c -routine_name mm_mult";

	 $status = &generate_level($generator_controls,
				   $prec, $matop, $alphatype,
				   $m1,$k1,$n1,
				   $m0, $k0, $n0,
				   $l0spopt)
	    if ($status == 0);

	 # compile generated source code
	 $status = &compile_code("mm_mult", $cflags)
	    if ($status == 0);

	 # link with timing routines to produce
	 # an executable timer program
	 $status = &link_timer
	    if ($status == 0);

	 # execute the timer program to benchmark
	 # the current blocking sizes
	 if ($status == 0)
	 {
	    $mflops_rate = &run_timing_test;
	 }
	 else
	 {
	    $mflops_rate = -1;
	 }

	 # delete the timer program to prevent
	 # subsequent failed cases from running it
	 # accidentally
	 &delete_timer;
	 
	 # do housekeeping work
	 $numdone++;
	 &update_best_stats;
	 &print_curr_stats;

      } # end loop for BLOCKING
      
      # close output file for this option
      printf(PERF_FILE &file_trailer_str);
      close(PERF_FILE);

      # rename file to indicate completion of this case
      rename("$perf_dir/$perf_file",
	     "$perf_dir/$search_mode.$prec.$matop.$alphatype.$l1codeuse");

   } # end loop for L1CODEUSE
} # end loop for CASE


##
## Termination
##

## Change back to working directory
chdir($working_dir);

chop($date = `date`);
print STDOUT "\n\nL1 Search ended on $hostname at $date\n\n";

exit(0);

##
## END
##




##
## sub create_L1_options_space
##
## Generate all points in the options space.  The options space
## is a subset of the following direct product:
##
## {cases} X {code use options}
##
## {code use options} is a (sub)set of L1 code uses specified by
## $l1_use_nofringe_general, $l1_use_fixed_nofringe_general.
##
## {cases} is a (sub)set of combinations of precision, matrix operation
## and alpha type which has some performance recorded in the L0
## top performance file.
##
## The returned value of this subroutine is the total number of
## points in the whole options space.
##
sub create_L1_options_space
{
   local($key);

   ##
   ## create array %cases
   ##

   %cases = ();

   open(L0TOP_FILE, "$working_dir/$l0_top_file")
      || die "***Can't open $working_dir/$l0_top_file.\n";

   while(<L0TOP_FILE>)
   {
      s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

      if (/^%/)
      {
	 # comments --- ignore
      }
      elsif (/^$/)
      {
	 # blank line --- ignore
      }
      else
      {
	 /^(\w+)\.(\w+)\.(\w+):\s+([0-9.]+)/;
	 # the fields are:
	 #   (1) precision, (2) matrix operation, (3) alpha type,
	 #   (4) performance

	 $key = "$1.$2.$3";

	 local($l0core, $cflags);

	 $_ = <L0TOP_FILE>;
	 while (substr($_,0,1) eq "#")
	 {
	    $_ =~ /^\# (.+) = (.*)$/;

	    if ($1 eq "L0CORE")
	    {
	       $l0core = $2;
	    }
	    elsif ($1 eq "cflags")
	    {
	       $cflags = $2;
	    }
	    else
	    {
	       warn "***Ignoring information $_.\n";
	    }

	    $_ = <L0TOP_FILE>;
	 }

	 $cases{$key} = "$l0core;$cflags";
      }
   }

   close(L0TOP_FILE);

   ##
   ## create array @l1codeuses
   ##
   
   @l1codeuses = ();

   if ($l1_use_nofringe_general)
   {
      push(@l1codeuses, ('nofringe_general'));
   }
   if ($l1_use_fixed_nofringe_general)
   {
      push(@l1codeuses, ('fixed_nofringe_general'));
   }

   ##
   ## return number of points in options space
   ##

   return(scalar(keys(%cases)) * scalar(@l1codeuses));
}



##
## sub init_best_stats
##
## Initializes the statistics for the best blocking sizes found
## so far.
##
sub init_best_stats
{
   $best_mflops_rate = 0;	# the maximum mflops_rate so far
   $best_blocking = "";		# and the corresponding blocking sizes
}


##
## sub update_best_stats
##
## Updates the statistics for the best blocking sizes found so
## far.
##
sub update_best_stats
{
   if ($mflops_rate >= $best_mflops_rate)
   {
      $best_mflops_rate = $mflops_rate;
      $best_blocking = "($m1,$k1,$n1)";
   }
}


##
## sub print_curr_stats
##
## Prints the current search statistics to STDOUT
##
sub print_curr_stats
{
   if ($verbosity > 1)
   {
      printf(STDOUT "Performance for $prec.$matop.$alphatype.$l1codeuse ($m1,$k1,$n1) : %7.1f MFLOPS.\n",
	     $mflops_rate);

      printf(STDOUT "Best Blocking for $prec.$matop.$alphatype.$l1codeuse so far : $best_blocking : %7.1f MFLOPS.\n",
	     $best_mflops_rate);

      printf(STDOUT "Done %d (%3.1f%%), %d (%3.1f%%) left; ",
	     $numdone, $numdone * 100 / $numpbms,
	     $numpbms - $numdone, ($numpbms - $numdone) * 100 / $numpbms);
      printf(STDOUT "Finishing by %s\n",
	     &ctime($^T + ($numpbms / $numdone) * (time() - $^T)));
      &flush(STDOUT);
   }
}


##
## sub file_header_str
##
## Returns a file header string to be printed at the beginning of
## the output file, documenting the parameters with which this
## search was conducted.
##
sub file_header_str
{
   local($output_str);
   local($opA, $opB);

   $opA = substr($matop,0,1);
   $opB = substr($matop,1,1);

   $output_str = "";
   
   $output_str .= "%%%% Performance from L1 search:\n";
   $output_str .= "%%%%\n";

   $output_str .= "%%%% (search started on $hostname at $date)\n";
   $output_str .= "%%%% File opened on ";
   $output_str .= `date`;
   $output_str .= "%%%%\n";

   $output_str .= "%%%% Precision:\t\t\t";
   $output_str .= "$prec\n";

   $output_str .= "%%%% Matrix Operations:\t\t";

   if ($opA eq "N")
   {
      $output_str .= "Normal-";
   }
   elsif ($opA eq "T")
   {
      $output_str .= "Transpose-";
   }

   if  ($opB eq "N")
   {
      $output_str .= "Normal\n";
   }
   elsif ($opB eq "T")
   {
      $output_str .= "Transpose\n";
   }

   $output_str .= "%%%% Alpha:\t\t\t";
   
   if ($alphatype eq "1")
   {
      $output_str .= "equals one (constant)\n";
   }
   elsif ($alphatype eq "c")
   {
      $output_str .= "arbitrary\n";
   }

   $output_str .= "%%%% Code Use:\t\t\t";
   $output_str .= "$l1codeuse\n";

   $output_str .= "%%%%\n";

   ##
   ## machine information
   ##

   $output_str .= "%%%% Number of single precision registers: $num_single_prec_registers\n";
   $output_str .= "%%%% Number of double precision registers: $num_double_prec_registers\n";
   $output_str .= "%%%%\n";
   $output_str .= "%%%% L1 cache size: $L1_cache_size bytes\n";
   if ($L2_cache_size > 0)
   {
      $output_str .= "%%%% L2 cache size: $L2_cache_size bytes\n";

      if ($L3_cache_size > 0)
      {
	 $output_str .= "%%%% L3 cache size: $L3_cache_size bytes\n";
      }
   }
   $output_str .= "%%%%\n";
   $output_str .= "%%%% Compiler: $cc\n";
   $output_str .= "%%%% Generator options: $generator_opts\n";
   $output_str .= "%%%% Timer arguments: $timer_args\n";
   $output_str .= "%%%%\n\n";

   ##
   ## environment information
   ##

   $output_str .= "# L0CORE = $l0core\n";
   $output_str .= "# cflags = $cflags\n";
   $output_str .= "\n";

   $output_str .= "%%%%   Mflops   \t m1  k1  n1\n";
   $output_str .= "\n";

   $output_str;			# return value
}



##
## sub file_trailer_str
##
## Returns a file trailer string to be printed at the end of
## the output file.
##
sub file_trailer_str
{
   local($output_str);

   $output_str  = "\n\n";
   $output_str .= "%%%%\n";

   $output_str .= "%%%% File closed on ";
   $output_str .= `date`;
   $output_str .= "%%%%\n";

   $output_str;			# return value
}



##
## sub create_blockings_space
##
## Generate all points in the blockings space.  The blockings
## space is the set of all triplets (m1,k1,n1) that will be
## used as the L1 blocking sizes for the generated
## code.
##
sub create_blockings_space
{
   # create array %blockings_space

   local(%blocking_space);

   %blockings_space = ();

   $sqsize = sqrt($L1maxelmts * ($fill_cache_percentage/100) / 3);

   foreach $m_ratio (@m_ratios)
   {
      foreach $k_ratio (@k_ratios)
      {
	 foreach $n_ratio (@n_ratios)
	 {
	    $m1 = int(0.5 + $m_ratio * $sqsize / $m0);
	    $k1 = int(0.5 + $k_ratio * $sqsize / $k0);
	    $n1 = int(0.5 + $n_ratio * $sqsize / $n0);

	    if (($m1 > 0) && ($k1 > 0) && ($n1 > 0))
	    {
	       $blockings_space{"$m1,$k1,$n1"} = 0;
	    }
	 }
      }
   }

   keys(%blockings_space);
}



##
## sub run_timing_test
##
## Executes the timer program to perform timing tests on the
## generated routine.  Returns the mflops-rating of the routine.
##
sub run_timing_test
{
   local($M, $K, $N);		# these are the matrix sizes

   local($mflops_rate);

   $sqsize = sqrt($L2maxelmts * ($fill_cache_percentage/100) / 3);

   $spadd = substr($l0spopt,0,1) - 1; # additional $K for software pipelining
   $M = $m0 * $m1 * int($sqsize / ($m0 * $m1));
   $K = ($k0 * $k1 + $spadd) * int($sqsize / ($k0 * $k1 + $spadd));
   $N = $n0 * $n1 * int($sqsize / ($n0 * $n1));

   $mflops_rate = &do_test;

   return(-1) if ($NO_ACTION);

   printf(PERF_FILE "%12.6f\t%3d %3d %3d\n",
	  $mflops_rate,
	  $m1,$k1,$n1);
   &flush(PERF_FILE);

   $mflops_rate;		# return value
}


