##
## SEARCH_READPARAMS.PL
##
## Read in parameter file for test configuration parameters.
##
## Written by:
##      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
##
##
## "Copyright (c) 1997 The Regents of the University of California.  All
## rights reserved."  Permission to use, copy, modify, and distribute
## this software and its documentation for any purpose, without fee, and
## without written agreement is hereby granted, provided that the above
## copyright notice and the following two paragraphs appear in all copies
## of this software.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
## THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
## SUCH DAMAGE.
## 
## THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
## PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
## CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
## ENHANCEMENTS, OR MODIFICATIONS.
## 


# The factor we assume that an L(n+1) cache is larger than
# an L(n) cache when the L(n+1) cache size parameter is
# not given.
$LN_factor=8;


sub read_machine_specs
{
   local($params_file) = @_;

   ##
   ## default values for machine specifications
   ##

   $path_PHiPAC = "";		# full path name for the PHiPAC directory

   $verbosity = 0;		# verbosity level (higher => wordier)
   $NO_ACTION = 0;		# 1 iff all unix commands are not executed
				# (use with verbosity to debug this script)

   $num_single_prec_registers = 0; # number of single precision registers
   $num_double_prec_registers = 0; # number of double precision registers

   $L1_cache_size = 0;		# byte size of L1 cache
   $L2_cache_size = 0;		# byte size of L2 cache
   $L3_cache_size = 0;		# byte size of L3 cache

   $cc = "";			# C compiler to be used

   $generator_opts = "";	# code generator options supplied for each call

   $timer_args = "-o 10 -i-auto"; # command-line arguments to be passed
				  # to the timer

   open(PARAMFILE,$params_file)
      || die "***Can't open machine specification file $params_file.\nUseage:\n$useage\n\n";

   $line = 0;
   
   while(<PARAMFILE>)
   {
      $line++;
      s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

      ##
      ## search configuration options
      ##

      if (m|^path_PHiPAC\s*=\s*(.*)/$|)
      {
	 $path_PHiPAC = $1;
      }
      elsif (m|^path_PHiPAC\s*=\s*(.*)$|)
      {
	 $path_PHiPAC = $1;
      }
      elsif (/^verbosity\s*=\s*(.*)$/)
      {
	 $verbosity = $1;
      }
      elsif (/^NO_ACTION$/)
      {
	 $NO_ACTION = ~0;
      }

      ##
      ## machine configuration information
      ##

      elsif (/^num_single_prec_registers\s*=\s*(.*)$/)
      {
	 $num_single_prec_registers = $1;
      }
      elsif (/^num_double_prec_registers\s*=\s*(.*)$/)
      {
	 $num_double_prec_registers = $1;
      }
      elsif (/^L1_cache_size\s*=\s*(.*)$/)
      {
	 $L1_cache_size = $1;
      }
      elsif (/^L2_cache_size\s*=\s*(.*)$/)
      {
	 $L2_cache_size = $1;
      }
      elsif (/^L3_cache_size\s*=\s*(.*)$/)
      {
	 $L3_cache_size = $1;
      }

      ##
      ## compiler, generator and timer information
      ##

      elsif (/^compiler\s*=\s*(.*)$/)
      {
	 $cc = $1;
      }
      elsif (/^generator_opts\s*=\s*(.*)$/)
      {
	 $generator_opts = $1;
      }
      elsif (/^timer_args\s*=\s*(.*)$/)
      {
	 $timer_args = $1;
      }

      ##
      ## housekeeping
      ##

      elsif (/^%/)
      {
	 # comments --- ignore
      }
      elsif (/^$/)
      {
	 # blank line --- ignore
      }
      else
      {
	 warn "*** Ignoring specification \"$_\"\n";
      }
   }

   close(PARAMFILE);


   ##
   ## check that all required variables have been defined
   ##

   $variables_OK = 1;

   if ($path_PHiPAC eq "")
   {
      warn "*** The 'path_PHiPAC' parameter must be defined.\n";
      $variables_OK = 0;
   }

   if ($num_single_prec_registers <= 0)
   {
      warn "*** The 'num_single_prec_registers' parameter must be defined as a positive integer.\n";
      $variables_OK = 0;
   }

   if ($num_double_prec_registers <= 0)
   {
      warn "*** The 'num_double_prec_registers' parameter must be defined as a positive integer.\n";
      $variables_OK = 0;
   }
   
   if ($cc eq "")
   {
      warn "*** The 'compiler' parameter must be defined.\n";
      $variables_OK = 0;
   }

   die "*** Can't proceed with search without properly defined parameters.\n\n"
      if ($variables_OK == 0);
}




sub read_l0_search_params
{
   local($params_file) = @_;

   ##
   ## default values for L0 search parameters
   ##

   $one_stage_holdstripe_A = 0;	# 1 iff "-sp1 -holdstripe A" is to be tested
   $one_stage_holdstripe_B = 0;	# 1 iff "-sp1 -holdstripe B" is to be tested
   $two_stage_load_mult = 0;	# 1 iff "-sp2lm" is to be tested
   $two_stage_mult_add = 0;	# 1 iff "-sp2ma" is to be tested
   $three_stage = 0;		# 1 iff "-sp3" is to be tested

   $alpha_equals_one = 0;	# 1 iff alpha=1 code is to be tested
   $alpha_arbitrary = 0;	# 1 iff alpha=c code is to be tested

   $matop_NN = 0;		# 1 iff NN is to be tested
   $matop_NT = 0;		# 1 iff NT is to be tested
   $matop_TN = 0;		# 1 iff TN is to be tested
   $matop_TT = 0;		# 1 iff TT is to be tested

   $auto_blockings = 1;		# 1 iff blocking sizes in the search
				# space are to be automatically generated
   @k0_set = ();		# set of values for k0 to benchmark
   $blockings_file = "";	# name of file containing register blocking
				# sizes to be searched if $auto_blockings
				# is 0

   $auto_matrixsizes = 1;	# 1 iff matrix sizes used for benchmarking
				# are to be automatically generated
   $fill_cache_percentage = 80;	# percentage cache to fill when timing
				# for large matrices
   $matrixsizes_file = "";	# name of file containing matrix sizes to be
				# used for benchmarking if $auto_matrixsizes
				# is 0
   $use_timing = "";		# timing to be used; either "fatdot" or "almstsq"
   $do_L0_gen_frng = 0;        # 1 iff we are to do the L0 gen fringe code search.

   open(PARAMFILE,$params_file)
      || die "Can't open L0 search parameters file $params_file.\nUseage:\n$useage\n\n";

   $line = 0;
   
   while(<PARAMFILE>)
   {
      $line++;
      s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

      ##
      ## L0 search options
      ##

      ## matrix operation options

      if (/^matop_NN$/)
      {
	 $matop_NN = 1;
      }
      elsif (/^matop_NT$/)
      {
	 $matop_NT = 1;
      }
      elsif (/^matop_TN$/)
      {
	 $matop_TN = 1;
      }
      elsif (/^matop_TT$/)
      {
	 $matop_TT = 1;
      }

      ## alpha-value options

      elsif (/^alpha_equals_one$/)
      {
	 $alpha_equals_one = 1;
      }
      elsif (/^alpha_arbitrary$/)
      {
	 $alpha_arbitrary = 1;
      }

      ## software pipelining options

      elsif (/^one_stage_holdstripe_A$/)
      {
	 $one_stage_holdstripe_A = 1;
      }
      elsif (/^one_stage_holdstripe_B$/)
      {
	 $one_stage_holdstripe_B = 1;
      }
      elsif (/^two_stage_load_mult$/)
      {
	 $two_stage_load_mult = 1;
      }
      elsif (/^two_stage_mult_add$/)
      {
	 $two_stage_mult_add = 1;
      }
      elsif (/^three_stage$/)
      {
	 $three_stage = 1;
      }

      ## blocking space options

      elsif (/^auto_blockings$/)
      {
	 $auto_blockings = 1;
      }
      elsif (/^k0_set\s*=\s*(.*)$/)
      {
	 @k0_set = split(/\,/, $1);
      }
      elsif (/^blockings_file\s*=\s*(.*)$/)
      {
	 $blockings_file = $1;
	 $auto_blockings = 0;
      }

      ## matrix sizes options

      elsif (/^auto_matrixsizes$/)
      {
	 $auto_matrixsizes = 1;
      }
      elsif (/^fill_cache_percentage\s*=\s*(.*)$/)
      {
	 $fill_cache_percentage = $1;
      }
      elsif (/^matrixsizes_file\s*=\s*(.*)$/)
      {
	 $matrixsizes_file = $1;
	 $auto_matrixsizes = 0;
      }

      ## use timing options

      elsif (/^use_fatdot$/)
      {
	 $use_timing = "fatdot";
      }
      elsif (/^use_almstsq$/)
      {
	 $use_timing = "almstsq";
      }

      elsif (/^do_L0_gen_frng$/)
      {
	  $do_L0_gen_frng = 1;
      }

      elsif (/^benchmark_l0_out_of_l2/)
      {
	  $benchmark_l0_out_of_l2 = 1;
      }

      elsif (/^benchmark_l0_out_of_l3/)
      {
	  $benchmark_l0_out_of_l3 = 1;
      }


      ##
      ## housekeeping
      ##

      elsif (/^%/)
      {
	 # comments --- ignore
      }
      elsif (/^$/)
      {
	 # blank line --- ignore
      }
      else
      {
	 warn "*** Ignoring specification \"$_\"\n";
      }
   }

   close(PARAMFILE);


   ## check that L1 cache size is defined

   if ($L1_cache_size <= 0)
   {
      die "*** The 'L1_cache_size' parameter must be defined as a positive integer.\n";
   }

   ## assign L2 cache size = $LN_factor * L1 cache size if necessary
   
   if ($L2_cache_size <= 0)
   {
      $L2_cache_size = $LN_factor * $L1_cache_size;
   }

   ## assign L3 cache size = $LN_factor * L2 cache size if necessary

   if ($L3_cache_size <= 0)
   {
      $L3_cache_size = $LN_factor * $L2_cache_size;
   }



}



sub read_l1_search_params
{
   local($params_file) = @_;

   ##
   ## default values for L1 search parameters
   ##

   $l1_use_nofringe_general = 0; # 1 iff L1 code use nofringe and
				   # general L0 code is to be tested
   $l1_use_fixed_nofringe_general = 0; # 1 iff L1 code use fixed, nofringe and
				       # general L0 code is to be tested

   @m_ratios = ();		# list of L1 ratios in the M direction
   @k_ratios = ();		# list of L1 ratios in the K direction
   @n_ratios = ();		# list of L1 ratios in the N direction

   $fill_cache_percentage = 80;	# percentage cache to fill when timing
				# for large matrices


   open(PARAMFILE,$params_file)
      || die "Can't open L1 search parameters file $params_file.\nUseage:\n$useage\n\n";

   $line = 0;
   
   while(<PARAMFILE>)
   {
      $line++;
      s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

      ##
      ## L1 search options
      ##

      ## code use options

      if (/^l1_use_nofringe_general$/)
      {
	 $l1_use_nofringe_general = 1;
      }
      elsif (/^l1_use_fixed_nofringe_general$/)
      {
	 $l1_use_fixed_nofringe_general = 1;
      }

      ## block ratios options

      elsif (/^m_ratios\s*=\s*(.*)$/)
      {
	 @m_ratios = split(/\,/, $1);
      }
      elsif (/^k_ratios\s*=\s*(.*)$/)
      {
	 @k_ratios = split(/\,/, $1);
      }
      elsif (/^n_ratios\s*=\s*(.*)$/)
      {
	 @n_ratios = split(/\,/, $1);
      }

      elsif (/^fill_cache_percentage\s*=\s*(.*)$/)
      {
	 $fill_cache_percentage = $1;
      }


      elsif (/^benchmark_l1_out_of_l3/)
      {
	  $benchmark_l1_out_of_l3 = 1;
      }


      ##
      ## housekeeping
      ##

      elsif (/^%/)
      {
	 # comments --- ignore
      }
      elsif (/^$/)
      {
	 # blank line --- ignore
      }
      else
      {
	 warn "*** Ignoring specification \"$_\"\n";
      }
   }

   close(PARAMFILE);


   ## check that L1 cache size is defined

   if ($L1_cache_size <= 0)
   {
      die "*** The 'L1_cache_size' parameter must be defined as a positive integer.\n";
   }

   ## assign L2 cache size = $LN_factor * L1 cache size if necessary
   
   if ($L2_cache_size <= 0)
   {
      $L2_cache_size = $LN_factor * $L1_cache_size;
   }

   ## assign L3 cache size = $LN_factor * L2 cache size if necessary

   if ($L3_cache_size <= 0)
   {
      $L3_cache_size = $LN_factor * $L2_cache_size;
   }

}



sub read_l2_search_params
{
   local($params_file) = @_;

   ##
   ## default values for L2 search parameters
   ##

   @m_ratios = ();		# list of L2 ratios in the M direction
   @k_ratios = ();		# list of L2 ratios in the K direction
   @n_ratios = ();		# list of L2 ratios in the N direction

   $fill_cache_percentage = 80;	# percentage cache to fill when timing
				# for large matrices


   open(PARAMFILE,$params_file)
      || die "Can't open L2 search parameters file $params_file.\nUseage:\n$useage\n\n";

   $line = 0;
   
   while(<PARAMFILE>)
   {
      $line++;
      s/^\s*(.*)\s*$/$1/;	# remove leading and trailing whitespaces

      ##
      ## L2 search options
      ##

      ## block ratios options

      if (/^m_ratios\s*=\s*(.*)$/)
      {
	 @m_ratios = split(/\,/, $1);
      }
      elsif (/^k_ratios\s*=\s*(.*)$/)
      {
	 @k_ratios = split(/\,/, $1);
      }
      elsif (/^n_ratios\s*=\s*(.*)$/)
      {
	 @n_ratios = split(/\,/, $1);
      }

      elsif (/^fill_cache_percentage\s*=\s*(.*)$/)
      {
	 $fill_cache_percentage = $1;
      }

      ##
      ## housekeeping
      ##

      elsif (/^%/)
      {
	 # comments --- ignore
      }
      elsif (/^$/)
      {
	 # blank line --- ignore
      }
      else
      {
	 warn "*** Ignoring specification \"$_\"\n";
      }
   }

   close(PARAMFILE);

   ## check that L2 cache sizes is defined

   if ($L2_cache_size <= 0)
   {
      die "*** The 'L2_cache_size' parameter must be defined as a positive integer.\n";
   }

   ## assign L3 cache size = $LN_factor * L2 cache size if necessary
   
   if ($L3_cache_size <= 0)
   {
      $L3_cache_size = $LN_factor * $L2_cache_size;
   }
}


1; # module successfully loaded

