#! /usr/local/bin/perl

##
## SEARCH_SORT.PL
##
## Perl script to sort and select the best blocking sizes 
## for matrix-matrix multiplication code generated by MM_GEN.
##
## Written by:
##      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
##
##
## "Copyright (c) 1997 The Regents of the University of California.  All
## rights reserved."  Permission to use, copy, modify, and distribute
## this software and its documentation for any purpose, without fee, and
## without written agreement is hereby granted, provided that the above
## copyright notice and the following two paragraphs appear in all copies
## of this software.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
## THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
## SUCH DAMAGE.
## 
## THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
## PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
## CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
## ENHANCEMENTS, OR MODIFICATIONS.
## 


##
## Initialization
##

## Add the directory of search to @INC
{
   local(@temp);

   @temp = split(/\//,$0);
   pop(@temp);
   push(@INC, join("/",@temp));
}

## Initialize test configuration parameters.
require "search_init.pl";
$search_mode = "";
$useage = "$0 <sort level> <precision>";

if (($ARGV[0] =~ /^(L[0-9].*)$/) && 
    (($ARGV[1] eq "single") || ($ARGV[1] eq "double")))
{
   $sort_level = $1;
   $prec = $ARGV[1];
}
else
{
   die "Useage:\n\t$useage\n\n";
}


chop($hostname = `hostname`);
chop($date = `date`);
print STDOUT "\n\n$sort_level Sort started on $hostname on $date\n\n";
&flush(STDOUT);

## Change to temporary directory
chdir($perf_dir);

if (-e "$perf_dir/$sort_level" . "_top.$prec")
{
   # top performance file already exist,
   # so skip this case altogether
  
   warn "***Nothing to be done for $sort_level.\n";
}
else
{
   $ls_call = "ls $sort_level.$prec.*";
   $ls_out = `$ls_call`;
   die "***Can't list performance files.\n"
      if ($? != 0);

   %cases = ();
   foreach $filename (sort(split(/\n/, $ls_out)))
   {
      @fields = split(/\./, $filename);
      $key = $fields[1] . "." . $fields[2] . "." . $fields[3];
      $cases{$key} = "";
   }

   ##
   ## Loop through all cases
   ## and conduct sorting for each case.
   ##

   CASE:
   foreach $case (sort(keys(%cases)))
   {
      $ls_call = "ls $sort_level.$case*";
      $ls_out = `$ls_call`;
      if ($? != 0)
      {
	 warn "***Can't list performance files for $case.\n";
	 next CASE;
      }
      print STDOUT $ls_out;

      # initialize sort variables
      $best_mflops_rate = 0;
      @best_options = ();
      $best_blocking = "";
      @best_info_lines = ();

      FILENAME:
      foreach $filename (sort(split(/\n/, $ls_out)))
      {
	 # open file for this option
	 open(PERF_FILE, "$perf_dir/$filename");
	 if ($? != 0)
	 {
	    warn "***Can't open $filename.\n";
	    next FILENAME;
	 }

	 @info_lines = ();

	 while(<PERF_FILE>)
	 {
	    s/^\s*(.*)\s*$/$1/; # remove leading and trailing whitespaces

	    if (/^%/)
	    {
	       # comments --- ignore
	    }
	    elsif (/^$/)
	    {
	       # blank line --- ignore
	    }
	    elsif (/^\#(.+)$/)
	    {
	       # information line --- store up and output to
	       # top performance file

	       push(@info_lines, "#$1");
	    }
	    else
	    {
	       /^([0-9.]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9]+)/;

	       if ($1 > $best_mflops_rate)
	       {
		  $best_mflops_rate = $1;
		  @best_options = split(/\./,$filename);
		  shift(@best_options);
		  shift(@best_options);
		  shift(@best_options);
		  shift(@best_options);
		  $best_blocking = "$2,$3,$4";
		  @best_info_lines = @info_lines;
	       }
	    }
	 }

	 # close file for this option
	 close(PERF_FILE);

      } # end loop for FILENAME


      $cases{$case}  = "$best_mflops_rate\n";

      $cases{$case} .= "# $sort_level = ";
      $cases{$case} .= join(",", split(/\s+/, $best_blocking));
      if (scalar(@best_options) > 0)
      {
	 $cases{$case} .= ",";
	 $cases{$case} .= join(",", @best_options);
      }
      $cases{$case} .= "\n";

      $cases{$case} .= join("\n", @best_info_lines);

   } # end loop for CASE


   open(TOP_FILE, ">$perf_dir/$sort_level" . "_top.$prec" )
      || die "***Can't open $perf_dir/$sort_level" . "_top.$prec.\n";
   printf(TOP_FILE &file_header_str);

   foreach $case (sort(keys(%cases)))
   {
      printf(TOP_FILE "$case:\t" . $cases{$case});
      printf(TOP_FILE "\n\n");
   }

   close(TOP_FILE);
}


##
## Termination
##

## Change back to working directory
chdir($working_dir);

chop($date = `date`);
print STDOUT "\n\n$sort_level Sort ended on $hostname at $date\n\n";

exit(0);

##
## END
##


##
## sub file_header_str
##
## Returns a file header string to be printed to the output file
## documenting the parameters with which this search was conducted.
##
sub file_header_str
{
   local($output_str);

   $output_str = "";
   
   $output_str .= "%%%% Best $sort_level blocking sizes:\n";
   $output_str .= "%%%%\n";

   $output_str .= "%%%% (sort started on $hostname on $date)\n";
   $output_str .= "%%%%\n";

   $output_str .= "%%%%\n\n";

   $output_str;			# return value
}
