#! /usr/local/bin/perl

##
## SEARCH_START.PL
##
## Perl script to prepare the environment for the search scripts.
## Creates subdirectories needed by the search scripts, and
## compiles IPM, RPRF and the code generators MM_CGEN and MM_LGEN.
## 
## Written by:
##      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
##
##
## "Copyright (c) 1997 The Regents of the University of California.  All
## rights reserved."  Permission to use, copy, modify, and distribute
## this software and its documentation for any purpose, without fee, and
## without written agreement is hereby granted, provided that the above
## copyright notice and the following two paragraphs appear in all copies
## of this software.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
## THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
## SUCH DAMAGE.
## 
## THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
## PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
## CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
## ENHANCEMENTS, OR MODIFICATIONS.
## 


##
## Initialization
##

## Add the directory of search to @INC
{
   local(@temp);
   print STDOUT $0;
   @temp = split(/\//,$0);
   pop(@temp);
   push(@INC, join("/",@temp));
}

## Initialize test configuration parameters.
require "search_init.pl";
$useage = "$0 <machine specification file> [clean]";

## Read search parameters.
require "search_readparams.pl";
&read_machine_specs($ARGV[0]);

## Tell the whole world we've started...
chop($hostname = `hostname`);
chop($date = `date`);
print STDOUT "\n\nSearch initialization started on $hostname on $date\n\n";
&flush(STDOUT);

&mkdir_subdir;
chdir($tmp_dir);

&compile_IPM;
&compile_RPRF;
&compile_core_generator;
&compile_level_generator;

chdir($working_dir);

## Tell the whole world we've finished! (what a miracle)

chop($date = `date`);
print STDOUT "\n\nSearch initialization completed on $hostname at $date\n\n";

exit(0);

##
## END
##




##
## sub mkdir_subdir
##
## Creates subdirectories for benchmarking.
##
sub mkdir_subdir
{
   local($cc_call);

   print STDOUT "Creating subdirectories...\n"
      if $verbosity > 0;

   $cc_call = "rm -rf $tmp_dir; mkdir $tmp_dir";
   if (!( (-e $perf_dir) && (-d $perf_dir) ) ||
       $ARGV[1] eq "clean")
   {
      $cc_call .= "; rm -rf $perf_dir; mkdir $perf_dir";
   }
   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   if (!$NO_ACTION)
   {
      system($cc_call);
      die "***Can't create temporary directory (exit status $?)\n"
	 if ($? != 0);
   }

   $cc_call = "mkdir $tmp_dir/include $tmp_dir/lib $tmp_dir/src $tmp_dir/obj $tmp_dir/bin $tmp_dir/doc";
   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   if (!$NO_ACTION)
   {
      system($cc_call);
      die "***Can't create installation directories (exit status $?)\n"
	 if ($? != 0);
   }
}



##
## sub compile_IPM
##
## Compiles the IPM library.
##
sub compile_IPM
{
   local($cc_call);
   local($compiler, $flags);
   
   print STDOUT "Compiling IPM...\n" if $verbosity > 0;

   ($compiler, $flags) = split(/\s+/,$cc,2);

   $cc_call = "env CC=$compiler CFLAGS='$flags' $path_PHiPAC/$dir_IPM/configure --srcdir=$path_PHiPAC/$dir_IPM --prefix=$tmp_dir";
   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   if (!$NO_ACTION)
   {
      system($cc_call);
      die "***Can't configure IPM (exit status $?)\n"
	 if ($? != 0);
   }

   $cc_call = "$make install";
   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   if (!$NO_ACTION)
   {
      system($cc_call);
      die "***Can't compile IPM (exit status $?)\n"
	 if ($? != 0);
   }
}



##
## sub compile_RPRF
##
## Compiles the RPRF library.
##
sub compile_RPRF
{
   local($cc_call);
   local($compiler, $flags);

   print STDOUT "Compiling RPRF...\n" if $verbosity > 0;

   ($compiler, $flags) = split(/\s+/,$cc,2);

   $cc_call = "env CC=$compiler CFLAGS='$flags' $path_PHiPAC/$dir_RPRF/configure --srcdir=$path_PHiPAC/$dir_RPRF --prefix=$tmp_dir";
   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   if (!$NO_ACTION)
   {
      system($cc_call);
      die "***Can't configure RPRF (exit status $?)\n"
	 if ($? != 0);
   }

   $cc_call = "$make install";
   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   if (!$NO_ACTION)
   {
      system($cc_call);
      die "***Can't compile RPRF (exit status $?)\n"
	 if ($? != 0);
   }
}



##
## sub compile_core_generator
##
## Compiles the core-code generator.
##
sub compile_core_generator
{
   local($cc_call);

   print STDOUT "Compiling core-code generator...\n" if $verbosity > 0;

   $cc_call = "$cc -O -o $core_generator $path_PHiPAC/$dir_GEN/$core_generator.c";

   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   if (!$NO_ACTION)
   {
      system($cc_call);
      die "***Can't compile core-code generator (exit status $?)\n"
	 if ($? != 0);
   }
}


##
## sub compile_level_generator
##
## Compiles the level-code generator.
##
sub compile_level_generator
{
   local($cc_call);

   print STDOUT "Compiling level-code generator...\n" if $verbosity > 0;

   $cc_call = "$cc -O -o $level_generator $path_PHiPAC/$dir_GEN/$level_generator.c";

   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   if (!$NO_ACTION)
   {
      system($cc_call);
      die "***Can't compile level-code generator (exit status $?)\n"
	 if ($? != 0);
   }
}

