##
## SERACH_SUPP.PL
##
## Supporting routines for the search script.  These routines
## provide the interface between the search script, IPM/RPRF,
## MM_GEN, and the C compiler.
##
## Written by:
##      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
##
##
## "Copyright (c) 1997 The Regents of the University of California.  All
## rights reserved."  Permission to use, copy, modify, and distribute
## this software and its documentation for any purpose, without fee, and
## without written agreement is hereby granted, provided that the above
## copyright notice and the following two paragraphs appear in all copies
## of this software.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
## THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
## SUCH DAMAGE.
## 
## THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
## PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
## CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
## ENHANCEMENTS, OR MODIFICATIONS.
## 



##
## sub compile_timer
##
## Compiles the timer routine which is the timer interface for the
## matrix multiplication code produced by mm_gen
##
sub compile_timer
{
   local($cc_call);

   print STDOUT "Compiling timer object code...\n" if $verbosity > 0;

   $cc_call = "$cc $cflags -I$path_PHiPAC/$dir_IPM -I$path_PHiPAC/$dir_RPRF -o ./obj/$timer_obj -c $path_PHiPAC/$dir_SEARCH/$timer_src";

   $cc_call .= " -DSEARCH_MODE_$search_mode";

   if ($alphatype eq "1")
   {
      $cc_call .= " -DALPHA_EQUALS_ONE";
   }
   elsif ($alphatype eq "c")
   {
      $cc_call .= " -DALPHA_ARBITRARY";
   }

   if ($prec eq "single")
   {
      $cc_call .= " -DSINGLE_PREC";
   }
   elsif ($prec eq "double")
   {
      $cc_call .= " -DDOUBLE_PREC";
   }

   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   return(0) if $NO_ACTION;

   system($cc_call);
   if ($? != 0)
   {
      warn "***Can't compile timer object code (exit status $?)\n";
      return $?;
   }

   return(0);		# OK status
}



##
## sub generate_core
##
## Calls the code generator to generate source code of
## matrix-matrix multiplication routine with the given register
## blocking sizes.
##
sub generate_core
{
   local($generator_controls,
	 $prec, $matop, $alphatype,
	 $codetype, $softpipeopt,
	 $m0,$k0,$n0, $m1,$k1,$n1) = @_;
   local($generator_call);

   print STDOUT "Generating $codetype core code...\n" if $verbosity > 2;

   $opA = substr($matop,0,1);
   $opB = substr($matop,1,1);

   $generator_call = "./$core_generator $generator_opts -prec $prec -opA $opA -opB $opB -alpha $alphatype";

   if ($softpipeopt ne "")
   {
      if (substr($softpipeopt,0,1) eq "1")
      {
	 $generator_call .= " -sp 1 -holdstripe ";
	 $generator_call .= substr($softpipeopt,1,1);
      }
      else
      {
	 $generator_call .= " -sp ";
	 $generator_call .= $softpipeopt;
      }
   }

   $generator_call .= " -l0 $m0 $k0 $n0";

   if ($codetype eq "nofringe")
   {
      $generator_call .= " -no_fringes";
   }

   if ($codetype eq "fixed")
   {
      $generator_call .= " -l1 $m1 $k1 $n1";
   }

   $generator_call .= " $generator_controls";

   print STDOUT ">>> $generator_call\n" if $verbosity > 5;

   return(0) if $NO_ACTION;

   system($generator_call);
   warn "***Can't generate $codetype core code for m0 = $m0, k0 = $k0, n0 = $n0 (exit status $?)\n"
      if ($? != 0);

   return($?);			# return value
}



##
## sub generate_level
##
## Calls the level generator to generate source code of
## matrix-matrix multiplication routine with the given
## L1 blocking sizes.
##
sub generate_level
{
   local($generator_controls,
	 $prec, $matop, $alphatype,
	 $m1,$k1,$n1,
	 $m0,$k0,$n0,
	 $l0spopt) = @_;
   local($generator_call);

   print STDOUT "Generating level code...\n" if $verbosity > 2;

   $opA = substr($matop,0,1);
   $opB = substr($matop,1,1);

   $generator_call = "./$level_generator $generator_opts -prec $prec -opA $opA -opB $opB -alpha $alphatype";

   if ($l0spopt ne "")
   {
      $generator_call .= " -sp ";

      if (substr($l0spopt,0,1) eq "1")
      {
	 $generator_call .= "1";
      }
      else
      {
	 $generator_call .= $l0spopt;
      }
   }
   
   $generator_call .= " -l0 $m0 $k0 $n0";
   $generator_call .= " -l1 $m1 $k1 $n1";

   $generator_call .= " $generator_controls";

   print STDOUT ">>> $generator_call\n" if $verbosity > 5;

   return(0) if $NO_ACTION;

   system($generator_call);
   warn "***Can't generate level code for m1 = $m1, k1 = $k1, n1 = $n1 (exit status $?)\n"
      if ($? != 0);

   return($?);			# return value
}



##
## sub compile_code
##
## Compiles the generated code.
##
sub compile_code
{
   local($name, $cflags) = @_;
   local($cc_call);

   print STDOUT "Compiling code for $name...\n" if $verbosity > 2;

   $cc_call = "$cc $cflags -o ./obj/$name.o -c ./src/$name.c";

   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   return(0) if $NO_ACTION;

   system($cc_call);
   warn "***Can't compile code for $name (exit status $?)\n"
      if ($? != 0);

   return($?);			# return value
}



##
## sub link_timer
##
## Links the compiled code with the timer routine
## to produce an executable timer program.
##
sub link_timer
{
   local($cc_call);

   print STDOUT "Linking timer...\n" if $verbosity > 2;

   $cc_call = "$cc $cflags -I./include -L./lib -o ./bin/$timer_prg ./obj/$timer_obj ./obj/mm_mult.o";

   if (($search_mode eq "L1") || ($search_mode eq "L2"))
   {
      $cc_call .= " ./obj/mm_general.o";

      if ($l1codeuse eq "nofringe_general")
      {
	 $cc_call .= " ./obj/mm_nofringe.o";
      }
      elsif ($l1codeuse eq "fixed_nofringe_general")
      {
	 $cc_call .= " ./obj/mm_nofringe.o ./obj/mm_fixed.o";
      }
   }

   if ($search_mode eq "L2")
   {
      $cc_call .= " ./obj/mm_l1.o";
   }

   $cc_call .= " -lm -lipm -lrprf";

   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   return(0) if $NO_ACTION;

   system($cc_call);
   warn "***Can't link timer (exit status $?)\n"
      if ($? != 0);

   return($?);			# return value
}



##
## sub do_test
##
## Executes the timer program with the given blocking sizes and
## matrix sizes, and returns the maximum performance (MFLOPS)
## reading.
##
## This subroutine is called only from sub run_timing_test.
##
sub do_test
{
   local($timer_output, @timer_outputs);
   local($timer_call);

   print STDOUT "Running timer...\n" if $verbosity > 2;

   if ($matop eq "NN")
   {
      $timer_call = "./bin/$timer_prg -S-max-rate $timer_args $M $K $N $M $K $K $N $M $N";
   }
   elsif ($matop eq "NT")
   {
      $timer_call = "./bin/$timer_prg -S-max-rate $timer_args $M $K $N $M $K $N $K $M $N";
   }
   elsif ($matop eq "TN")
   {
      $timer_call = "./bin/$timer_prg -S-max-rate $timer_args $M $K $N $K $M $K $N $M $N";
   }
   elsif ($matop eq "TT")
   {
      $timer_call = "./bin/$timer_prg -S-max-rate $timer_args $M $K $N $K $M $N $K $M $N";
   }
   else
   {
      $timer_call = "";		# something is wrong
   }

   print STDOUT ">>> $timer_call\n" if $verbosity > 5;

   $timer_output = `$timer_call` if !$NO_ACTION;

   return(-1) if $NO_ACTION;

   # extract last value of output from timer,
   # which is the maximum performance reading (MFLOPS)
   # because of the option "-S-max-rate"

   @timer_outputs = split(/\s+/,$timer_output);
   $timer_output = pop(@timer_outputs);

   if ($timer_output == 0)
   {
      warn "***Failed test case\t(M = $M, K = $K, N = $N)\n";
   }
   else
   {
      print STDOUT "<<< (M = $M, K = $K, N = $N) : $timer_output\n"
	 if $verbosity > 4;
   }

   $timer_output;		# return value
}


##
## sub delete_timer
##
## Delete executable timer program.
##
sub delete_timer
{
   local($cc_call);

   print STDOUT "Deleting timer...\n" if $verbosity > 2;

   $cc_call = "rm -f ./bin/$timer_prg";

   print STDOUT ">>> $cc_call\n" if $verbosity > 5;

   return(0) if $NO_ACTION;

   system($cc_call);
   warn "***Can't delete timer (exit status $?)\n"
      if ($? != 0);

   return($?);			# return value
}




1; # module successfully loaded
