#! /usr/local/bin/perl

##
## TIME_GEMM.pl
##
## Perl script to benchmark the PHiPAC GEMM library.
##
## Written by:
##      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
##
##
## "Copyright (c) 1997 The Regents of the University of California.  All
## rights reserved."  Permission to use, copy, modify, and distribute
## this software and its documentation for any purpose, without fee, and
## without written agreement is hereby granted, provided that the above
## copyright notice and the following two paragraphs appear in all copies
## of this software.
## 
## IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
## THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
## SUCH DAMAGE.
## 
## THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
## PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
## CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
## ENHANCEMENTS, OR MODIFICATIONS.
## 

##
## Initialization
##

## Add the directory of search to @INC
{
   local(@temp);

   @temp = split(/\//,$0);
   pop(@temp);
   push(@INC, join("/",@temp));
}

## Initialize test configuration parameters.
require "search_init.pl";
$search_mode = "";
$useage = "$0 <machine specification file> <single|double>";

## Read search parameters.
require "search_readparams.pl";
&read_machine_specs($ARGV[0]);
$prec = $ARGV[1];

## Load supporting routines.
require "search_supp.pl";


$gemm = "";

if ($prec eq "single")
{
   $gemm = "sgemm";
   $maxsize = int(sqrt((16 * 1024 * 1024 / 4) / 3));
}
elsif ($prec eq "double")
{
   $gemm = "dgemm";
   $maxsize = int(sqrt((16 * 1024 * 1024 / 8) / 3));
}
else
{
   die "***Precision must be either \"single\" or \"double\".\n";
}


if (!(-e "$working_dir/PHiPAC_$gemm/lib$gemm.a"))
{
   # PHiPAC GEMM library doesn't exist

   die "***PHiPAC GEMM library lib$gemm.a does not exists.\n";
}



chop($hostname = `hostname`);
chop($date = `date`);
print STDOUT "\n\nTime_GEMM started on $hostname on $date\n\n";
&flush(STDOUT);

## Change to temporary directory
chdir($tmp_dir);

##
## Compile timers
##

print STDOUT "Compiling PHiPAC GEMM timers...\n" if $verbosity > 0;

$cc_call = "cp" ;
$cc_call .= " $path_PHiPAC/$dir_SEARCH/timer_$gemm.c";
$cc_call .= " ./src/";
print STDOUT ">>> $cc_call\n" if $verbosity > 5;
system($cc_call);
warn "***Can't copy PHiPAC GEMM timer source codes (exit status $?)\n"
   if ($? != 0);

$cc_call = "$cc -O -I$path_PHiPAC/$dir_IPM -I$path_PHiPAC/$dir_RPRF -o ./obj/timer_$gemm.o -c ./src/timer_$gemm.c";
print STDOUT ">>> $cc_call\n" if $verbosity > 5;
system($cc_call);
die "***Can't compile timer object code (exit status $?)\n" 
   if ($? != 0);
   
$cc_call = "$cc -O -I./include -L$working_dir/PHiPAC_$gemm -L./lib -o ./bin/timer_$gemm ./obj/timer_$gemm.o -l$gemm -lrprf -lipm";
print STDOUT ">>> $cc_call\n" if $verbosity > 5;
system($cc_call);
die "***Can't link timer with GEMM library (exit status $?)\n" 
   if ($? != 0);


##
## open and initialize output file
##

open(PERF_FILE,">$perf_dir/$perf_file")
   || die "***Can't open $perf_dir/$perf_file.\n";
printf(PERF_FILE &file_header_str);


##
## Benchmark on square matrices with size <= $maxsize
##

@sizes = ();
for ($size = 10; $size <= $maxsize; )
{
   push(@sizes, $size);
   # Use multiples of 10 up to 300x300, multiples
   # of 100 thereafter.
   if ($size < 300) {
       $size += 10;
   } else  {
       $size += 100;
   }
}

$numpbms = scalar(@sizes);
$numdone = 0;

foreach $size (@sizes)
{
   print STDOUT "\nTesting size = $size...\n";

   # call GEMM for Normal Normal alpha=beta=1 matrix multiply
   # for square matrices of size $size
   $timer_call = "./bin/timer_$gemm -S-max-rate $timer_args n n $size $size $size 1 1";

   print STDOUT ">>> $timer_call\n" if $verbosity > 5;

   $timer_output = `$timer_call` if !$NO_ACTION;

   # extract last value of output from timer,
   # which is the maximum performance reading (MFLOPS)
   # because of the option "-S-max-rate"

   @timer_outputs = split(/\s+/,$timer_output);
   $mflops_rate = pop(@timer_outputs);

   if ($mflops_rate == 0)
   {
      warn "***Failed test case $size\n";
   }
   else
   {
      printf(PERF_FILE "%3d\t%12.6f\n",
	     $size, $mflops_rate);
      &flush(PERF_FILE);

      $numdone++;

      printf(STDOUT "Size $size : %7.1f MFLOPS\n", $mflops_rate);

      printf(STDOUT "Done %d (%3.1f%%), %d (%3.1f%%) left; ",
	     $numdone, $numdone * 100 / $numpbms,
	     $numpbms - $numdone, ($numpbms - $numdone) * 100 / $numpbms);
      printf(STDOUT "Finishing by %s\n",
	     &ctime($^T + ($numpbms / $numdone) * (time() - $^T)));
      &flush(STDOUT);
   }
}


##
## close output file
##

printf(PERF_FILE &file_trailer_str);
close(PERF_FILE);

# rename file to indicate completion of this case
rename("$perf_dir/$perf_file",
       "$perf_dir/PERF_$gemm");


##
## Termination
##

## Change back to working directory
chdir($working_dir);

print STDOUT "\n\nTo run GEMM timer, type:\n";
print STDOUT "\t./tmp/bin/timer_$gemm <timer arguments>\n";


print STDOUT "\n\nTime_GEMM ended on $hostname at $date\n\n";

exit(0);

##
## END
##



##
## sub file_header_str
##
## Returns a file header string to be printed at the beginning of
## the output file.
##
sub file_header_str
{
   local($output_str);

   $output_str = "";
   
   $output_str .= "%%%% Performance for PHiPAC lib$gemm.a:\n";
   $output_str .= "%%%%\n";

   $output_str .= "%%%% File opened on ";
   $output_str .= `date`;
   $output_str .= "%%%%\n";

   ##
   ## machine information
   ##

   $output_str .= "%%%% Number of single precision registers: $num_single_prec_registers\n";
   $output_str .= "%%%% Number of double precision registers: $num_double_prec_registers\n";
   $output_str .= "%%%%\n";
   $output_str .= "%%%% L1 cache size: $L1_cache_size bytes\n";
   if ($L2_cache_size > 0)
   {
      $output_str .= "%%%% L2 cache size: $L2_cache_size bytes\n";

      if ($L3_cache_size > 0)
      {
	 $output_str .= "%%%% L3 cache size: $L3_cache_size bytes\n";
      }
   }
   $output_str .= "%%%%\n";
   $output_str .= "%%%% Compiler: $cc\n";
   $output_str .= "%%%% Generator options: $generator_opts\n";
   $output_str .= "%%%% Timer arguments: $timer_args\n";
   $output_str .= "%%%%\n\n";

   $output_str .= "%%%% Size\t   Mflops   \n";
   $output_str .= "\n";

   $output_str;			# return value
}


##
## sub file_trailer_str
##
## Returns a file trailer string to be printed at the end of
## the output file.
##
sub file_trailer_str
{
   local($output_str);

   $output_str  = "\n\n";
   $output_str .= "%%%%\n";

   $output_str .= "%%%% File closed on ";
   $output_str .= `date`;
   $output_str .= "%%%%\n";

   $output_str;			# return value
}



