/*
** timer_sgemm.c
**      RPRF interface for timing PHiPAC SGEMM
**
** Written by:
**      CheeWhye Chin <cheewhye@icsi.berkeley.edu>
**
**
** "Copyright (c) 1997 The Regents of the University of California.  All
** rights reserved."  Permission to use, copy, modify, and distribute
** this software and its documentation for any purpose, without fee, and
** without written agreement is hereby granted, provided that the above
** copyright notice and the following two paragraphs appear in all copies
** of this software.
** 
** IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
** FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
** THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
** 
** THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
** INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
** PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
** CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
** ENHANCEMENTS, OR MODIFICATIONS.
** 
*/


#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

#include "rprf.h"


extern void
sgemm(char*, char*,		/* transA, transB */
      int*, int*, int*,	        /* M, N, K */
      float*,			/* alpha */
      void*, int*,		/* A, Astride */
      void*, int*,		/* B, Bstride */
      float*,			/* beta */
      void*, int*);		/* C, Cstride */

/* These SGEMM arguments work for FORTRAN programs;
 * since we are calling from C, the roles of A and B
 * have to be switched; see MM_MULT below.
 */



/* Simple matrix initialize routine */

void
init_float(size_t len, float val, float *ptr)
{
  size_t i;

  for (i=0; i<len; i++)
    *ptr++ = val;
}



/* Main program */

int
main(int argc, char *argv[])
{
  float *A, *B, *cC;
  float alpha, beta;

  char *opA = "", *opB = "";
  size_t M = 0, K = 0, N =0;
  size_t rowA = 0, colA = 0, rowB =0, colB = 0, rowC = 0, colC = 0;
  size_t strideA = 0, strideB = 0, strideC = 0;

  if (argc < 5)
    {
      printf("Usage: %s [<perf_args>] <opA> <opB> <M> <K> <N> <alpha> <beta>\n",
	     argv[0]);
      printf("\tcalls SGEMM for matrix multiplication\n");
      printf("\t  alpha*opA(A) * opB(B) + beta*C  -->  C\n");
      printf("\twhere:\n");
      printf("\t\t <opA>      is \"Normal\" or \"Transpose\";\n");
      printf("\t\t <opB>      is \"Normal\" or \"Transpose\";\n");
      printf("\t\t <alpha>    is a float precision scalar;\n");
      printf("\t\t <beta>     is a float precision scalar;\n");
      printf("\t\t opA(A)     is a MxK matrix;\n");
      printf("\t\t opB(A)     is a KxN matrix;\n");
      printf("\t\t C          is a MxN matrix;\n");
      printf("\t\t <perf_args> are arguments passed to RPRF.\n\n");
	
      exit(0);
    }

  rprf_parse_args(argc, argv, "ssiiiff", &opA, &opB, &M, &K, &N,
		  &alpha, &beta);

  A = rprf_malloc_vf(M*K);
  B = rprf_malloc_vf(K*N);
  cC = rprf_malloc_vf(M*N);

  init_float(M*K, 1.0f, A);
  init_float(K*N, 1.0f, B);

  if ((opA[0] == 'n') || (opA[0] == 'N'))
    {
      if ((opB[0] == 'n') || (opB[0] == 'N'))
	{
	  rowA = M; colA = K;
	  rowB = K; colB = N;
	  rowC = M; colC = N;
	  strideA = K; strideB = N; strideC = N;
	}
      else if ((opB[0] == 't') || (opB[0] == 'T'))
	{
	  rowA = M; colA = K;
	  rowB = N; colB = K;
	  rowC = M; colC = N;
	  strideA = K; strideB = K; strideC = N;
	}
    }
  else if ((opA[0] == 't') || (opA[0] == 'T'))
    {
      if ((opB[0] == 'n') || (opB[0] == 'N'))
	{
	  rowA = K; colA = M;
	  rowB = K; colB = N;
	  rowC = M; colC = N;
	  strideA = M; strideB = N; strideC = N;
	}
      else if ((opB[0] == 't') || (opB[0] == 'T'))
	{
	  rowA = K; colA = M;
	  rowB = N; colB = K;
	  rowC = M; colC = N;
	  strideA = M; strideB = K; strideC = N;
	}
    }


/* Run the performance test */

#define MM_MULT                       \
    sgemm(opB, opA,                   \
	  (int*)&N,(int*)&M,(int*)&K, \
	  &alpha,                     \
	  (void*)B,(int*)&strideB,    \
	  (void*)A,(int*)&strideA,    \
	  &beta,                      \
	  (void*)cC,(int*)&strideC)

/* These SGEMM arguments work for FORTRAN programs;
 * since we are calling from C, the roles of A and B
 * have to be switched; see declaration of dgemm above.
 */

    RPRF_BODY(MM_MULT, M*N*(2*K-1))

    return 0;
}
